
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/* traceback system stolen from Sasha Vasko <sasha at aftercode.net> */

/*!
 * \file agl_error.c
 *
 * \brief manage internal errors
 * \ingroup tools
 *
 * \author Stephane APIOU
*/

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include "config.h"
#include "agl_private.h"

#ifdef HAVE_EXECINFO_H
#include <execinfo.h>
char **__backtrace_symbols(void **, int);
#endif

#ifdef HAVE_LINK_H
#include <link.h>
#endif

#include <signal.h>


//! this structure defines the error message structure
typedef struct agl_err
{
	char string[AGL_ERROR_SIZE];	//!< the error message is this string
	struct agl_err *error;	//!< points to the next error message or NULL
}
agl_err;

//! this structure defines the memory zone allocation
// The structure definition is the following:
//      - pnext: pointer to next item
//      - pprev: pointer to previous item
//      - bpointer: the pointer to the beginning of the zone
//      - size: the size of the useable zone
//      - magic1 : contains "cabedafe" if allocated, "73170537" if freed
//      - zone : the useable zone (size defined by malloc rounded to upper 4)
//      - magic2 : the magic of the end "cabedafe" if allocated, "73170537" if freed
//      - string: a string that describes the file, line and function allocated

typedef struct agl_memalloc
{
	struct agl_memalloc *pnext;
	struct agl_memalloc *pprev;
	void *bpointer;
	unsigned long size;
	unsigned long magic1;
}
agl_memalloc;


static agl_memalloc *agl_root_mem = NULL;

static agl_err *agl_root_err = NULL;

static char *agl_error_string;

#ifdef HAVE_SIGCONTEXT
#include <asm/sigcontext.h>
#endif

typedef struct _agl_proc_tables
{

#ifdef HAVE_LINK_H
	struct r_debug *debug;
#endif

#ifdef _ELF_H
	Elf32_Sym *symbols;
	Elf32_Word *sym_hash;
#endif
	int sym_ent_size;
	int sym_ent_num;

	char *strings;
	int str_tabl_size;

	void *pltgot;

}
agl_proc_tables;

char *agl_elf_start = (char *) 0x08048000;


//! treats a fatal error

/*! this function treats a fatal error. this function exits the program
 * \param string the error message
 * \param file the file name that creates this error
 * \param line the line of C code that creates this error
 * \param function the function that create this error
 *
 * \return void we don't return
 */

void
_agl_fatal(char *string, char *file, int line, char *function)
{
	if (function)
		fprintf(stderr, "AGL FATAL ERROR : %s in file %s at line %d, function %s\n", string, file, line, function);
	else
		fprintf(stderr, "AGL FATAL ERROR : %s in file %s at line %d\n", string, file, line);
	exit(1);
}

//! treats a common error

/*! this function treats a common error and keeps this error in an error buffer
 * the user can print the buffer at any time
 * \param string the error message
 * \param file the name of the file that generates this error
 * \param line the line in this file
 * \param function the function that generates this error
 *
 * \return void nothing to return
 * \note this function exits if there is not enough memory to keep the error
 */
void
_agl_error(char *string, char *file, int line, char *function)
{
	agl_err *perr;

	if (agl_error_string != NULL)
		free(agl_error_string);
	agl_error_string = NULL;
	if ((perr = malloc(sizeof(agl_err))) == NULL)
	{
		_agl_fatal(string, file, line, function);
	}
	perr->error = agl_root_err;
	agl_root_err = perr;
	if (function)
		snprintf(perr->string, AGL_ERROR_SIZE, "AGL Error: %s in %s at line %d, function %s\n", string, file, line, function);
	else
		snprintf(perr->string, AGL_ERROR_SIZE, "AGL Error: %s in %s at line %d\n", string, file, line);
}

//! this function asserts

/*! this function tests an assertion. if the assertion fails, it prints
 * the assertion that fails
 * \param expression the assertion to test
 * \param file the name of the file where the assertion is
 * \param line the line in the file
 * \param function the function that test this assertion
 *
 * \return int return not null
 * \note this function exits if there is not enough memory to keep the error
 */

int
_agl_assert(char *expression, char *file, int line, char *function)
{
	agl_err *perr;

	if (agl_error_string != NULL)
		free(agl_error_string);
	agl_error_string = NULL;
	if ((perr = malloc(sizeof(agl_err))) == NULL)
	{
		_agl_fatal(expression, file, line, function);
	}
	perr->error = agl_root_err;
	agl_root_err = perr;
	if (function)
		snprintf(perr->string, AGL_ERROR_SIZE, "AGL Assert: %s failed in %s at line %d, function %s\n", expression, file, line,
			 function);
	else
		snprintf(perr->string, AGL_ERROR_SIZE, "AGL Assert: %s failed in %s at line %d\n", expression, file, line);
	return 1;
}

//! this function prints only a debugging message

/*! this function prints a debugging message. This not an error message. this is not fatal.
 * \param fmt the formatting string
 * \param  ... a list of parameters separated by commas
 *
 * \return void  nothing
 * \note this function exits if there is not enough memory to keep the error
 */

void
agl_msg(char *fmt, ...)
{
	va_list ap;
	agl_err *perr;

	if (agl_error_string != NULL)
		free(agl_error_string);
	agl_error_string = NULL;
	if ((perr = malloc(sizeof(agl_err))) == NULL)
	{
		agl_memerr();
	}
	perr->error = agl_root_err;
	agl_root_err = perr;
	strcpy(perr->string, "AGL Msg: ");
	va_start(ap, fmt);
	vsnprintf(perr->string + 9, AGL_ERROR_SIZE - sizeof(perr->string) - 1, fmt, ap);
	va_end(ap);
}

//! flush error buffer

/*! this function flushes the error buffer.
 * \param void nothing
 *
 * \return void nothing
 */

void
agl_flusherror(void)
{
	agl_err *perr, *nperr;

	for (perr = agl_root_err; perr != NULL; perr = nperr)
	{
		nperr = perr->error;
		free(perr);
	}
	agl_root_err = NULL;
}

//! this function only gets the last error message

/*! this function gets the last error message. if this message is the result of
 * multiple cascading errors, only the last message is taken. the previous messages
 * are dropped.
 * \param void nothing
 *
 * \return char * the error message string
 */

char *
agl_getlasterror(void)
{
	char *str;

	if (agl_error_string != NULL)
		free(agl_error_string);
	agl_error_string = NULL;
	if (agl_root_err == NULL)
	{
		if ((str = malloc(1)) == NULL)
			agl_memerr();
		else
			*str = 0;
	}
	else
	{
		if ((str = malloc(strlen(agl_root_err->string) + 1)) == NULL)
			agl_memerr();
		else
			strcpy(str, agl_root_err->string);
	}
	agl_error_string = str;
	agl_flusherror();
	return str;
}

//! this function gets all the error messages

/*! this function gets all the error messages. the error buffers are clened afterwards
 * \param void nothing
 *
 * \return char * the error messages.
 */

char *
agl_geterror(void)
{
	char *str, *pch, *pce;
	int len;
	agl_err *perr;

	if (agl_error_string != NULL)
		free(agl_error_string);
	agl_error_string = NULL;
	for (len = 0, perr = agl_root_err; perr != NULL; perr = perr->error)
		len += strlen(perr->string);
	if ((str = malloc(len + 1)) == NULL)
		agl_memerr();
	for (pch = str, perr = agl_root_err, pce = perr->string; perr != NULL; perr = perr->error, pce = perr->string, pch--)
		while ((*pch++ = *pce++)) ;
	*pch = 0;
	agl_error_string = str;
	agl_flusherror();
	return str;
}

//! this function is useful to debug malloc errors

/*! it detects out of bound problems double free errors and lost block errors
 * \warning this function is VERY memory and time consuming. assumes pointers and 
 * long are 4 octets long
 * \param size a size to allocate
 * \param file the source file that called the function
 * \param line the line in the file that called the function
 * \param function the function name that called this function
 * \return void * NULL if no mem else a pointer to the allocated zone
 */

void *
_agl_malloc(unsigned long size, char *file, int line, char *function)
{
	agl_memalloc *pmem;
	int usz, stsz, sz;
	unsigned char str[120];
	unsigned long *lptr;

	if (function)
		snprintf(str, 120, "%ld in %s at line %d, function %s", size, file, line, function);
	else
		snprintf(str, 120, "%ld in %s at line %d", size, file, line);

	stsz = ((strlen(str) + 4) & ~3);
	usz = ((size + 3) & ~3);

	sz = sizeof(agl_memalloc) + stsz + usz + sizeof(unsigned long);
	pmem = (agl_memalloc *) malloc(sz);

	if (pmem == NULL)
	{
		_agl_error("out of memory", file, line, function);
		return NULL;
	}

	strcpy((char *) pmem + sizeof(agl_memalloc) + usz + sizeof(unsigned long), str);
	pmem->bpointer = (void *) pmem;
	pmem->size = size;
	pmem->magic1 = 0XCABEDAFE;
	lptr = (unsigned long *) ((char *) pmem + sizeof(agl_memalloc) + usz);	/* magic 2 */
	*lptr = 0XCABEDAFE;

	if (agl_root_mem == NULL)
		pmem->pnext = NULL;
	else
	{
		pmem->pnext = agl_root_mem;
		agl_root_mem->pprev = pmem;
	}
	agl_root_mem = pmem;
	pmem->pprev = NULL;

	return (char *) pmem + sizeof(agl_memalloc);
}

//! this function is useful to debug calloc errors

/*! it detects out of bound problems double free errors and lost block errors
 * \warning this function is VERY memory and time consuming. assumes pointers and 
 * long are 4 octets long
 * \param nmem the number of record to allocated
 * \param size a size to allocate
 * \param file the source file that called the function
 * \param line the line in the file that called the function
 * \param function the function name that called this function
 * \return void * NULL if no mem else a pointer to the allocated zone
 */

void *
_agl_calloc(unsigned long nmem, unsigned long size, char *file, int line, char *function)
{
	agl_memalloc *pmem;
	int usz, stsz, sz;
	unsigned char str[120];
	unsigned long *lptr;

	if (function)
		snprintf(str, 120, "%ld in %s at line %d, function %s", size, file, line, function);
	else
		snprintf(str, 120, "%ld in %s at line %d", size, file, line);

	stsz = ((strlen(str) + 4) & ~3);
	usz = ((size * nmem + 3) & ~3);

	sz = sizeof(agl_memalloc) + stsz + usz + sizeof(unsigned long);
	pmem = (agl_memalloc *) calloc(1, sz);

	if (pmem == NULL)
	{
		_agl_error("out of memory", file, line, function);
		return NULL;
	}

	strcpy((char *) pmem + sizeof(agl_memalloc) + usz + sizeof(unsigned long), str);
	pmem->bpointer = (void *) pmem;
	pmem->size = size * nmem;
	pmem->magic1 = 0XCABEDAFE;
	lptr = (unsigned long *) ((char *) pmem + sizeof(agl_memalloc) + usz);	/* magic 2 */
	*lptr = 0XCABEDAFE;

	if (agl_root_mem == NULL)
		pmem->pnext = NULL;
	else
	{
		pmem->pnext = agl_root_mem;
		agl_root_mem->pprev = pmem;
	}
	agl_root_mem = pmem;
	pmem->pprev = NULL;

	return (char *) pmem + sizeof(agl_memalloc);
}

//! this function is useful to debug realloc errors

/*! it detects out of bound problems double free errors and lost block errors
 * \warning this function is VERY memory and time consuming. assumes pointers and 
 * long are 4 octets long
 * \param ptr the pointer to the memory location
 * \param size a size to allocate
 * \param file the source file that called the function
 * \param line the line in the file that called the function
 * \param function the function name that called this function
 * \return void * NULL if no mem else a pointer to the allocated zone
 */

void *
_agl_realloc(void *ptr, unsigned long size, char *file, int line, char *function)
{
	agl_memalloc *pmem, *pmem1;
	int usz, stsz, sz;
	unsigned char str[180], *pt;
	unsigned long *lptr;

	// test memory integrity

	if (function)
		snprintf(str, 120, "%ld in %s at line %d, function %s", size, file, line, function);
	else
		snprintf(str, 120, "%ld in %s at line %d", size, file, line);

	stsz = ((strlen(str) + 4) & ~3);

	pmem = (agl_memalloc *) ((char *) ptr - sizeof(agl_memalloc));

	usz = ((pmem->size + 3) & ~3);

	lptr = (unsigned long *) ((char *) pmem + sizeof(agl_memalloc) + usz);	/* magic 2 */
	pt = (void *) (lptr + 1);

	if (pmem->magic1 == 0X73170537)
		goto err1;
	if (pmem->magic1 != 0XCABEDAFE)
		goto err;
	if (pmem->bpointer != pmem)
		goto err;

	if (*lptr == 0X73170537)
		goto err1;
	if (*lptr != 0XCABEDAFE)
		goto err;

	if (pmem->pprev != NULL)
		pmem->pprev->pnext = pmem->pnext;
	else
		agl_root_mem = pmem->pnext;

	if (pmem->pnext != NULL)
		pmem->pnext->pprev = pmem->pprev;

	pmem->pnext = pmem->pprev = NULL;
	pmem->magic1 = 0X73170537;
	*lptr = 0X73170537;

	if (size == 0)
		return NULL;

	usz = ((size + 3) & ~3);

	sz = sizeof(agl_memalloc) + stsz + usz + sizeof(unsigned long);
	pmem1 = (agl_memalloc *) realloc(pmem, sz);

	if (pmem1 == NULL)
	{
		_agl_error("out of memory", file, line, function);
		return NULL;
	}

	strcpy((char *) pmem1 + sizeof(agl_memalloc) + usz + sizeof(unsigned long), str);
	pmem1->bpointer = (void *) pmem1;
	pmem1->size = size;
	pmem1->magic1 = 0XCABEDAFE;
	lptr = (unsigned long *) ((char *) pmem1 + sizeof(agl_memalloc) + usz);	/* magic 2 */
	*lptr = 0XCABEDAFE;

	if (agl_root_mem == NULL)
		pmem1->pnext = NULL;
	else
	{
		pmem1->pnext = agl_root_mem;
		agl_root_mem->pprev = pmem1;
	}
	agl_root_mem = pmem1;
	pmem1->pprev = NULL;

	return (char *) pmem1 + sizeof(agl_memalloc);
      err:
	if (function)
		fprintf(stderr, "memory corruption in %s at line %d, function %s\n", file, line, function);
	else
		fprintf(stderr, "memory corruption in %s at line %d\n", file, line);
	fprintf(stderr, "memory allocated : size %s\n", pt);
	exit(1);
      err1:
	if (function)
		fprintf(stderr, "memory already freed in %s at line %d, function %s\n", file, line, function);
	else
		fprintf(stderr, "memory already freed in %s at line %d\n", file, line);
	exit(1);
}

//! this function is useful to debug free errors

/*! it detects out of bound problems double free errors and lost block errors
 * \warning this function is VERY memory and time consuming.
 * \param ptr the pointer to the memory location
 * \param file the source file that called the function
 * \param line the line in the file that called the function
 * \param function the function name that called this function
 * \return void nothing
 */

void
_agl_free(void *ptr, char *file, int line, char *function)
{
	agl_memalloc *pmem;
	int usz;
	int stsz;
	unsigned char *pt;
	unsigned long *lptr;

	// test memory integrity

	if (ptr == NULL)
	{
		fprintf(stderr, "freeing NULL pointer in %s at line %d, function %s\n", file, line, function);
		exit(1);
	}

	pmem = (agl_memalloc *) ((char *) ptr - sizeof(agl_memalloc));

	usz = ((pmem->size + 3) & ~3);

	lptr = (unsigned long *) ((char *) pmem + sizeof(agl_memalloc) + usz);	/* magic 2 */
	pt = (char *) (lptr + 1);

	if (pmem->magic1 == 0X73170537)
		goto err1;
	if (pmem->magic1 != 0XCABEDAFE)
		goto err;

	if (pmem->bpointer != pmem)
		goto err;

	if (*lptr == 0X73170537)
		goto err1;
	if (*lptr != 0XCABEDAFE)
		goto err;

	if (pmem->pprev != NULL)
		pmem->pprev->pnext = pmem->pnext;
	else
		agl_root_mem = pmem->pnext;
	if (pmem->pnext != NULL)
		pmem->pnext->pprev = pmem->pprev;
	pmem->pnext = pmem->pprev = NULL;
	pmem->magic1 = 0X73170537;
	*lptr = 0X73170537;
	stsz = strlen(pt);
	if (function)
		snprintf(pt, stsz, "%ld in %s at line %d, function %s", pmem->size, file, line, function);
	else
		snprintf(pt, stsz, "%ld in %s at line %d", pmem->size, file, line);

	free(pmem);
	return;
      err:
	if (function)
		fprintf(stderr, "memory corruption in %s at line %d, function %s\n", file, line, function);
	else
		fprintf(stderr, "memory corruption in %s at line %d\n", file, line);
	fprintf(stderr, "memory allocated size %s\n", pt);
	exit(1);
      err1:
	if (function)
		fprintf(stderr, "memory already freed in %s at line %d, function %s\n", file, line, function);
	else
		fprintf(stderr, "memory already freed in %s at line %d\n", file, line);
	fprintf(stderr, "memory allocated size %s\n", pt);
	exit(1);

}

//! this function detects memory area not freed at the end of adgali

/*! this is always memory allocated by adgali not free by it (memory leakage)
 * \return void nothing
 */

int
agl_memgarbage(agl_container * handle)
{
	agl_memalloc *pmem, *npmem;
	unsigned char *pt;
	unsigned long *lptr;
	int usz;
	int ret;

	handle = NULL;
	ret = 1;
	for (pmem = agl_root_mem; pmem != NULL; pmem = npmem)
	{
		usz = ((pmem->size + 3) & ~3);
		lptr = (unsigned long *) ((char *) pmem + sizeof(agl_memalloc) + usz);	/* magic 2 */
		pt = (char *) (lptr + 1);
		fprintf(stderr, "memory allocated size %s\n", pt);

		if ((pmem->magic1 != 0XCABEDAFE) || (pmem->bpointer != pmem) || (*lptr != 0XCABEDAFE))
		{
			fprintf(stderr, "moreover, memory is corrupted\n");
			exit(1);
		}

		npmem = pmem->pnext;
		free(pmem);
		ret = 0;
	}
	return ret;
}

static void
agl_get_proc_tables(agl_proc_tables * ptabs)
{
#ifdef _ELF_H
	ElfW(Dyn) * dyn;

	memset(ptabs, 0x00, sizeof(agl_proc_tables));
	for (dyn = _DYNAMIC; dyn != NULL && dyn->d_tag != DT_NULL; ++dyn)
		switch (dyn->d_tag)
		{
			/* symbols */
		case DT_SYMTAB:
			ptabs->symbols = (Elf32_Sym *) dyn->d_un.d_ptr;
			break;
		case DT_HASH:
			ptabs->sym_hash = (Elf32_Word *) dyn->d_un.d_ptr;
			break;
		case DT_SYMENT:
			ptabs->sym_ent_size = dyn->d_un.d_val;
			break;
			/* strings */
		case DT_STRTAB:
			ptabs->strings = (char *) dyn->d_un.d_ptr;
			break;
		case DT_STRSZ:
			ptabs->str_tabl_size = dyn->d_un.d_val;
			break;
			/* debug info */
		case DT_DEBUG:
			ptabs->debug = (struct r_debug *) dyn->d_un.d_ptr;
			break;
			/* GOT/PLT */
		case DT_PLTGOT:
			ptabs->pltgot = (void *) dyn->d_un.d_ptr;
			break;
		}
	if (ptabs->sym_hash != NULL)
		ptabs->sym_ent_num = *(ptabs->sym_hash + 1);

#endif
}

static char *agl_unknown = "unknown";

static char *
agl_find_func_symbol(agl_proc_tables * ptabs, void *addr, long *offset)
{
#ifdef _ELF_H
	int i;
	long min_offset = 0x0fffffff, curr_offset;
	char *selected = agl_unknown;
	Elf32_Sym *ptr32 = ptabs->symbols + 1;
	Elf32_Addr addr32 = (unsigned int) addr;
	Elf64_Sym *ptr64 = (Elf64_Sym *) (ptabs->symbols + 1);
	Elf64_Addr addr64 = (unsigned int) addr;

	*offset = -1;
	if (ptabs->symbols == NULL || ptabs->strings == NULL)
		return agl_unknown;

	if (ptabs->sym_ent_size == sizeof(Elf32_Sym))
	{
		for (i = 1; i < ptabs->sym_ent_num; i++)
		{
			if (ptr32->st_value <= addr32 && ELF32_ST_TYPE(ptr32->st_info) == STT_FUNC)
			{
				curr_offset = addr32 - ptr32->st_value;
				if ((curr_offset < min_offset) && (curr_offset > 0) && ((unsigned long) curr_offset < ptr32->st_size))
				{
					selected = ptabs->strings + (ptr32->st_name);
					min_offset = curr_offset;
				}
			}
			ptr32++;
		}
	}
	else if (ptabs->sym_ent_size == sizeof(Elf64_Sym))
	{
		for (i = 1; i < ptabs->sym_ent_num; i++)
		{
			if (ptr64->st_value <= addr64 && ELF64_ST_TYPE(ptr64->st_info) == STT_FUNC)
			{
				curr_offset = addr64 - ptr64->st_value;
				if ((curr_offset < min_offset) && (curr_offset > 0) && ((unsigned long) curr_offset < ptr32->st_size))
				{
					selected = ptabs->strings + (ptr64->st_name);
					min_offset = curr_offset;
				}
			}
			ptr64++;
		}
	}
	*offset = min_offset;
	return selected;
#else
	*offset = 0;
	return agl_unknown;
#endif
}

static void
agl_print_lib_list(agl_proc_tables * ptabs)
{
#ifdef _ELF_H
	if (ptabs->debug != NULL)
	{
		struct link_map *plm = ptabs->debug->r_map;

		fprintf(stderr, " Loaded dynamic libraries :\n");
		for (; plm != NULL; plm = plm->l_next)
		{
			if (plm->l_addr != 0x00 && plm->l_name != NULL)
				fprintf(stderr, "   [0x%8.8X]:[%s]\n", plm->l_addr, plm->l_name);
		}
	}
#endif
}

#ifdef HAVE_SIGCONTEXT
static void
agl_print_signal_context(struct sigcontext *psc)
{
	if (psc)
	{
		int i;

		fprintf(stderr, " Signal Context :\n");
		fprintf(stderr, "  Registers\n");
		i = 0;		/* to avoid warnings */
#if defined(_ASMi386_SIGCONTEXT_H)
		fprintf(stderr, "   EAX: 0x%8.8lX", (unsigned long) (psc->eax));
		fprintf(stderr, "  EBX: 0x%8.8lX", (unsigned long) (psc->ebx));
		fprintf(stderr, "  ECX: 0x%8.8lX", (unsigned long) (psc->ecx));
		fprintf(stderr, "  EDX: 0x%8.8lX\n", (unsigned long) (psc->edx));
		fprintf(stderr, "   ESP: 0x%8.8lX", (unsigned long) (psc->esp));
		fprintf(stderr, "  EBP: 0x%8.8lX", (unsigned long) (psc->ebp));
		fprintf(stderr, "  EIP: 0x%8.8lX\n", (unsigned long) (psc->eip));
		fprintf(stderr, "   ESI: 0x%8.8lX", (unsigned long) (psc->esi));
		fprintf(stderr, "  EDI: 0x%8.8lX\n", (unsigned long) (psc->edi));
#else
#if defined(_ASMAXP_SIGCONTEXT_H) || defined(__ASM_MIPS_SIGCONTEXT_H) || defined(__ASM_SH_SIGCONTEXT_H)
		for (; i < 32; i++)
			fprintf(stderr, "  #%d: 0x%8.8lX\n", i, (unsigned long) (psc->sc_reg[i]));
#else
#if defined(_ASM_PPC_SIGCONTEXT_H)
		for (; i < 32; i++)
			fprintf(stderr, "  #%d: 0x%8.8lX\n", i, (unsigned long) (psc->regs->gpr[i]));
#endif
#endif
#endif
	}
}
#endif

static void
agl_print_my_backtrace(agl_proc_tables * ptabs, long *ebp, long *esp, long *eip)
{
#if defined(_ASMi386_SIGCONTEXT_H)
	int frame_no = 0;
	char **dummy;
	long offset;
	char *func_name = NULL;

	fprintf(stderr, " Stack Backtrace :\n");
	if (ebp < (long *) 0x08074000	/* stack can't possibly go below that */
		)
	{

		fprintf(stderr, "  heh, looks like we've got corrupted stack, so no backtrace for you.\n");
		if (eip != NULL)
		{
			fprintf(stderr, "  all I can say is that we failed at 0x%lX", (unsigned long) eip);
			func_name = agl_find_func_symbol(ptabs, (void *) eip, &offset);
			if (func_name != agl_unknown)
				fprintf(stderr, " in [%s+0x%lX(%lu)]", func_name, offset, offset);
			else
			{
#ifdef HAVE_EXECINFO_H
				dummy = backtrace_symbols((void **) &eip, 1);
				func_name = *dummy;
				if (*func_name != '[')
					fprintf(stderr, " in %s()", func_name);
#endif
			}
			fprintf(stderr, "\n");
		}
		return;
	}
	esp = eip;
	fprintf(stderr, "   FRAME       NEXT FRAME  FUNCTION\n");
	while (ebp != NULL && ebp > (long *) 0x08074000	/* stack can't possibly go below that */
		)
	{
		frame_no++;
		fprintf(stderr, "   0x%8.8lX", (unsigned long) ebp);
		fprintf(stderr, "  0x%8.8lX", (unsigned long) *(ebp));
		fprintf(stderr, "  0x%8.8lX", (unsigned long) esp);
		if (*(ebp) == 0)
			fprintf(stderr, "  [program entry point]\n");
		else
		{
			func_name = agl_find_func_symbol(ptabs, (void *) esp, &offset);
			if (func_name == agl_unknown && frame_no == 1 && eip != NULL)	/* good fallback for current frame */
				func_name = agl_find_func_symbol(ptabs, (void *) eip, &offset);

			if (func_name == agl_unknown)
			{
#ifdef HAVE_EXECINFO_H
				dummy = backtrace_symbols((void **) &esp, 1);
				func_name = *dummy;
				if (*func_name != '[')
					fprintf(stderr, "  [%s]", func_name);
				else
#endif
					fprintf(stderr, "  [some silly code]");
			}
			else
				fprintf(stderr, "  [%s+0x%lX(%lu)]", func_name, offset, offset);
			fprintf(stderr, "\n");
		}
		esp = (long *) *(ebp + 1);
		ebp = (long *) *(ebp);
	}
#endif
}

#define MAX_CALL_DEPTH 32	/* if you change it here  - then change below as well!!! */

#ifndef HAVE_SIGCONTEXT

static long
agl_get_call_list(long **call_list, int nb)
{
	int i;

	call_list[0] = NULL;
	i = 0;

#if defined(__GNUC__)

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(2) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(2)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(3) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(3)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(4) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(4)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(5) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(5)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(6) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(6)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(7) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(7)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(8) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(8)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(9) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(9)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(10) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(10)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(11) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(11)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(12) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(12)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(13) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(13)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(14) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(14)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(15) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(15)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(16) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(16)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(17) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(17)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(18) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(18)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(19) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(19)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(20) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(20)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(21) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(21)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(22) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(22)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(23) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(23)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(24) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(24)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(25) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(25)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(26) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(26)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(27) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(27)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(28) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(28)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(29) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(29)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(30) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(30)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(31) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(31)) == NULL)
		goto done;

	if (i >= (nb - 1))
		goto done;
	if (__builtin_frame_address(32) == NULL)
		goto done;
	if ((call_list[i++] = __builtin_return_address(32)) == NULL)
		goto done;

	call_list[i] = NULL;
      done:
#endif
	return i;
}

static void
agl_print_simple_backtrace(agl_proc_tables * ptabs)
{
#if defined(__GNUC__)
	int call_no = 0;
	long *call_list[MAX_CALL_DEPTH];

	agl_get_call_list(call_list, MAX_CALL_DEPTH);
	if (call_list[0] == NULL)
		return;
	agl_get_proc_tables(ptabs);
	fprintf(stderr, " Call Backtrace :\n");
	fprintf(stderr, " CALL#: ADDRESS:    FUNCTION:\n");
	while (call_list[call_no] != NULL)
	{
		long offset = 0;
		char *func_name = NULL;

		fprintf(stderr, " %5u  0x%8.8lX", call_no, (unsigned long) (call_list[call_no]));

		func_name = agl_find_func_symbol(ptabs, (void *) call_list[call_no], &offset);
		if (func_name == agl_unknown)
		{
#ifdef HAVE_EXECINFO_H
			char **dummy = backtrace_symbols((void **) &(call_list[call_no]), 1);

			func_name = *dummy;
			if (*func_name != '[')
				fprintf(stderr, "  [%s]", func_name);
			else
#endif
				fprintf(stderr, "  [some silly code]");
		}
		else
			fprintf(stderr, "  [%s+0x%lX(%lu)]", func_name, offset, offset);
		fprintf(stderr, "\n");
		call_no++;
	}
#endif
}

#endif

#ifdef HAVE_SIGCONTEXT

static void
agl_print_diag_info(agl_proc_tables * ptabs, struct sigcontext *psc)
{
	agl_get_proc_tables(ptabs);
	fprintf(stderr, "Printing Debug Information :\n");
	agl_print_lib_list(ptabs);
	if (psc)
	{
		agl_print_signal_context(psc);
#if defined(_ASMi386_SIGCONTEXT_H)
		agl_print_my_backtrace(ptabs, (long *) (psc->ebp), (long *) (psc->esp), (long *) (psc->eip));
#endif
	}
}

#endif

static void
agl_sigsegv_handler(int signum
#ifdef HAVE_SIGCONTEXT
		    , struct sigcontext sc
#endif
	)
{
	static int level = 0;
	agl_proc_tables agl_ptabs;

	if (signum == SIGSEGV)
	{
		fprintf(stderr, "Segmentation Fault trapped.\n");
		if (level > 0)
			exit(1);	/* sigsegv in sigsegv */
		level++;
	}
	else
		fprintf(stderr, "Non-critical Signal %d trapped.\n", signum);

#ifdef HAVE_SIGCONTEXT
	agl_print_diag_info(&agl_ptabs, &sc);
#else
	agl_print_simple_backtrace(&agl_ptabs);
#endif
	if (signum == SIGSEGV)
	{
		fprintf(stderr,
			"Please collect all the listed information and submit a bug report to <http://savannah.nongnu.org/support/index.php?group=adgali>.\n");
		fprintf(stderr,
			"If core dump was generated by this fault, please examine it with gdb and attach results to your report.\n");
		fprintf(stderr, " You can use the following sequence to do so :\n");
		fprintf(stderr, "   gdb -core core your_program\n");
		fprintf(stderr, "   gdb>backtrace\n");
		fprintf(stderr, "   gdb>info frame\n");
		fprintf(stderr, "   gdb>info all-registers\n");
		fprintf(stderr, "   gdb>disassemble\n");
		exit(1);
	}
}

struct agl_sighandler
{
	int handle;
	void *handler;
	void *ohandler;
};

static struct agl_sighandler agl_handler[] = {
	{SIGSEGV, agl_sigsegv_handler, NULL},
	{SIGBUS, agl_sigsegv_handler, NULL},
	{SIGFPE, agl_sigsegv_handler, NULL},
	{0, NULL, NULL}
};

int
agl_initparachute(void)
{
	struct agl_sighandler *phand;

	for (phand = agl_handler; phand->handle; phand++)
	{
		phand->ohandler = signal(phand->handle, phand->handler);

		/* keep handler set by calling library */
		if (phand->ohandler == SIG_ERR)
		{
			agl_error("unable to set signal handler");
			return -1;
		}
		if (phand->ohandler != SIG_DFL)
		{
			signal(phand->handle, phand->ohandler);
			phand->ohandler = NULL;
		}
	}
	return 1;
}

void
agl_exitparachute(void)
{
	struct agl_sighandler *phand;

	for (phand = agl_handler; phand->handle; phand++)
	{
		signal(phand->handle, phand->ohandler);
	}
}
