
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! The objects cache system

/*! \file
 * \ingroup module
 * this file contains the cache management system. This system contains all the
 * objects used by adgali (only the objects loaded). The cache system is yet just
 * a repository. There is no memory management and cache ageing system. The cache
 * search system is very basic (linear search)
 */


#include <string.h>
#include "config.h"
#include "agl_private.h"

static int agl_objfree(agl_object * pobj);

int
agl_cacheinit(agl_container * phandle)
{
	phandle->cacheroot = NULL;

	return (1);
}

int
agl_cacheexit(agl_container * phandle)
{
	agl_object *ps, *pnext;
	int ret;


	ret = 1;
	ps = phandle->cacheroot;

	for (; ps != NULL; ps = pnext)
	{
		pnext = ps->pnext;
		agl_objdelete(ps);
	}

	phandle->cacheroot = NULL;

	return (1);
}

// TODO: add a name search to avoid duplicated names. today
// if names are duplicated, the last declared is the one found

int
agl_cacheadd(agl_object * pobj)
{
	agl_object *cacheroot;

	agl_rvassert(pobj != NULL, -1);

	cacheroot = pobj->pmod->handle->cacheroot;

	pobj->pprev = NULL;
	if (cacheroot == NULL)
	{
		pobj->pnext = NULL;
	}
	else
	{
		pobj->pnext = cacheroot;
		cacheroot->pprev = pobj;
	}
	cacheroot = pobj;

	return 1;
}

int
agl_cachesearch(char *name, agl_handle handle, agl_object ** pobj)
{
	agl_object *ps;
	agl_object *cacheroot;

	agl_rvassert(name != NULL, -1);
	agl_rvassert(pobj != NULL, -1);

	cacheroot = handle->cacheroot;

	for (ps = cacheroot; ps != NULL; ps = ps->pnext)
	{
		if (strcmp(ps->name, name) == 0)
		{
			*pobj = ps;
			return 1;
		}
	}

	return 0;
}

int
agl_cacheobjexists(agl_object * pobj)
{
	agl_object *ps;
	agl_object *cacheroot;

	agl_rvassert(pobj != NULL, -1);

	cacheroot = pobj->pmod->handle->cacheroot;

	for (ps = cacheroot; ps != NULL; ps = ps->pnext)
		if (ps == pobj)
			return 1;
	return 0;
}

int
agl_cacheobjevent(agl_object * pobj, int event, agl_object * pref)
{
	switch (event)
	{
	case AGL_OBJEVT_FLUSHED:
	case AGL_OBJEVT_UPDATE:
		break;
	case AGL_OBJEVT_UNLOAD:
		if (pobj->flags & AGL_OBJ_LOCKED)
			return 0;
		break;
	}

	agl_rvassert(pobj->pmod->mod_dependevt != NULL, -1);

	if (pobj->pmod->mod_dependevt(pobj, event, pobj->name, pref) < 0)
	{
		agl_error3("object %s : event failed: %d", pobj->name, event);
		return -1;
	}

	return 1;
}

int
agl_cacheobjget(char *name, agl_handle handle, agl_object ** pobj)
{
	agl_object *ps;
	int ret;

	agl_rvassert(name != NULL, -1);
	agl_rvassert(pobj != NULL, -1);

	if ((ret = agl_cachesearch(name, handle, &ps)) != 1)
	{
		agl_error2("variable %s not found in cache", name);
		return ret;
	}
	if ((ps->flags & AGL_OBJ_LOADED) == 0)
	{
		if (agl_cacheobjevent(ps, AGL_OBJEVT_UPDATE, ps) < 0)
		{
			agl_error("object load/update event failed");
			return -1;
		}
		agl_rvassert((ps->flags & AGL_OBJ_LOADED) == 0, 0);
	}
	*pobj = ps;
	return 1;
}

//! walk through cache and find object of type module \a "name"

/*! this function is useful to find objects of a selected module name
 * \param name the name of the module to find
 * \param pobj the address of a pointer to an object. if the pointer is NULL
 * then it would contain on return the first object of type \a name (if found)
 * from the beginning of the cache. if the pointer is not NULL, the pointer would
 * contain on return the next object after the current of type \a name found
 * in the cache. the initial object must be in the cache (beware when suppressing
 * object in the cache).
 * \return nothing
 */

int
agl_cachemodwalk(agl_container * phandle, char *name, agl_object ** pobj)
{
	agl_object *ps;
	agl_object *cacheroot;

	agl_rvassert(name != NULL, -1);
	agl_rvassert(pobj != NULL, -1);

	if (*pobj == NULL)
		cacheroot = phandle->cacheroot;
	else
		cacheroot = *pobj;

	for (ps = cacheroot; ps != NULL; ps = ps->pnext)
	{
		if (strcmp(ps->name, name) == 0)
		{
			*pobj = ps;
			return 1;
		}
	}

	return 0;
}

static agl_object *
agl_objnew(void)
{
	agl_object *pobj;

	pobj = (agl_object *) agl_malloc(sizeof(agl_object));
	if (pobj == NULL)
	{
		agl_memerr();
		return (NULL);
	}
	pobj->name = NULL;
	pobj->flags = 0;
	pobj->priv = NULL;
	pobj->pmod = NULL;
	pobj->usedby = NULL;
	pobj->defmax = 0;
	pobj->defcur = 0;
	pobj->defidx = 0;
	pobj->defelem = NULL;
	pobj->pnext = NULL;
	pobj->pprev = NULL;

	return pobj;
}

agl_object *
agl_objcreate(char *name, agl_module * pmod)
{
	agl_object *pobj;
	agl_modelem *pelem;
	int nb;

	agl_rvassert((name != NULL), NULL);
	agl_rvassert((pmod != NULL), NULL);

	pobj = agl_objnew();
	if (pobj == NULL)
	{
		return (NULL);
	}
	pobj->name = (char *) agl_malloc(strlen(name) + 1);
	strcpy(pobj->name, name);
	pobj->pmod = pmod;
	pelem = pmod->pmodelem;
	if (pelem != NULL)
	{
		nb = 0;
		while (pelem->name)
		{
			pelem++;
			nb++;
		}
		pobj->defmax = nb;
		if (nb != 0)
		{		// allocate first at all modelem element for object definition
			pobj->defelem = (agl_elemtype *) agl_malloc(sizeof(agl_elemtype) * nb);
			if (pobj->defelem == NULL)
			{
				agl_objfree(pobj);
				agl_memerr();
				return NULL;
			}
			memset(pobj->defelem, 0, sizeof(agl_elemtype) * nb);
			pobj->defidx = (int *) agl_malloc(sizeof(int) * nb);
			if (pobj->defidx == NULL)
			{
				agl_objfree(pobj);
				agl_memerr();
				return NULL;
			}
			memset(pobj->defidx, 0, sizeof(int) * nb);

		}
	}
	return pobj;
}

static int
agl_objfree(agl_object * pobj)
{
	agl_modelem *pelem;
	agl_elemtype *pdef;

	agl_rvassert(pobj != NULL, -1);
	if (pobj->name != NULL)
		agl_free(pobj->name);
	if (pobj->defidx != NULL)
		agl_free(pobj->defidx);

	pelem = pobj->pmod->pmodelem;
	pdef = pobj->defelem;
	if (pdef != NULL)
	{
		if (pelem != NULL)
		{
			while (pelem->name)
			{
				agl_defnsubrecur(pdef, 0);
				pelem++;
				pdef++;
			}
		}
		agl_free(pobj->defelem);
	}
	agl_free(pobj);
	return 1;
}

int
agl_objdelete(agl_object * pobj)
{
	agl_object *cacheroot;
	agl_objectlist *pobjl;

	agl_rvassert(pobj != NULL, -1);

	for (pobjl = pobj->usedby; pobjl != NULL; pobjl = pobjl->pnext)
		if (agl_cacheobjevent(pobjl->pobj, AGL_OBJEVT_FLUSHED, pobj) < 0)
		{
			agl_error("parent object flush event failed");
			return -1;
		}

	if (pobj->pmod->mod_delete != NULL)
		if (pobj->pmod->mod_delete(pobj) < 0)
		{
			agl_error("object deletion failed");
			return -1;
		}

	cacheroot = pobj->pmod->handle->cacheroot;

	if (pobj->pprev == NULL)
		cacheroot = pobj->pnext;
	else
		pobj->pprev->pnext = pobj->pnext;
	if (pobj->pnext != NULL)
		pobj->pnext->pprev = pobj->pprev;

	if (agl_objfree(pobj) < 0)
	{
		agl_error("object deletion failed");
		return -1;
	}
	return 1;
}

int
agl_objaddelem(agl_object * pobj, int idx, agl_elemtype * pv)
{
	agl_elemtype *pdef;
	int i;

	agl_rvassert(pobj != NULL, -1);
	agl_rvassert(pv != NULL, -1);

	agl_rvassert(pobj->defelem != NULL, -1);

	for (i = 0; i < pobj->defcur; i++)
		if (pobj->defidx[i] == idx)
		{
			pdef = pobj->defelem + pobj->defcur;
			break;
		}
	if (i == pobj->defcur)
	{
		agl_rvassert(pobj->defcur < pobj->defmax, -1);
		pdef = pobj->defelem + pobj->defcur;
		pobj->defidx[pobj->defcur++] = idx;
	}

	agl_defnaddrecur(&pdef, pv, 0);

	return 1;
}
