
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/*
 * Hash library from
 * Copyright (c) 2000 Andrew Ferguson <andrew@owsla.cjb.net>
 * Copyright (c) 1998 Sasha Vasko <sasha@aftercode.net>
 * and translated for adgali
 */

#include "config.h"
#include "agl_private.h"
#include <ctype.h>
#include <string.h>


static unsigned short
default_hash_func(unsigned long value, unsigned short hash_size)
{
	return value % hash_size;
}

static long
default_compare_func(unsigned long value1, unsigned long value2)
{
	return ((long) value1 - (long) value2);
}

static void
agl_hashinit(agl_hashtab * hash, int freeres)
{
	if (hash)
	{
		if (freeres)
			if (hash->buckets)
				agl_free(hash->buckets);
		memset(hash, 0x00, sizeof(agl_hashtab));
	}
}

agl_hashtab *
agl_hashcreate(unsigned short size, unsigned short (*hash_func) (unsigned long, unsigned short),
	       long (*compare_func) (unsigned long, unsigned long), void (*item_destroy_func) (unsigned long, void *))
{
	agl_hashtab *hash;

	if (size <= 0)
		size = DEFAULT_HASH_SIZE;

	hash = agl_malloc(sizeof(agl_hashtab));
	agl_hashinit(hash, 0);

	hash->buckets = agl_malloc(sizeof(agl_hashitem *) * size);
	memset(hash->buckets, 0x00, sizeof(agl_hashitem *) * size);

	hash->size = size;

	if (hash_func)
		hash->hash_func = hash_func;
	else
		hash->hash_func = default_hash_func;

	if (compare_func)
		hash->compare_func = compare_func;
	else
		hash->compare_func = default_compare_func;

	hash->item_destroy_func = item_destroy_func;

	return hash;
}

static void
agl_hashdestroybucket(agl_hashitem ** bucket, void (*item_destroy_func) (unsigned long, void *))
{
	agl_hashitem *item, *next;

	for (item = *bucket; item != NULL; item = next)
	{
		next = item->next;
		if (item_destroy_func)
			item_destroy_func(item->value, item->data);
		agl_free(item);
	}
	*bucket = NULL;
}

void
agl_hashdestroy(agl_hashtab ** hash)
{
	int i;

	if (*hash)
	{
		for (i = (*hash)->size - 1; i >= 0; i--)
			if ((*hash)->buckets[i])
				agl_hashdestroybucket(&((*hash)->buckets[i]), (*hash)->item_destroy_func);

		agl_hashinit(*hash, 1);
		agl_free(*hash);
		*hash = NULL;
	}
}

static int
add_item_to_bucket(agl_hashitem ** bucket, agl_hashitem * item, long (*compare_func) (unsigned long, unsigned long))
{
	agl_hashitem **tmp;

	/* first check if we already have this item */
	for (tmp = bucket; *tmp != NULL; tmp = &((*tmp)->next))
	{
		long res = compare_func((*tmp)->value, item->value);

		if (res == 0)
			return ((*tmp)->data == item->data) ? 0 : -1;
		else if (res > 0)
			break;
	}
	/* now actually add this item */
	item->next = (*tmp);
	*tmp = item;
	return 1;
}

#define DEALLOC_CACHE_SIZE      1024
static agl_hashitem *deallocated_mem[DEALLOC_CACHE_SIZE + 10];
static unsigned int deallocated_used = 0;

int
agl_hashadditem(agl_hashtab * hash, unsigned long value, void *data)
{
	unsigned short key;
	agl_hashitem *item;
	int res;

	if (hash == NULL)
		return -3;

	key = hash->hash_func(value, hash->size);
	if (key >= hash->size)
		return -3;

	if (deallocated_used > 0)
		item = deallocated_mem[--deallocated_used];
	else
		item = agl_malloc(sizeof(agl_hashitem));

	item->next = NULL;
	item->value = value;
	item->data = data;

	res = add_item_to_bucket(&(hash->buckets[key]), item, hash->compare_func);
	if (res == 1)
	{
		hash->most_recent = item;
		hash->items_num++;
		if (hash->buckets[key]->next == NULL)
			hash->buckets_used++;
	}
	else
		agl_free(item);
	return res;
}

static agl_hashitem **
find_item_in_bucket(agl_hashitem ** bucket, unsigned long value, long (*compare_func) (unsigned long, unsigned long))
{
	agl_hashitem **tmp;
	long res;

	/* first check if we already have this item */
	for (tmp = bucket; *tmp != NULL; tmp = &((*tmp)->next))
	{
		res = compare_func((*tmp)->value, value);
		if (res == 0)
			return tmp;
		else if (res > 0)
			break;
	}
	return NULL;
}

int
agl_hashgetitem(agl_hashtab * hash, unsigned long value, void **trg)
{
	unsigned short key;
	agl_hashitem **pitem = NULL;

	if (hash)
	{
		key = hash->hash_func(value, hash->size);
		if (key < hash->size)
			pitem = find_item_in_bucket(&(hash->buckets[key]), value, hash->compare_func);
	}
	if (pitem)
		if (*pitem)
		{
			if (trg)
				*trg = (*pitem)->data;
			return 1;
		}
	return -2;
}

void
flush_ashash_memory_pool()
{
	while (deallocated_used > 0)
		agl_free(deallocated_mem[--deallocated_used]);
}

int
agl_hashremoveitem(agl_hashtab * hash, unsigned long value, void **trg, int destroy)
{
	unsigned short key = 0;
	agl_hashitem **pitem = NULL;

	if (hash)
	{
		key = hash->hash_func(value, hash->size);
		if (key < hash->size)
			pitem = find_item_in_bucket(&(hash->buckets[key]), value, hash->compare_func);
	}
	if (pitem)
		if (*pitem)
		{
			agl_hashitem *next;

			if (hash->most_recent == *pitem)
				hash->most_recent = NULL;

			if (trg)
				*trg = (*pitem)->data;

			next = (*pitem)->next;
			if (hash->item_destroy_func && destroy)
				hash->item_destroy_func((*pitem)->value, (trg) ? NULL : (*pitem)->data);

			if (deallocated_used < DEALLOC_CACHE_SIZE)
			{
				deallocated_mem[deallocated_used++] = *pitem;
			}
			else
				agl_free(*pitem);

			*pitem = next;
			if (hash->buckets[key] == NULL)
				hash->buckets_used--;
			hash->items_num--;

			return 1;
		}
	return -2;
}

unsigned long
agl_hashsortitems(agl_hashtab * hash, unsigned long *values, void **data, unsigned long max_items)
{
	if (hash)
	{
		agl_hashitem **buckets;
		unsigned short i;
		unsigned short k = 0, top = hash->buckets_used - 1, bottom = 0;
		unsigned long count_in = 0;

		if (hash->buckets_used == 0 || hash->items_num == 0)
			return 0;

		if (max_items == 0)
			max_items = hash->items_num;

		buckets = agl_malloc(hash->buckets_used * sizeof(agl_hashitem *));
		for (i = 0; i < hash->size; i++)
			if (hash->buckets[i])
				buckets[k++] = hash->buckets[i];
		while (max_items-- > 0)
		{
			k = bottom;
			for (i = bottom + 1; i <= top; i++)
			{
				if (buckets[i])
					if (hash->compare_func(buckets[k]->value, buckets[i]->value) > 0)
						k = i;
			}
			if (values)
				*(values++) = buckets[k]->value;
			if (data)
				*(data++) = buckets[k]->data;
			count_in++;
			buckets[k] = buckets[k]->next;
			/* a little optimization : */
			while (buckets[bottom] == NULL && bottom < top)
				bottom++;
			while (buckets[top] == NULL && top > bottom)
				top--;
			if (buckets[top] == NULL)
				break;
		}
		agl_free(buckets);
		return count_in;
	}
	return 0;
}

unsigned long
agl_hashlistitems(agl_hashtab * hash, unsigned long *values, void **data, unsigned long max_items)
{
	unsigned long count_in = 0;

	if (hash)
		if (hash->buckets_used > 0 && hash->items_num > 0)
		{
			agl_hashitem *item;
			unsigned short i;

			if (max_items == 0)
				max_items = hash->items_num;

			for (i = 0; i < hash->size; i++)
				for (item = hash->buckets[i]; item != NULL; item = item->next)
				{
					if (values)
						*(values++) = item->value;
					if (data)
						*(data++) = item->data;
					if (++count_in >= max_items)
						return count_in;
				}
		}
	return count_in;
}

/***** Iterator functionality *****/

int
agl_hashstartiter(agl_hashtab * hash, agl_hashiter * iterator)
{
	if (iterator && hash)
	{
		int i;

		for (i = 0; i < hash->size; i++)
			if (hash->buckets[i] != NULL)
				break;
		if (i < hash->size)
		{
			iterator->hash = hash;
			iterator->curr_bucket = i;
			iterator->curr_item = &(hash->buckets[i]);
			return 1;
		}
	}
	return 0;
}

int
agl_hashnextiter(agl_hashiter * iterator)
{
	if (iterator)
		if (iterator->hash && iterator->curr_item)
		{
			agl_hashitem **curr = iterator->curr_item;

			if (*curr)
				curr = &((*curr)->next);

			iterator->curr_item = curr;

			if (*curr == NULL)
			{
				int i;

				for (i = iterator->curr_bucket + 1; i < iterator->hash->size; i++)
					if (iterator->hash->buckets[i] != NULL)
						break;
				if (i < iterator->hash->size)
				{
					iterator->curr_item = &(iterator->hash->buckets[i]);
					iterator->curr_bucket = i;
				}
			}

			return (*(iterator->curr_item) != NULL);
		}
	return 0;
}

void
agl_hashremovecurrval(agl_hashiter * iterator, int destroy)
{
	if (iterator)
		if (iterator->hash && iterator->curr_item)
		{
			agl_hashitem *removed = *(iterator->curr_item);

			if (removed)
			{
				agl_hashtab *hash = iterator->hash;
				unsigned short key = iterator->curr_bucket;

				*(iterator->curr_item) = removed->next;
				removed->next = NULL;

				if (hash->item_destroy_func && destroy)
					hash->item_destroy_func(removed->value, removed->data);
				agl_free(removed);
				if (hash->buckets[key] == NULL)
					hash->buckets_used--;
				hash->items_num--;
			}
		}
}

unsigned long
agl_hashcurrval(agl_hashiter * iterator)
{
	if (iterator)
	{
		if (iterator->curr_item && *(iterator->curr_item))
			return (*(iterator->curr_item))->value;
	}
	return (unsigned long) ((char *) NULL);
}

void *
agl_hashcurrdata(agl_hashiter * iterator)
{
	if (iterator)
	{
		if (iterator->curr_item && *(iterator->curr_item))
			return (*(iterator->curr_item))->data;
	}
	return NULL;
}

/************************************************************************/

/************************************************************************/

/* 	Some usefull implementations 					*/

/************************************************************************/
unsigned short
agl_hashptrhash(unsigned long value, unsigned short hash_size)
{
	union
	{
		void *ptr;
		unsigned short key[2];
	} mix;
	unsigned short key;

	mix.ptr = (void *) value;
	key = mix.key[0] ^ mix.key[1];
	if (hash_size == 256)
		return (key >> 4) & 0x0FF;
	return (key >> 4) % hash_size;
}

/* case sensitive strings hash */
unsigned short
agl_hashstrhash(unsigned long value, unsigned short hash_size)
{
	unsigned short hash_key = 0;
	unsigned int i = 0;
	char *string = (char *) value;
	char c;

	do
	{
		c = string[i];
		if (c == '\0')
			break;
		hash_key += (((unsigned short) c) << i);
		++i;
	}
	while (i < ((sizeof(unsigned short) - sizeof(char)) << 3));
	return hash_key % hash_size;
}

long
agl_hashstrcmp(unsigned long value1, unsigned long value2)
{
	char *str1 = (char *) value1;
	char *str2 = (char *) value2;
	int i = 0;

	if (str1 == str2)
		return 0;
	if (str1 == NULL)
		return -1;
	if (str2 == NULL)
		return 1;
	do
	{
		if (str1[i] != str2[i])
			return (long) (str1[i]) - (long) (str2[i]);

	}
	while (str1[i++]);
	return 0;
}

void
agl_hashstrdestroy(unsigned long value, void *data)
{
	if ((char *) value != NULL)
		agl_free((char *) value);
	if (data != (void *) value && data != NULL)
		agl_free(data);
}

/* variation for case-unsensitive strings */
unsigned short
agl_hashcasestrhash(unsigned long value, unsigned short hash_size)
{
	unsigned short hash_key = 0;
	unsigned int i = 0;
	char *string = (char *) value;
	int c;

	do
	{
		c = string[i];
		if (c == '\0')
			break;
		if (isupper(c))
			c = tolower(c);
		hash_key += (((unsigned short) c) << i);
		++i;
	}
	while (i < ((sizeof(unsigned short) - sizeof(char)) << 3));

	return hash_key % hash_size;
}

long
agl_hashcasestrcmp(unsigned long value1, unsigned long value2)
{
	char *str1 = (char *) value1;
	char *str2 = (char *) value2;
	int i = 0;

	if (str1 == str2)
		return 0;
	if (str1 == NULL)
		return -1;
	if (str2 == NULL)
		return 1;
	do
	{
		int u1, u2;

		u1 = str1[i];
		u2 = str2[i];
		if (islower(u1))
			u1 = toupper(u1);
		if (islower(u2))
			u2 = toupper(u2);
		if (u1 != u2)
			return (long) u1 - (long) u2;
	}
	while (str1[i++]);
	return 0;
}

/* colors hash */

unsigned short
color_hash_value(unsigned long value, unsigned short hash_size)
{
	unsigned long h;
	unsigned long g;
	unsigned int i = 1;
	unsigned long long_val = value;

	h = (long_val & 0x0ff);
	long_val = long_val >> 8;

	while (i++ < 4)
	{			/* for the first 4 bytes we can use simplier loop as there can be no overflow */
		h = (h << 4) + (long_val & 0x0ff);
		long_val = long_val >> 8;
	}
	while (i++ < sizeof(unsigned long))
	{			/* this loop will only be used on computers with long > 32bit */

		h = (h << 4) + (long_val & 0x0ff);
		long_val = long_val >> 8;
		if ((g = h & 0xf0000000))
		{
			h ^= g >> 24;
			h &= 0x0fffffff;
		}
	}
	return (unsigned short) (h % (unsigned long) hash_size);
}
