
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! This is the central module management system

/*! \file
 * \ingroup module
 * This file contains the module management system. This system is the central
 * nervous of adgali. It registers modules, manage objects and generates objects
 * tree dynamically. It's purpose is not to manage instanciated objects relationships
 */

#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "agl_private.h"

extern agl_module *agl_moduletab[];

int agl_modderegister(agl_container * handle, char *name);

agl_module *
agl_modsearch(agl_container * handle, char *name)
{
	agl_module *pmod;

	agl_rvassert(handle != NULL, NULL);
	agl_rvassert(name != NULL, NULL);

	for (pmod = handle->moduleroot; pmod != NULL; pmod = pmod->pnext)
	{
		if (strcmp(pmod->name, name) == 0)
		{
			return pmod;
		}
	}
	return NULL;
}

//! register a module

/*! the module registered is then usable by the library
 * \param modfunc a pointer to the module to register
 * \return
 * - -1 if a fatal error occurs
 * - 0 if it fails but there is no fatal error
 * - 1 if success
 */

int
agl_modregister(agl_container * handle, agl_module * pinitmod)
{
	agl_module *pmod, *pcurmod, *pprev;
	agl_caller *pcal;
	int i;

	agl_rvassert(handle != NULL, -1);
	agl_rvassert(pinitmod != NULL, -1);

	if (pinitmod->name == NULL)
		return 1;	// module without object handler
	pcurmod = (agl_module *) agl_malloc(sizeof(agl_module));
	if (pcurmod == NULL)
	{
		agl_memerr();
		return -1;
	}
	memcpy(pcurmod, pinitmod, sizeof(agl_module));

	pcurmod->class_struct = NULL;
	pcurmod->handle = handle;
	pcurmod->pnext = handle->moduleroot;
	handle->moduleroot = pcurmod;

	if (pinitmod->mod_init != NULL)
	{

		if ((pcurmod->mod_init) (pcurmod) < 0)
		{
			agl_error(" Unable to initialise module");
			return -1;
		}
	}

	for (pcal = pcurmod->caller, i = 0; (pcal->name) && (i < AGL_MODER_SZ); pcal++, i++)
	{
		pmod = agl_modsearch(handle, pcal->name);
		if (pmod == NULL)
		{
			continue;	// caller not known waiting next register
		}
		if (pmod->mod_caller == NULL)
		{
			agl_error2("caller module %s has no caller function", pcal->name);
			agl_modderegister(handle, pcurmod->name);
			return -1;
		}
		if (pmod->mod_caller(pmod, 1, pcurmod, pcal->priv) != 1)
		{
			agl_error3("caller module %s failed initialising called module %s", pcal->name, pcurmod->name);
			agl_modderegister(handle, pcurmod->name);
			return -1;
		}
	}

	for (pmod = handle->moduleroot; pmod != NULL; pprev = pmod, pmod = pmod->pnext)
	{
		for (pcal = pmod->caller, i = 0; (pcal->name) && (i < AGL_MODER_SZ); pcal++, i++)
		{
			if (strcmp(pcal->name, pcurmod->name) == 0)
			{	// found a new called not registered
				if (pcurmod->mod_caller == NULL)
				{
					agl_error2("caller module %s has no caller function", pcurmod->name);
					agl_modderegister(handle, pcurmod->name);
					return -1;
				}
				if (pcurmod->mod_caller(pcurmod, 1, pmod, pcal->priv) != 1)
				{
					agl_error3("caller module %s failed initialising called module %s", pcurmod->name, pmod->name);
					agl_modderegister(handle, pcurmod->name);
					return -1;
				}
			}
		}
	}

	return 1;
}

int
agl_modderegister(agl_container * handle, char *name)
{
	agl_module *pmod, *pprev, *pref;
	agl_object *pobj, *pnobj;
	agl_caller *pcal;
	int i;
	int ret;

	agl_rvassert(handle != NULL, -1);
	agl_rvassert(name != NULL, -1);

	pprev = NULL;
	for (pmod = handle->moduleroot; pmod != NULL; pprev = pmod, pmod = pmod->pnext)
	{
		if (strcmp(pmod->name, name) == 0)
			break;
	}
	if (pmod == NULL)
		return 0;
	// free all objects before unloading module

	pobj = NULL;
	while (1)
	{
		if ((ret = agl_cachemodwalk(handle, name, &pobj)) == -1)
		{
			agl_error("error during object release; aborting");
			return -1;
		}
		if ((ret == 0) && (pobj == NULL))
			break;	// no more objects found
		pnobj = pobj->pnext;
		if (agl_objdelete(pobj) != 1)
		{
			agl_error("error during object release; aborting");
			return -1;
		}
		pobj = pnobj;
	}

	// unregistering caller module

	for (pcal = pmod->caller, i = 0; (pcal->name) && (i < AGL_MODER_SZ); pcal++, i++)
	{
		pref = agl_modsearch(handle, pcal->name);
		if (pref == NULL)
		{
			agl_error2("unknown caller module %s", pcal->name);
			return -1;
		}
		if (pref->mod_caller == NULL)
		{
			agl_error2("caller module %s has no caller function", pcal->name);
			return -1;
		}
		if (pref->mod_caller(pref, 0, pmod, pcal->priv) != 1)
		{
			agl_error3("caller module %s failed de-registering called module %s", pcal->name, pmod->name);
			return -1;
		}
	}

	// unregistering called module

	for (pref = handle->moduleroot; pref != NULL; pref = pref->pnext)
	{
		for (pcal = pref->caller, i = 0; (pcal->name) && (i < AGL_MODER_SZ); pcal++, i++)
		{
			if (strcmp(pcal->name, name) == 0)
			{	// found a new called not de-registered
				if (pmod->mod_caller == NULL)
				{
					agl_error2("caller module %s has no caller function", name);
					return -1;
				}
				if (pmod->mod_caller(pmod, 1, pref, pcal->priv) != 1)
				{
					agl_error3("caller module %s failed de-registering called module %s", name, pref->name);
					return -1;
				}
			}
		}
	}

	// at last free the module

	if (pprev == NULL)
		handle->moduleroot = pmod->pnext;
	else
		pprev->pnext = pmod->pnext;

	if (pmod->mod_exit != 0)
	{
		if (pmod->mod_exit(pmod) < 0)
		{
			agl_error("error during object release; aborting");
			return -1;
		}
	}

	agl_free(pmod);

	return 1;
}


int
agl_modinit(agl_container * handle)
{
	int i;

	agl_rvassert(handle != NULL, -1);

	handle->moduleroot = NULL;
	for (i = 0; agl_moduletab[i]; i++)
	{
		if (agl_modregister(handle, agl_moduletab[i]) < 0)
		{
			agl_error("unable to initialise module system");
			return -1;
		}
	}
	return 1;
}

int
agl_modexit(agl_container * handle)
{
	agl_rvassert(handle != NULL, -1);

	while (handle->moduleroot != NULL)
	{
		if (agl_modderegister(handle, handle->moduleroot->name) < 0)
		{
			agl_error("unable to free module system");
		}
	}

	return 1;
}
