
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! This file implements the lexical interpreter and the scanner

/*! \file
 * \ingroup module
 * here is described the lexical analyser (integer, characters, float
 * octal, binary and hexadecimal numbers, strings, identifier, keywords
 * and token).
 * the rest of this file contains the tokenised string analyser (created after
 * parsing). each token is interpreted and an action is executed. The interpretation
 * of the string instanciates all the objects.
 * Temporary variables and the tokenised string are then freed
 */


#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "config.h"
#include "agl_private.h"

#define CHAR_TOKEN	0
#define STRING_TOKEN	1
#define ZDEC_TOKEN	2
#define NUMBER_TOKEN	3
#define FLOAT_TOKEN	4
#define INT_TOKEN	5
#define HEXA_TOKEN	6
#define BIN_TOKEN	7
#define OCT_TOKEN	8
#define SYMB_TOKEN	9
#define SKIP_TOKEN	10

unsigned char *
agl_getnum(unsigned char *ch, long *value)
{
	int type, status;
	int msign;
	unsigned char c;
	unsigned long uval;

	agl_rvassert(ch != NULL, NULL);
	agl_rvassert(value != NULL, NULL);

	type = -1;
	uval = 0;
	status = 0;
	msign = 1;
	while (1)
	{
		c = *ch++;
		if (c == '\0')
			return NULL;
		else
			switch (type)
			{
			case -1:
				switch (c)
				{
				case '0':
					type = ZDEC_TOKEN;
					break;
				case '-':
					msign = -1;
				case '+':
					type = NUMBER_TOKEN;
					break;
				default:
					if (isdigit(c))
					{
						type = INT_TOKEN;
						uval = c - '0';
					}
					else
						return NULL;
				}
				break;
			case NUMBER_TOKEN:
				if (c == '0')
				{
					type = ZDEC_TOKEN;
				}
				else
				{
					if (isdigit(c))
					{
						type = INT_TOKEN;
						uval = c - '0';
					}
					else
						return NULL;
				}
				break;
			case ZDEC_TOKEN:
				switch (c)
				{
				case 'x':
				case 'X':
					type = HEXA_TOKEN;
					break;
				case 'b':
				case 'B':
					type = BIN_TOKEN;
					break;
				default:
					if (c < '0')
						status = 3;
					else if (c <= '7')
					{
						type = OCT_TOKEN;
						uval = c - '0';
						status = 1;
					}
					else if (c <= '9')
					{
						type = INT_TOKEN;
						uval = c - '0';
					}
					else
						status = 3;
					break;
				}
				break;
			case HEXA_TOKEN:
				if (isdigit(c))
				{
					uval = uval * 16 + c - '0';
					status = 1;
				}
				else
				{
					c = toupper(c);
					if ((c >= 'A') && (c <= 'F'))
					{
						uval = uval * 16 + c - 'A' + 10;
						status = 1;
					}
					else if (status == 0)
						return NULL;
					else
						status = 3;
				}
				break;
			case OCT_TOKEN:
				if ((c >= '0') && (c <= '7'))
					uval = uval * 8 + c - '0';
				else
					status = 3;
				break;
			case INT_TOKEN:
				if ((c >= '0') && (c <= '9'))
				{
					uval = uval * 10 + c - '0';
				}
				else
					status = 3;
				break;
			case BIN_TOKEN:
				if ((c >= '0') && (c <= '1'))
				{
					uval = uval * 2 + c - '0';
					status = 1;
				}
				else if (status == 0)
					return NULL;
				else
					status = 3;
				break;
			default:
				agl_error2("illegal lexical type (%d)", type);
				return NULL;
				break;
			}
		if (status == 3)
		{
			*value = (long) uval *msign;

			return ch;
		}
	}
	agl_error("unexpected break");
	return NULL;
}
