
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! create a font descriptor

/*! \file
 * \ingroup video
 * this file contains all the procedure to create a font descriptor.
 * This descriptor could be a bitmap font with or without alpha blending
 * or a true type font. The fonts create are fixed or proportionnal fonts
 */

#include <ctype.h>
#include <string.h>
#include "config.h"
#include "agl_private.h"

extern agl_fontio agl_bfnt_fontio;

//! structure used to hold fonts parameters
typedef struct
{
	int created;		//!< if 1 this structure is filled
	agl_pix pixmap;		//!< the channel structure that contains the font
	agl_size outline_size;	//!< defines the rectangle that is around the character
	int baseline;		//!< vertical offset to the baseline
}
agl_bfntglyph;

/*! The structure used to hold internal ttf font information
 * element of size int and nexty must be the third
 */

typedef struct
{
	agl_bfntglyph glyphs[256];	//!< the glyphs table
	int nextx;		//!< origin increment for an horizontal writing
	int nexty;		//!< origin increment for a vertical writing
	agl_image *img;		//!< image containing the font
}
bmp_font;

static int
get_bfntchar(agl_ios * ops, unsigned char *car)
{
	unsigned char c;

	while (1)
	{
		if (agl_ios_read(ops, &c, 1, 1, NULL) != 1)
			return 1;
		if (c == '#')
			while (c != '\n')
				if (agl_ios_read(ops, &c, 1, 1, NULL) != 1)
					return 1;
		if (!isspace(c))
		{
			*car = c;
			return 0;
		}
	}
	return 1;
}

#define get_value(c) (((c>='0') && (c <='9')) ? c - '0' : ((isupper(c) || islower(c)) ? toupper(c) - 'A' + 10 : -1))

static int
read_bfontnumber(agl_ios * ops, unsigned long *val, int base)
{
	unsigned long digits;
	int v;
	unsigned char c;

	digits = 0;
	while (1)
	{
		if (agl_ios_read(ops, &c, 1, 1, NULL) != 1)
			return 1;
		if (c == '#')
			while (c != '\n')
				if (agl_ios_read(ops, &c, 1, 1, NULL) != 1)
					return 1;
		if (((v = get_value(c)) >= 0) && (v < base))
			break;
	}
	while (1)
	{
		v = get_value(c);
		if ((v >= 0) && (v < base))
			digits = digits * base + v;
		else
			break;
		if (agl_ios_read(ops, &c, 1, 1, NULL) != 1)
			break;
	}
	*val = digits;
	return 0;

}

static int
agl_bfntis(agl_module * pmod, agl_ios * ops)
{
	char buf[20];
	int i;
	unsigned char c;

	agl_rvassert((ops != NULL), -1);
	agl_rvassert((pmod != NULL), -1);

	if (get_bfntchar(ops, &c) < 0)
		return 0;
	i = 0;
	while ((!isspace(c)) && (i < 5))
	{
		buf[i++] = c;
		if (agl_ios_read(ops, &c, 1, 1, NULL) != 1)
			return 0;
	}
	buf[i] = 0;
	if (!isspace(c))
		return 0;
	if (strncmp(buf, "BFONT", 5) != 0)
		return 0;
	return 1;
}

static agl_font *
agl_bfntopen(agl_module * pmod, agl_ios * ops, int size)
{
	agl_font *font;
	agl_ios *pio;
	int i;
	agl_image *img;
	char buf[256];
	unsigned char c;
	unsigned long upper, lower, base, next, space;
	unsigned long first, second, ind;

	agl_rvassert((ops != NULL), NULL);

	size = 0;		// make compiler happy
	if (get_bfntchar(ops, &c) < 0)
	{
		agl_ioserr();
		return NULL;
	}
	i = 0;
	while ((!isspace(c)) && (i < 5))
	{
		buf[i++] = c;
		if (agl_ios_read(ops, &c, 1, 1, NULL) != 1)
		{
			agl_ioserr();
			return NULL;
		}
	}
	buf[i] = 0;
	if (!isspace(c))
	{
		agl_error("not a bitmap font file");
		return NULL;
	}
	if (strncmp(buf, "BFONT", 5) != 0)
	{
		agl_error("not a bitmap font file");
		return NULL;
	}
	if (get_bfntchar(ops, &c) < 0)
	{
		agl_ioserr();
		return NULL;
	}
	if (c != '"')
	{
		agl_error("invalid bitmap font file");
		return NULL;
	}
	i = 0;
	while (i < 255)
	{
		if (agl_ios_read(ops, &c, 1, 1, NULL) != 1)
		{
			agl_ioserr();
			return NULL;
		}
		if (c == '"')
			break;
		buf[i++] = c;
	}
	buf[i] = 0;
	if (c != '"')
	{
		agl_error("invalid bitmap font file");
		return NULL;
	}
	if (read_bfontnumber(ops, &upper, 10))
	{
		agl_error("invalid bitmap font file");
		return NULL;
	}
	if (read_bfontnumber(ops, &lower, 10))
	{
		agl_error("invalid bitmap font file");
		return NULL;
	}
	if (read_bfontnumber(ops, &base, 10))
	{
		agl_error("invalid bitmap font file");
		return NULL;
	}
	if (read_bfontnumber(ops, &next, 10))
	{
		agl_error("invalid bitmap font file");
		return NULL;
	}
	if (read_bfontnumber(ops, &space, 10))
	{
		agl_error("invalid bitmap font file");
		return NULL;
	}
	if ((base < upper) || (lower < base) || (next < lower))
	{
		agl_error("invalid bitmap font file");
		return NULL;
	}
	pio = agl_ios_open(pmod->handle, buf);
	if (pio == NULL)
	{
		agl_error("unable to open font file");
		return NULL;
	}
	img = agl_imgioload(pio);
	agl_ios_close(pio);

	if (img == NULL)
	{
		agl_error("invalid bitmap font file");
		return NULL;
	}
	if (img->pix->size.height < next)
	{
		agl_error("invalid bitmap font file");
		agl_imgfree(img);
		return NULL;
	}

	font = agl_fontnew();
	if (font == NULL)
	{
		agl_imgfree(img);
		agl_memerr();
		return (NULL);
	}

	font->height = next - upper;
	font->ascent = base - upper;
	font->descent = base - lower;
	font->linegap = next - lower;
	while (1)
	{
		if (read_bfontnumber(ops, &ind, 16))
		{
			break;
		}
		if (ind > 255)
		{
			agl_error("invalid bitmap font file");
			agl_imgfree(img);
			agl_free(font);
			return NULL;
		}
		if (read_bfontnumber(ops, &first, 10))
		{
			agl_error("invalid bitmap font file");
			agl_imgfree(img);
			agl_free(font);
			return NULL;
		}
		if (read_bfontnumber(ops, &second, 10))
		{
			agl_error("invalid bitmap font file");
			agl_imgfree(img);
			agl_free(font);
			return NULL;
		}
		if ((img->pix->size.width < second) || (second < first))
		{
			agl_error("invalid bitmap font file");
			agl_imgfree(img);
			agl_free(font);
			return NULL;
		}
		font->glyphs[ind].bbox.x = -first;
		font->glyphs[ind].bbox.y = -upper;
		font->glyphs[ind].bbox.width = second - first;
		font->glyphs[ind].bbox.height = lower - upper;
		font->glyphs[ind].advance = second - first + space;
		font->glyphs[ind].yoffset = base + upper;
		font->glyphs[ind].created = 1;
	}
	font->pfio = (agl_fontio *) pmod->priv_desc;
	font->priv = img;

	return ((agl_font *) font);
}

static int
agl_bfntloadglyph(agl_font * font, unsigned char ch)
{
	agl_rvassert(font != NULL, -1);

	agl_error4("Character \"%c\", %d, %x does not exist in font file", ch, ch, ch);
	return -1;
}

static int
agl_bfntblit(agl_pix * pix, agl_font * font, unsigned char c, agl_pos * pos)
{
	agl_image *img;
	agl_rect prect;

	agl_rvassert(pix != NULL, -1);
	agl_rvassert(font != NULL, -1);
	agl_rvassert(pos != NULL, -1);
	agl_rvassert(font->glyphs[c].created, -1);

	img = (agl_image *) font->priv;
	prect.x = -font->glyphs[c].bbox.x;
	prect.y = -font->glyphs[c].bbox.y;
	prect.width = font->glyphs[c].bbox.width;
	prect.height = font->glyphs[c].bbox.height;

	if (agl_pixalpha(img->pix, pix, &prect, NULL, pos) < 0)
	{
		agl_error("unable to draw character");
		return -1;
	}
	return 0;
}

static int
agl_bfntclose(agl_font * fnt)
{
	agl_image *img;

	agl_rvassert((fnt != NULL), -1);

	img = (agl_image *) fnt->priv;
	agl_imgfree(img);
	fnt->priv = NULL;
	return 0;
}

static agl_fontio agl_bfnt_fontio = {
	agl_bfntis,
	agl_bfntopen,
	agl_bfntloadglyph,
	agl_bfntblit,
	agl_bfntclose
};

agl_module agl_modtab_bfntio = {
	"BFONT",
	"loader for bitmap fonts",
	AGL_MOD_FONT,
	{{NULL, NULL}},
	(void *) &agl_bfnt_fontio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};
