
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! create a font descriptor for a true type font

/*! \file
 * \ingroup video
 * this file contains all the procedure to create a font descriptor.
 * This descriptor points to a true type font with or without alpha blending
 * This package uses freetype to create fonts.
 * The fonts create are fixed or proportionnal fonts
 */

#include <string.h>
#include "config.h"
#include "agl_private.h"

#if HAVE_TTF
#ifdef FREETYPE_DIR
#include <freetype/freetype.h>
#include <freetype/ftoutln.h>
#include <freetype/ttnameid.h>
#else
#include <freetype.h>
#include <ftoutln.h>
#include <ttnameid.h>
#endif

#undef __FTERRORS_H__
#define FT_ERRORDEF( e, v, s )  { e, s },
#define FT_ERROR_START_LIST      {
#define FT_ERROR_END_LIST        { 0, 0 } };

static const struct agl_ttferrstr
{
	int err_code;
	const char *err_msg;
}
agl_ttferrors[] =
#ifdef FREETYPE_DIR
#include <freetype/fterrors.h>
#else
#include <fterrors.h>
#endif
#define FT_FLOOR(X)	((X & -64) / 64)
#define FT_CEIL(X)	(((X + 63) & -64) / 64)
extern agl_fontio agl_ttf_fontio;

/*! The structure used to hold internal ttf font information
 * element of size int and nexty must be the third
 */
typedef struct
{
	FT_Face face;		//!< describes the font for freetype
	struct FT_StreamRec_ stream;	//!< the stream structure
	agl_channel chan[256];	//!< the channel table for each glyph
}
ttf_font;

static void
agl_ttferror(const char *msg, FT_Error error)
{
	const struct agl_ttferrstr *perr;

	for (perr = agl_ttferrors; perr->err_msg; perr++)
	{
		if (error == perr->err_code)
			break;
	}
	if (perr->err_msg)
		agl_error3("%s: %s", msg, perr->err_msg);
	else
		agl_error2("%s: Unknown FreeType error", msg);
}

static unsigned long
agl_ttfsread(FT_Stream stream, unsigned long offset, unsigned char *buffer, unsigned long count)
{
	agl_ios *ops;
	int buf;

	if (count == 0)
		buffer = (char *) &buf;

	ops = stream->descriptor.pointer;

	if (agl_ios_seek(ops, offset, SEEK_SET) < 0)
	{
		return 0;
	}
	return (unsigned long) agl_ios_read(ops, buffer, 1, count, NULL);
}

static void
agl_ttfsclose(FT_Stream stream)
{
	stream->descriptor.pointer = NULL;
	stream->size = 0;
	stream->base = 0;

	return;
}


//! initialise the free type font engine

/*! this fonction initialise the freetype font engine and the engine variable
 * \return 
 * nothing
 */

static int
agl_ttfinit(agl_module * pmod)
{
	FT_Error error;
	FT_Library library;

	if ((error = FT_Init_FreeType(&library)))
	{
		agl_ttferror("Couldn't init FreeType engine", error);
		return -1;
	}
	pmod->class_struct = library;
	return 1;
}

static int
agl_ttfis(agl_module * pmod, agl_ios * ops)
{
	FT_Open_Args args;
	FT_Error error;
	FT_Face face;
	struct FT_StreamRec_ stream;
	long size;

	agl_rvassert((pmod != NULL), -1);
	agl_rvassert((ops != NULL), -1);
	stream.base = NULL;
	if (agl_ios_seek(ops, 0, SEEK_SET) < 0)
	{
		return -1;
	}
	if ((size = agl_ios_seek(ops, 0, SEEK_END)) < 0)
	{
		return -1;
	}
	stream.size = size;
	if (agl_ios_seek(ops, 0, SEEK_SET) < 0)
	{
		return -1;
	}
	stream.pos = 0;
	stream.descriptor.pointer = ops;
	stream.read = agl_ttfsread;
	stream.close = agl_ttfsclose;

	args.flags = FT_OPEN_STREAM;
	args.stream = &stream;

	if ((error = FT_Open_Face((FT_Library) pmod->class_struct, &args, 0, &face)))
	{
		return 0;
	}
	FT_Done_Face(face);
	return 1;
}

//! open a ttf font file and create the font descriptor for the font size

/*! open a ttf font file and create the font descriptor for the font size
 * \param file the filename of the font to open
 * \param size the size of the font (pica)
 * \return
 * - NULL if the function fails
 * - the font descriptor
 */

static agl_font *
agl_ttfopen(agl_module * pmod, agl_ios * ops, int size)
{
	agl_font *font;
	ttf_font *ttfnt;
	FT_Open_Args args;
	FT_Fixed scale;
	long sz;
	FT_Error error;

	agl_rvassert((pmod != NULL), NULL);
	agl_rvassert((ops != NULL), NULL);

	font = agl_fontnew();
	if (font == NULL)
	{
		agl_memerr();
		return (NULL);
	}

	if (agl_ios_seek(ops, 0, SEEK_SET) < 0)
	{
		return NULL;
	}
	if ((sz = agl_ios_seek(ops, 0, SEEK_END)) < 0)
	{
		return NULL;
	}
	if (agl_ios_seek(ops, 0, SEEK_SET) < 0)
	{
		return NULL;
	}
	ttfnt = (ttf_font *) agl_malloc(sizeof(ttf_font));
	if (ttfnt == NULL)
	{
		agl_fontfree(font);
		agl_memerr();
		return (NULL);
	}
	ttfnt->stream.base = NULL;
	ttfnt->stream.size = sz;
	ttfnt->stream.pos = 0;
	ttfnt->stream.descriptor.pointer = ops;
	ttfnt->stream.read = agl_ttfsread;
	ttfnt->stream.close = agl_ttfsclose;

	args.flags = FT_OPEN_STREAM;
	args.stream = &ttfnt->stream;


	if ((error = FT_Open_Face((FT_Library) pmod->class_struct, &args, 0, &ttfnt->face)))
	{
		agl_ttferror("unable to load font file", error);
		agl_free(ttfnt);
		agl_free(font);
		return (NULL);
	}

	// set the display resolution (may vary on Xwindows from 72 DPI to 110 DPI)
	// we uses a 72 dpi resolution and hope the Xserver is 72 dpi, set 0 and 0 for resolutions

	if ((error = FT_Set_Char_Size(ttfnt->face, 0, size * 64, 0, 0)))
	{
		agl_ttferror("unable to set font resolution", error);
		FT_Done_Face(ttfnt->face);
		agl_free(ttfnt);
		agl_fontfree(font);
		return (NULL);
	}

	/* Get the font metrics for this font */


	scale = ttfnt->face->size->metrics.y_scale;
	font->ascent = FT_CEIL(FT_MulFix(ttfnt->face->bbox.yMax, scale));
	font->descent = FT_CEIL(FT_MulFix(ttfnt->face->bbox.yMin, scale));
	font->height = font->ascent - font->descent + /* baseline */ 1;
	font->linegap = FT_CEIL(FT_MulFix(ttfnt->face->height, scale));

	font->priv = ttfnt;
	font->pfio = (agl_fontio *) pmod->priv_desc;

	return font;
}

//! load if necessary the channel for a character

/*! load the true type font for a character
 * if the caracter should be loaded, a pixmap
 * is generated. This pixmap is stored into a channel
 * \param font the font descriptor structure.
 * \param ch the character to load
 * \return
 * - 0 if the font is loaded or if a pixmap already exists
 * - -1 if an error occurs
 */

static int
agl_ttfloadglyph(agl_font * font, unsigned char ch)
{
	agl_glyph *glyph;
	agl_channel *channel;
	ttf_font *ttfnt;
	unsigned int findex;
	FT_Error error;
	FT_GlyphSlot ftglyph;
	FT_Glyph_Metrics *metrics;
	unsigned char *sbuf, *dbuf, *pdbuf, *psbuf;
	int i, j;
	long minx, maxx, miny, maxy;

	agl_rvassert((font != NULL), -1);
	ttfnt = font->priv;
	glyph = font->glyphs + ch;
	channel = ttfnt->chan + ch;
	if (glyph->created)
		return 0;
	/* Load the glyph */
	findex = FT_Get_Char_Index(ttfnt->face, ch);
	if (findex == 0)
		return -1;
	if ((error = FT_Load_Glyph(ttfnt->face, findex, FT_LOAD_DEFAULT)))
	{
		agl_ttferror("Couldn't find glyph", error);
		return -1;
	}

	ftglyph = ttfnt->face->glyph;
	metrics = &ftglyph->metrics;
	minx = FT_FLOOR(metrics->horiBearingX);
	maxx = minx + FT_CEIL(metrics->width);
	maxy = FT_FLOOR(metrics->horiBearingY);
	miny = maxy - FT_CEIL(metrics->height);
	glyph->bbox.x = minx;
	glyph->bbox.y = maxy;
	glyph->bbox.width = maxx - minx;
	glyph->bbox.height = maxy - miny;
	glyph->yoffset = 0;
	glyph->advance = FT_CEIL(metrics->horiAdvance);

	if ((error = FT_Render_Glyph(ftglyph, ft_render_mode_normal)))
	{
		agl_ttferror("Couldn't render glyph", error);
		return -1;
	}

	channel->size.width = ftglyph->bitmap.width;
	channel->size.height = ftglyph->bitmap.rows;
	channel->pixels = NULL;
	channel->row_size = channel->size.width;
	if ((channel->size.width == 0) || (channel->size.height == 0))
	{
		glyph->created = 1;
		return 0;
	}
	channel->pixels = agl_malloc(channel->size.height * channel->row_size);
	if (channel->pixels == NULL)
	{
		agl_memerr();
		return -1;
	}
	/* beware, alpha channel is reversed */
	for (i = 0, sbuf = ftglyph->bitmap.buffer, dbuf = channel->pixels; i < channel->size.height; i++)
	{
		for (pdbuf = dbuf, psbuf = sbuf, j = 0; j < channel->row_size; j++)
			*pdbuf++ = 255 - *psbuf++;
		dbuf += channel->row_size;
		sbuf += ftglyph->bitmap.pitch;
	}
	/*
	 * { int i, j; 
	 * char *bm = channel->pixels;
	 * for(i = 0; i < channel->size.height; i++) {
	 * for (j = 0; j < channel->row_size; j++)
	 * printf("%2.2x ", (*bm++) & 0XFF);
	 * printf("\n");
	 * }
	 * }
	 */
	/* We've done, mark this glyph cached */
	glyph->created = 1;
	return 0;
}

static int
agl_ttfblit(agl_pix * pix, agl_font * font, unsigned char c, agl_pos * pos)
{
	int pw, ph, x, y, xpos, ypos;
	unsigned long *pr;
	unsigned char *pp, *pa, *ppa;
	ttf_font *ttfnt;
	agl_channel *chan;

	agl_rvassert(pix != NULL, -1);
	agl_rvassert(font != NULL, -1);
	agl_rvassert(pos != NULL, -1);
	agl_rvassert(font->glyphs[c].created, -1);

	ttfnt = (ttf_font *) font->priv;
	chan = ttfnt->chan + c;
	pw = pix->size.width;
	ph = pix->size.height;
	xpos = pos->x;
	if (((xpos + chan->size.width) <= 0) || (xpos >= pw))
		return 0;
	ypos = pos->y;
	if (((ypos + chan->size.height) <= 0) || (ypos >= ph))
		return 0;
	if ((xpos + chan->size.width) >= pw)
		pw -= xpos;
	else
		pw = chan->size.width;
	if ((ypos + chan->size.height) >= ph)
		ph -= ypos;
	else
		ph = chan->size.height;
	if ((pw <= 0) || (ph <= 0))
		return 0;
	agl_rvassert(pix->pixels != NULL, -1);
	pa = chan->pixels;
	if (pa == NULL)
		return 0;
	//printf("ypos:%d xpos:%d ph:%d pw:%d\n", ypos, xpos, ph, pw);
	pp = (unsigned char *) (pix->pixels + pix->row_size * ypos + xpos) + 3;
	for (y = 0; y < ph; y++, pp = (unsigned char *) (pr + pix->row_size))
	{
		pr = (unsigned long *) pp;
		ppa = pa;
		for (x = 0; x < pw; x++)
		{
			if ((y >= (-ypos)) && (x >= (-xpos)) && (*pa < 255))
				*pp = *pa;
			pa++;
			pp += 4;
		}
		pa = ppa + chan->row_size;
	}

	return 0;
}

static int
agl_ttfclose(agl_font * font)
{
	int i;
	ttf_font *ttfnt;

	agl_rvassert(font != NULL, -1);
	ttfnt = (ttf_font *) font->priv;
	for (i = 0; i < 256; i++)
		if (font->glyphs[i].created)
			if (ttfnt->chan[i].pixels != NULL)
				agl_free(ttfnt->chan[i].pixels);
	FT_Done_Face(ttfnt->face);
	agl_free(ttfnt);
	return 0;
}

//! exit from the ttf system

/*! exit from the ttf system. free all resources.
 * \warning the ttf font structures are not freed by this function
 * \return nothing
 */

static int
agl_ttfexit(agl_module * pmod)
{

	agl_rvassert(pmod != NULL, -1);

	if (pmod->class_struct != NULL)
	{
		FT_Done_FreeType((FT_Library) (pmod->class_struct));
	}
	pmod->class_struct = NULL;

	return 1;
}

static agl_fontio agl_ttf_fontio = {
	agl_ttfis,
	agl_ttfopen,
	agl_ttfloadglyph,
	agl_ttfblit,
	agl_ttfclose
};

agl_module agl_modtab_ttfio = {
	"TTF",
	"loader for truetype fonts",
	AGL_MOD_FONT,
	{{NULL, NULL}},
	(void *) &agl_ttf_fontio,
	NULL,
	agl_ttfinit,
	NULL,
	NULL,
	NULL,
	NULL,
	agl_ttfexit,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#else

agl_module agl_modtab_ttfio = {
	NULL,
	NULL,
	AGL_MOD_FONT,
	{{NULL, NULL}
	 }
	,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#endif /* HAVE_TTF */
