
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! This file manages inputs from all the medias used by adgali

/*! \file
 * \ingroup ios
 * This file manges the input and output from various medias into adgali.
 * yet, it manages file stream (files), file ios, memory zones
 * The output is not implemented.
 * \todo sockets ios (TCP and UDP) should be implemented in the future
 */


#include <string.h>

#include "config.h"
#include "agl_private.h"

struct priv_mem
{
	unsigned char *mem;
	unsigned long size;
	unsigned long pos;
};

static int
fmem_is(agl_module * pmod, char *name, agl_ios * parent)
{
	long mem, size;
	char *p;

	agl_rvassert(pmod != NULL, -1);
	agl_rvassert(parent == NULL, -1);

	p = agl_getnum(name, &mem);
	if (p == NULL)
	{
		agl_error("MEM: unable to get mem desc");
		return -1;
	}
	if (*p++ != ':')
	{
		agl_error("MEM: unable to get mem desc");
		return -1;
	}
	p = agl_getnum(name, &size);
	if (p == NULL)
	{
		agl_error("MEM: unable to get mem desc");
		return -1;
	}
	return 1;
}

static agl_ios *
fmem_open(agl_module * pmod, char *name, agl_ios * parent)
{
	long mem, size;
	char *p;
	agl_ios *ios;
	struct priv_mem *pmem;

	agl_rvassert(pmod != NULL, NULL);
	agl_rvassert(parent == NULL, NULL);

	p = agl_getnum(name, &mem);
	if (p == NULL)
	{
		agl_error("MEM: unable to get mem desc");
		return NULL;
	}
	if (*p++ != ':')
	{
		agl_error("MEM: unable to get mem desc");
		return NULL;
	}
	p = agl_getnum(name, &size);
	if (p == NULL)
	{
		agl_error("MEM: unable to get mem desc");
		return NULL;
	}

	if ((ios = agl_malloc(sizeof(agl_ios))) == NULL)
	{
		agl_memerr();
		return NULL;
	}
	if ((pmem = agl_malloc(sizeof(struct priv_mem))) == NULL)
	{
		agl_memerr();
		agl_free(ios);
		return NULL;
	}
	pmem->mem = (unsigned char *) mem;
	pmem->size = size;
	pmem->pos = 0;
	if ((ios->filename = agl_malloc(p - name + 1)) == NULL)
	{
		agl_memerr();
		agl_free(ios);
		agl_free(pmem);
		return NULL;
	}
	strncpy(ios->filename, name, p - name);
	ios->priv = pmem;
	ios->pio = (agl_iosio *) pmod->priv_desc;
	ios->handle = pmod->handle;

	return ios;

}


static long
fmem_seek(agl_ios * src, long offset, int whence)
{
	long pos;
	struct priv_mem *pmem;

	agl_rvassert(src != NULL, -1);

	pmem = (struct priv_mem *) src->priv;

	agl_rvassert(pmem != NULL, -1);

	switch (whence)
	{
	case SEEK_SET:
		pos = offset;
		break;
	case SEEK_CUR:
		pos = pmem->pos + offset;
		break;
	case SEEK_END:
		pos = pmem->size + offset;
		break;
	default:
		agl_ioserr();
		return -1;
	}

	if (pos < 0)
		pos = 0;

	if ((unsigned long) pos > pmem->size)
		pos = pmem->size;

	pmem->pos = pos;

	return pos;
}

static long
fmem_read(agl_ios * src, void *buf, long size, long maxnum, void *supp)
{
	long sz, rec;
	struct priv_mem *pmem;

	agl_rvassert(src != NULL, -1);

	pmem = (struct priv_mem *) src->priv;

	agl_rvassert(pmem != NULL, -1);

	agl_rvassert(buf != NULL, -1);

	supp = NULL;

	rec = (pmem->size - pmem->pos) / size;
	rec = (rec < maxnum) ? rec : maxnum;
	sz = rec * size;

	memcpy(buf, pmem->mem + pmem->pos, sz);

	pmem->pos += sz;

	return rec;
}

static long
fmem_write(agl_ios * src, void *buf, long size, long maxnum, void *supp)
{
	long sz, rec;
	struct priv_mem *pmem;

	agl_rvassert(src != NULL, -1);

	pmem = (struct priv_mem *) src->priv;

	agl_rvassert(pmem != NULL, -1);

	agl_rvassert(buf != NULL, -1);

	supp = NULL;

	rec = (pmem->size - pmem->pos) / size;
	rec = (rec < maxnum) ? rec : maxnum;
	sz = rec * size;

	memcpy(pmem->mem + pmem->pos, buf, sz);

	pmem->pos += sz;

	return rec;
}

static int
fmem_close(agl_ios * src)
{
	agl_rvassert(src != NULL, -1);

	agl_free(src->priv);
	agl_free(src->filename);
	agl_free(src);
	return 0;
}

static agl_iosio agl_fmem_iosio = {
	"MEM",
	fmem_is,
	fmem_open,
	fmem_seek,
	fmem_read,
	fmem_write,
	fmem_close,
};

agl_module agl_modtab_fmem = {
	"MEMORY",
	"loader for memory",
	AGL_MOD_IO,
	{{NULL, NULL}},
	(void *) &agl_fmem_iosio,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};
