
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! This file implements the lexical interpreter and the scanner

/*! \file
 * \ingroup module
 * here is described the lexical analyser (integer, characters, float
 * octal, binary and hexadecimal numbers, strings, identifier, keywords
 * and token).
 * the rest of this file contains the tokenised string analyser (created after
 * parsing). each token is interpreted and an action is executed. The interpretation
 * of the string instanciates all the objects.
 * Temporary variables and the tokenised string are then freed
 */


#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "config.h"
#include "agl_private.h"
#include "agl_gramint.h"
#include "agl_gram.h"

#define CHAR_TOKEN	0
#define STRING_TOKEN	1
#define ZDEC_TOKEN	2
#define NUMBER_TOKEN	3
#define FLOAT_TOKEN	4
#define INT_TOKEN	5
#define HEXA_TOKEN	6
#define BIN_TOKEN	7
#define OCT_TOKEN	8
#define SYMB_TOKEN	9
#define SKIP_TOKEN	10

int agllex(YYSTYPE * pval, agl_analyse * analyse);
int aglparse(agl_analyse *);

static int
agl_graminitcode(agl_analyse * analyse, agl_module * modlist, agl_ios * ops)
{
	agl_rvassert(analyse != NULL, -1);
	agl_rvassert(ops != NULL, -1);

	analyse->symlist = NULL;

	analyse->garblist = NULL;

	analyse->codelist = NULL;
	analyse->codeptr = NULL;

	analyse->stacklist = NULL;

	analyse->pmodule = modlist;
	analyse->curmodule = NULL;
	analyse->cursymb = NULL;

	analyse->lineno = 1;

	analyse->ops = ops;
	return 1;
}

static int
agl_searchmod(agl_analyse * anal, char *s, agl_module ** pmod)
{
	agl_module *pcur;

	agl_rvassert(anal != NULL, -1);
	agl_rvassert(s != NULL, -1);
	agl_rvassert(pmod != NULL, -1);

	for (pcur = anal->pmodule; pcur; pcur = pcur->pnext)
	{
		if ((strcmp(s, pcur->name) == 0) && (pcur->pmodelem != NULL))
		{
			*pmod = pcur;
			return 1;
		}
	}
	return 0;

}

static int
agl_searchkeyw(agl_analyse * anal, char *s, agl_modelem ** pkeyw)
{
	agl_modelem *pcur;

	agl_rvassert(anal != NULL, -1);
	agl_rvassert(s != NULL, -1);
	agl_rvassert(pkeyw != NULL, -1);

	if (anal->curmodule == NULL)
		return 0;

	if (anal->curmodule->pmodelem == NULL)
		return 0;

	for (pcur = anal->curmodule->pmodelem; pcur->name; pcur++)
	{
		if (strcmp(s, pcur->name) == 0)
		{
			*pkeyw = pcur;
			return 1;
		}
	}
	return 0;
}

static int
agl_searchdefn(agl_analyse * anal, char *s, agl_moddefn ** pdefn)
{
	agl_module *pmod;
	agl_moddefn *pcur;

	agl_rvassert(anal != NULL, -1);
	agl_rvassert(s != NULL, -1);

	if (anal->curmodule == NULL)
		return 0;

	if (anal->curmodule->pmoddefn == NULL)
		return 0;

	for (pmod = anal->pmodule; pmod; pmod = pmod->pnext)
	{
		for (pcur = pmod->pmoddefn; pcur->name; pcur++)
		{
			if (strcmp(s, pcur->name) == 0)
			{
				*pdefn = pcur;
				return 1;
			}
		}
	}
	return 0;
}

static int
agl_lookupsymb(agl_analyse * analyse, char *s, agl_gramsymb ** psymb)
{
	agl_gramsymb *sp;

	agl_rvassert(analyse != NULL, -1);
	agl_rvassert(s != NULL, -1);

	sp = analyse->symlist;
	for (; sp; sp = sp->pnext)
		if (strncmp(sp->name, s, 30) == 0)
		{
			*psymb = sp;
			return 1;
		}
	return 0;
}

static int
agl_createsymb(agl_analyse * analyse, char *s, agl_gramsymb ** psymb)
{
	int ret;
	agl_object *pobj;
	char *ptr;

	agl_rvassert(analyse != NULL, -1);
	agl_rvassert(s != NULL, -1);
	agl_rvassert(psymb != NULL, -1);

	ret = agl_lookupsymb(analyse, s, psymb);
	if (ret != 0)
		return ret;

	*psymb = (agl_gramsymb *) agl_malloc(sizeof(agl_gramsymb));
	if (*psymb == NULL)
	{
		agl_memerr();
		return -1;
	}
	strncpy((*psymb)->name, s, 30);
	(*psymb)->val.vtype = AGL_STYPENONE;
	(*psymb)->val.itype = AGL_STYPENONE;
	(*psymb)->pnext = analyse->symlist;
	analyse->symlist = *psymb;

	if (agl_cachesearch(s, analyse->ops->handle, &pobj) == 1)
	{
		(*psymb)->val.vtype = AGL_STYPEOBJREF;

		ptr = (char *) agl_malloc(strlen(pobj->name) + 1);
		if (ptr == NULL)
		{
			agl_memerr();
			return -1;
		}
		(*psymb)->val.vu.ref.name = ptr;
		strcpy(ptr, pobj->name);

		ptr = (char *) agl_malloc(strlen(pobj->pmod->name) + 1);
		if (ptr == NULL)
		{
			agl_memerr();
			return -1;
		}
		(*psymb)->val.vu.ref.mod = ptr;
		strcpy(ptr, pobj->pmod->name);
	}
	return 1;
}

static int
agl_deletesymb(agl_analyse * analyse, int destroy)
{
	agl_gramsymb *pcur, *pnext;
	int ret;

	ret = 0;
	for (pcur = analyse->symlist; pcur; pcur = pnext)
	{

		pnext = pcur->pnext;
		if (agl_freeval(&pcur->val, destroy) < 0)
			ret = -1;
		agl_free(pcur);
	}
	return ret;
}

static int
agl_cachesymb(agl_analyse * analyse)
{
	agl_gramsymb *pcur;
	int ret;

	ret = 0;
	for (pcur = analyse->symlist; pcur; pcur = pcur->pnext)
	{
		if (!(pcur->val.vtype & AGL_STYPESTATIC) && (pcur->val.vtype & AGL_STYPEUSER)
		    && ((pcur->val.itype & AGL_STYPEALL) == AGL_STYPENONE))
			if (agl_cacheadd(pcur->val.vu.user) < 0)
				return -1;
	}
	return ret;
}

static int
get_char(agl_ios * ops, unsigned char *ch)
{
	unsigned char c;
	int val, ret;
	int autom;
	long pos;

	agl_rvassert(ch != NULL, -1);
	autom = 0;
	val = 0;
	pos = agl_ios_tell(ops);
	if (pos < 0)
	{
		agl_ioserr();
		return -1;
	}
	while (1)
	{
		if ((ret = agl_ios_read(ops, &c, 1, 1, NULL)) <= 0)
		{
			if (ret < 0)
			{
				agl_ioserr();
				return -1;
			}
			else
			{
				if (agl_ios_seek(ops, pos, SEEK_SET) < 0)
				{
					agl_ioserr();
					return -1;
				}
				return 0;
			}

		}
		switch (autom)
		{
		case 0:
			if (c != '\\')
			{
				*ch = c;
				return 1;
			}
			else
				autom = 2;
			break;
		case 2:
			switch (c)
			{
			case 'a':
				*ch = 7;	/* BEL */
				return 1;
			case 'b':
				*ch = 8;	/* BACKSPACE */
				return 1;
			case 'f':
				*ch = 12;	/* FORM FEED */
				return 1;
			case 'n':
				*ch = 10;	/* LINE FEED */
				return 1;
			case 'r':
				*ch = 13;	/* CARRIAGE RETURN */
				return 1;
			case 't':
				*ch = 9;	/* HORIZONTAL TAB */
				return 1;
			case 'v':
				*ch = 11;	/* VERTICAL TAB */
				return 1;
			case 'x':
			case 'X':
				autom = 3;
				break;
			default:
				if ((c >= '0') && (c <= '3'))
				{
					val = c - '0';
					autom = 6;
				}
				else
				{
					*ch = c;
					return 1;
				}
			}
			break;
		case 3:
		case 4:
			autom++;
			if ((c >= '0') && (c <= '9'))
				val = val * 16 + c - '0';
			else if ((c >= 'a') && (c <= 'f'))
				val = val * 16 + c - 'a' + 10;
			else if ((c >= 'A') && (c <= 'F'))
				val = val * 16 + c - 'A' + 10;
			else
				return 0;
			if (autom == 5)
			{
				*ch = val;
				return 1;
			}
			break;
		case 6:
		case 7:
			autom++;
			if ((c >= '0') && (c <= '7'))
				val = val * 8 + c - '0';
			else
			{
				return 0;
			}
			if (autom == 8)
			{
				*ch = val;
				return 1;
			}
			break;
		default:
			agl_error2("illegal number autom (%d)", autom);
		}

	}
	return 0;
}

int
agllex(YYSTYPE * pval, agl_analyse * analyse)
{
	int type, autom, status, ret;
	int expo, sign, msign, dec, cnt;
	double val;
	unsigned char *string, *ptr;
	unsigned char c;
	long pos, uval;
	agl_ios *ops;

	agl_rvassert(pval != NULL, 0);
	agl_rvassert(analyse != NULL, 0);

	ops = analyse->ops;

	agl_rvassert(ops != NULL, 0);

	type = -1;
	val = 0;
	uval = 0;
	status = 0;
	autom = 0;
	expo = 0;
	sign = 1;
	msign = 1;
	dec = 0;
	string = NULL;
	cnt = 0;
	pos = agl_ios_tell(ops);
	if (pos < 0)
	{
		agl_ioserr();
		return -1;
	}
	while (1)
	{
		if (agl_ios_read(ops, &c, 1, 1, NULL) <= 0)
			status = 6;	/* BAD; treat end of file */
		else
			switch (type)
			{
			case -1:
				switch (c)
				{
				case '\'':
					type = CHAR_TOKEN;
					break;
				case '"':
					type = STRING_TOKEN;
					break;
				case '0':
					type = ZDEC_TOKEN;
					break;
				case '-':
					msign = -1;
				case '+':
					type = NUMBER_TOKEN;
					break;
				case '.':
					type = FLOAT_TOKEN;
					break;
				case '#':
					type = SKIP_TOKEN;
					break;
				default:
					if (isdigit(c))
					{
						type = INT_TOKEN;
						uval = c - '0';
					}
					else
					{
						switch (c)
						{
						case '\n':
							analyse->lineno++;
							break;
						case '\r':
							break;
						default:
							if (isspace(c))
								break;
							if (isalpha(c) || (c == '_'))
							{
								type = SYMB_TOKEN;
								string = (unsigned char *) agl_malloc(100);
								if (string == NULL)
								{
									agl_memerr();
									status = 6;
									break;
								}
								string[cnt++] = c;
							}
							else
								return c;	/* other characters */
						}
					}
				}
				break;
			case SKIP_TOKEN:
				if (c == '\n')
				{
					analyse->lineno++;
					type = -1;
				}
				break;
			case SYMB_TOKEN:
				if (isalnum(c) || (c == '_'))
				{
					string[cnt++] = c;
					if ((cnt % 100) == 0)
					{
						ptr = (unsigned char *) agl_realloc(string, cnt + 100);
						if (ptr == NULL)
						{
							agl_memerr();
							status = 6;
							break;
						}
						string = ptr;
					}
				}
				else
					status = 7;
				break;
			case CHAR_TOKEN:
				switch (autom)
				{
				case 0:
					autom++;
					if (agl_ios_seek(ops, -1, SEEK_CUR) < 0)
					{
						status = 6;
						break;
					}
					if ((ret = get_char(ops, &c)) < 0)
						status = 6;
					else if (ret == 0)
						status = 2;
					else
						uval = c;
					break;
				case 1:
					if (c != '\'')
						status = 2;
					else
						status = 8;
					break;
				default:
					agl_error2("illegal number autom (%d)", autom);
					status = 6;
				}
				break;
			case STRING_TOKEN:
				if (string == NULL)
				{
					string = (unsigned char *) agl_malloc(100);
					if (string == NULL)
					{
						agl_memerr();
						status = 6;
						break;
					}
				}
				if (c == '"')
				{
					status = 5;
					break;
				}
				if (agl_ios_seek(ops, -1, SEEK_CUR) < 0)
				{
					status = 6;
					break;
				}
				if ((ret = get_char(ops, &c)) < 0)
					status = 6;
				if (ret == 0)
					status = 2;
				string[cnt++] = c;
				if ((cnt % 100) == 0)
				{
					ptr = (unsigned char *) agl_realloc(string, cnt + 100);
					if (ptr == NULL)
					{
						agl_memerr();
						status = 6;
						break;
					}
					string = ptr;
				}
				break;
			case NUMBER_TOKEN:
				if (c == '0')
				{
					type = ZDEC_TOKEN;
				}
				else
				{
					if (isdigit(c))
					{
						type = INT_TOKEN;
						uval = c - '0';
					}
					else
						status = 2;
				}
				break;
			case ZDEC_TOKEN:
				switch (c)
				{
				case '.':
					type = FLOAT_TOKEN;
					break;
				case 'x':
				case 'X':
					type = HEXA_TOKEN;
					break;
				case 'b':
				case 'B':
					type = BIN_TOKEN;
					break;
				default:
					if (c < '0')
						status = 3;
					else if (c <= '7')
					{
						type = OCT_TOKEN;
						uval = c - '0';
						status = 1;
					}
					else if (c <= '9')
					{
						type = INT_TOKEN;
						uval = c - '0';
					}
					else
						status = 3;
					break;
				}
				break;
			case HEXA_TOKEN:
				if (isdigit(c))
				{
					uval = uval * 16 + c - '0';
					status = 1;
				}
				else
				{
					c = toupper(c);
					if ((c >= 'A') && (c <= 'F'))
					{
						uval = uval * 16 + c - 'A' + 10;
						status = 1;
					}
					else if (status == 0)
						status = 2;
					else
						status = 3;
				}
				break;
			case OCT_TOKEN:
				if ((c >= '0') && (c <= '7'))
					uval = uval * 8 + c - '0';
				else
					status = 3;
				break;
			case INT_TOKEN:
				if ((c >= '0') && (c <= '9'))
				{
					uval = uval * 10 + c - '0';
				}
				else if (c == '.')
				{
					type = FLOAT_TOKEN;
					val = uval;
				}
				else if ((c == 'e') || (c == 'E'))
				{
					type = FLOAT_TOKEN;
					val = uval;
					autom = 1;
				}
				else
					status = 3;
				break;
			case BIN_TOKEN:
				if ((c >= '0') && (c <= '1'))
				{
					uval = uval * 2 + c - '0';
					status = 1;
				}
				else if (status == 0)
					status = 2;
				else
					status = 3;
				break;
			case FLOAT_TOKEN:
				switch (autom)
				{
				case 0:
					if ((c >= '0') && (c <= '9'))
					{
						dec++;
						val = val * 10 + c - '0';
					}
					else if ((c == 'e') || (c == 'E'))
						autom = 1;
					else
						autom = -1;
					break;
				case 1:
					autom = 2;
					if (c == '-')
						sign = -1;
					else if (c == '+')
						sign = 1;
					else if (isdigit(c))
						expo = c - '0';
					else
						status = 2;
					break;
				case 2:
					if (isdigit(c))
					{
						expo = expo * 10 + c - '0';
					}
					else
						autom = -1;
					break;
				default:
					agl_error("bad switch state");
					status = 6;
				}
				if (autom == -1)
				{
					expo *= sign;
					expo -= dec;
					val *= pow(10.0, (double) expo);
					status = 4;
				}
				break;
			default:
				agl_error2("illegal lexical type (%d)", type);
				status = 6;
				break;
			}
		if (status == 2)
		{
			if (string != NULL)
				agl_free(string);
			agl_error("lexical analysis error");
			if (agl_ios_seek(ops, pos, SEEK_SET) < 0)
			{
				agl_ioserr();
				return -1;
			}
			return 0;
		}
		if (status == 3)
		{
			if (agl_ios_seek(ops, -1, SEEK_CUR) < 0)
			{
				agl_ioserr();
				return -1;
			}
			pval->val.pnext = NULL;
			pval->val.itype = AGL_STYPENONE;
			pval->val.vtype = AGL_STYPEINT;
			pval->val.vu.num = (long) uval *msign;

			return (NUMBER);
		}
		if (status == 4)
		{
			if (agl_ios_seek(ops, -1, SEEK_CUR) < 0)
			{
				agl_ioserr();
				return -1;
			}
			pval->val.pnext = NULL;
			pval->val.itype = AGL_STYPENONE;
			pval->val.vtype = AGL_STYPEDBL;
			pval->val.vu.dbl = val * msign;
			return (FNUMBER);
		}
		if (status == 5)
		{
			string[cnt++] = '\0';
			ptr = (unsigned char *) agl_realloc(string, cnt);
			if (ptr == NULL)
			{
				agl_memerr();
				agl_free(string);
				return -1;
			}
			pval->val.pnext = NULL;
			pval->val.itype = AGL_STYPENONE;
			pval->val.vtype = AGL_STYPESTR;
			pval->val.vu.str = ptr;
			return (STRING);
		}
		if (status == 6)
		{
			if (type < 0)
				return 0;	/* end of file */
			agl_error("IO error");
			if (string != NULL)
				agl_free(string);
			if (agl_ios_seek(ops, pos, SEEK_SET) < 0)
				agl_ioserr();
			return -1;
		}
		if (status == 7)
		{
			if (agl_ios_seek(ops, -1, SEEK_CUR) < 0)
			{
				agl_ioserr();
				return -1;
			}
			string[cnt++] = '\0';
			ptr = (unsigned char *) agl_realloc(string, cnt);
			if (ptr == NULL)
			{
				agl_memerr();
				agl_free(string);
				return -1;
			}
			ret = agl_searchkeyw(analyse, string, &pval->keyw);
			if (ret == 1)
			{
				agl_free(string);
				return KEYW;
			}
			if (ret < 0)
			{
				agl_free(string);
				return -1;
			}
			ret = agl_searchdefn(analyse, string, &pval->defn);
			if (ret == 1)
			{
				agl_free(string);
				return DEFN;
			}
			if (ret < 0)
			{
				agl_free(string);
				return -1;
			}
			ret = agl_searchmod(analyse, string, &pval->mod);
			if (ret == 1)
			{
				agl_free(string);
				return MOD;
			}
			if (ret < 0)
			{
				agl_free(string);
				return -1;
			}
			ret = agl_createsymb(analyse, string, &pval->sym);
			agl_free(string);
			if (ret == 1)
				return IDENT;
			return -1;
		}
		if (status == 8)
		{
			pval->val.pnext = NULL;
			pval->val.itype = AGL_STYPENONE;
			pval->val.vtype = AGL_STYPEINT;
			pval->val.vu.num = (long) uval;
			return (NUMBER);
		}

	}
	agl_error("unexpected break");
	return (-1);
}


static agl_graminstr *
agl_gramnewinst(agl_analyse * analyse)
{
	agl_graminstr *nptr;

	agl_rvassert(analyse != NULL, NULL);

	nptr = (agl_graminstr *) agl_malloc(sizeof(agl_graminstr));
	if (nptr == NULL)
	{
		agl_memerr();
		return NULL;
	}
	nptr->pgnext = analyse->garblist;
	analyse->garblist = nptr;

	return nptr;
}

static int
agl_gramfreeinst(agl_analyse * analyse, int destroy)
{
	agl_graminstr *pcur, *pnext;
	int ret;

	agl_rvassert(analyse != NULL, -1);

	ret = 0;
	for (pcur = analyse->garblist; pcur; pcur = pnext)
	{
		pnext = pcur->pgnext;
		if (pcur->type == AGL_GRAMVAL)
			if (agl_freeval(&pcur->u.val, destroy) < 0)
				ret = -1;
		agl_free(pcur);
	}
	analyse->garblist = NULL;
	return ret;
}

static int
agl_grampush(agl_analyse * anal, agl_graminstr * s)
{
	agl_rvassert(anal != NULL, -1);
	agl_rvassert(s != NULL, -1);

	s->pnext = anal->stacklist;
	anal->stacklist = s;
	return 1;
}

static int
agl_grampop(agl_analyse * anal, agl_graminstr ** s)
{
	agl_rvassert(anal != NULL, -1);
	agl_rvassert(s != NULL, -1);

	if (anal->stacklist == NULL)
		return 0;
	*s = anal->stacklist;
	anal->stacklist = (*s)->pnext;
	return 1;
}

int
agl_valpush(agl_analyse * anal)
{
	agl_graminstr *p;

	agl_rvassert(anal != NULL, -1);

	p = anal->codeptr->pnext;

	agl_rvassert(p != NULL, -1);

	anal->codeptr->pnext = p->pnext;

	return agl_grampush(anal, p);
}

int
agl_keywstmt2(agl_analyse * anal)
{
	agl_graminstr *s;

	agl_rvassert(anal != NULL, -1);

	agl_rvassert(anal->curmodule == NULL, -1);

	s = anal->codeptr->pnext;

	agl_rvassert(s != NULL, -1);

	anal->codeptr->pnext = s->pnext;

	agl_rvassert(s->type == AGL_GRAMMOD, -1);

	anal->curmodule = s->u.mod;

	s = anal->codeptr->pnext;

	agl_rvassert(s != NULL, -1);

	anal->codeptr->pnext = s->pnext;

	if (s->type != AGL_GRAMSYMB)
		return -1;

	agl_rvassert(s->u.symb != NULL, -1);

	if ((s->u.symb->val.vtype & AGL_STYPEALL) != AGL_STYPENONE)
	{
		agl_error2("object %s already declared", s->u.symb->name);
		return -1;
	}

	s->u.symb->val.vu.user = agl_objcreate(s->u.symb->name, anal->curmodule);
	if (s->u.symb->val.vu.user == NULL)
		return -1;
	s->u.symb->val.vtype = AGL_STYPEUSER;
	anal->cursymb = s->u.symb;

	return 0;
}

int
agl_keywadd(agl_analyse * anal)
{
	agl_object *pobj;
	agl_graminstr *s;

	agl_rvassert(anal != NULL, -1);

	agl_rvassert(anal->curmodule == NULL, -1);

	s = anal->codeptr->pnext;

	agl_rvassert(s != NULL, -1);

	anal->codeptr->pnext = s->pnext;

	agl_rvassert(s->type == AGL_GRAMMOD, -1);

	anal->curmodule = s->u.mod;

	s = anal->codeptr->pnext;

	agl_rvassert(s != NULL, -1);

	anal->codeptr->pnext = s->pnext;

	if (s->type != AGL_GRAMSYMB)
		return -1;

	agl_rvassert(s->u.symb != NULL, -1);

	if ((s->u.symb->val.vtype & AGL_STYPEALL) == AGL_STYPENONE)
	{
		agl_error2("object %s never defined", s->u.symb->name);
		return -1;
	}

	switch (s->u.symb->val.vtype & AGL_STYPEALL)
	{
	case AGL_STYPEUSER:
		pobj = s->u.symb->val.vu.user;
		break;
	case AGL_STYPEOBJREF:
		if (agl_cacheobjget(s->u.symb->val.vu.ref.name, anal->ops->handle, &pobj) != 1)
		{
			agl_error2("object %s not found or error", s->u.symb->val.vu.ref.name);
			return -1;
		}
		agl_free(s->u.symb->val.vu.ref.name);
		agl_free(s->u.symb->val.vu.ref.mod);
		break;
	default:
		agl_error("object of invalid type");
		return -1;
	}

	if (agl_cacheobjevent(pobj, AGL_OBJEVT_UNLOAD, pobj) < 0)
	{
		agl_error2("object %s : event failed", pobj->name);
		return -1;
	}

	anal->cursymb = s->u.symb;

	anal->cursymb->val.vtype = AGL_STYPEUSER;
	anal->cursymb->val.vu.user = pobj;

	return 0;
}

int
agl_keywlike(agl_analyse * anal)
{
	agl_graminstr *s;
	agl_object *pobj, *pcurobj;
	agl_elemtype *pdef;
	agl_modelem *pelem;
	int idx;

	agl_rvassert(anal != NULL, -1);

	agl_rvassert(anal->curmodule == NULL, -1);

	s = anal->codeptr->pnext;

	agl_rvassert(s != NULL, -1);

	anal->codeptr->pnext = s->pnext;

	agl_rvassert(s->type == AGL_GRAMSYMB, -1);

	agl_rvassert(s->u.symb == NULL, -1);

	switch (s->u.symb->val.vtype & AGL_STYPEALL)
	{
	case AGL_STYPEUSER:
		pobj = s->u.symb->val.vu.user;
		break;
	case AGL_STYPEOBJREF:
		if (agl_cacheobjget(s->u.symb->val.vu.ref.name, anal->ops->handle, &pobj) != 1)
		{
			agl_error2("object %s not found or error", s->u.symb->val.vu.ref.name);
			return -1;
		}
		break;
	default:
		agl_error("object of invalid type");
		return -1;
	}

	pcurobj = anal->cursymb->val.vu.user;
	pelem = pcurobj->pmod->pmodelem;
	pdef = pobj->defelem;
	idx = 0;

	while (pelem->name)
	{			// copy all elements
		if (agl_objaddelem(pcurobj, idx, pdef) < 0)
		{
			return -1;
		}
		pelem++;
		pdef++;
		idx++;
	}

	return 0;
}

static int
agl_validstr(char **str, int *pidx, int nb)
{
	int valid;
	int ret, i;
	int state = 0;
	unsigned char c;
	int bit;

	valid = 1;
	for (; **str; (*str)++)
	{
		c = **str;
		switch (state)
		{
		case 0:	// init ( -> ch (
			state = 1;
			if (c == '(')
			{
				valid = agl_validstr(str, pidx, nb);
				if (valid < 0)
					return -1;
				if (**str != ')')
					return -1;
				continue;
			}
			break;
		case '1':	// ) ch -> ch ( | )
			if (c == ')')
				return valid;
			if (c == '(')
			{
				ret = agl_validstr(str, pidx, nb);
				if (ret < 0)
					return -1;
				valid &= ret;
				if (**str != ')')
					return -1;
				continue;
			}
			if (c == '|')
			{
				ret = agl_validstr(str, pidx, nb);
				if (ret < 0)
					return -1;
				valid |= ret;
				return valid;
			}
		default:
			agl_error("invalid state");
			return -1;
		}
		if ((c >= '0') || (c <= '9'))
			bit = c - '0';
		else if ((c >= 'A') || (c <= 'U'))
			bit = c - 'A' + 10;
		else if ((c >= 'a') || (c <= 'u'))
			bit = c - 'a' + 10;
		else
			return -1;
		for (i = 0; i < nb; i++)
			if (pidx[i] == bit)
				valid &= 1;
		valid = 0;
	}
	return valid;
}

int
agl_keywend(agl_analyse * anal)
{
	int ret;
	char *ptr;

	agl_rvassert(anal != NULL, -1);

	agl_rvassert(anal->curmodule != NULL, -1);

	if ((anal->cursymb->val.vtype & AGL_STYPEALL) == AGL_STYPEUSER)
	{
		if (anal->curmodule->validstr != NULL)
		{
			ptr = anal->curmodule->validstr;
			ret = agl_validstr(&ptr, anal->cursymb->val.vu.user->defidx, anal->cursymb->val.vu.user->defmax);
			if ((ret < 0) || (*ptr != '\0'))
			{
				anal->cursymb->val.vtype = AGL_STYPENONE;
				agl_objdelete(anal->cursymb->val.vu.user);
				agl_error2("validation string error in %s", anal->curmodule->name);
				return -1;
			}
			if (ret == 0)
			{
				anal->cursymb->val.vtype = AGL_STYPENONE;
				agl_objdelete(anal->cursymb->val.vu.user);
				agl_error3("The statement %s named %s is not valid", anal->curmodule->name, anal->cursymb->name);
				return -1;
			}
		}
		if (anal->curmodule->mod_validate != NULL)
		{
			ret = anal->curmodule->mod_validate(anal->cursymb->val.vu.user);
			if (ret <= 0)
			{
				anal->cursymb->val.vtype = AGL_STYPENONE;
				agl_objdelete(anal->cursymb->val.vu.user);
				agl_error3("The statement %s named %s is not valid", anal->curmodule->name, anal->cursymb->name);
				return -1;
			}
		}
		anal->cursymb->val.vu.user->flags |= AGL_OBJ_VALID;
	}
	anal->curmodule = NULL;
	anal->cursymb = NULL;

	return 0;
}

int
agl_ieval(agl_analyse * anal)
{
	agl_graminstr *s;
	agl_elemtype *pval = NULL;
	agl_elemtype lval;
	int ret;

	agl_rvassert(anal != NULL, -1);

	if ((ret = agl_grampop(anal, &s)) < 0)
	{
		agl_error("stack empty");
		return -1;
	}
	switch (s->type)
	{
	case AGL_GRAMNONE:
	case AGL_GRAMINSTR:
	case AGL_GRAMMOD:
	case AGL_GRAMLSTSZ:
	case AGL_GRAMKEYW:
	default:
		agl_error2("cannot evaluate: %d", s->type);
		return -1;
	case AGL_GRAMVAL:
		break;
	case AGL_GRAMSYMB:
		if ((s->u.symb->val.vtype & AGL_STYPEALL) == AGL_STYPENONE)
		{
			agl_error2("cannot evaluate: %d", s->type);
			return -1;
		}
		pval = &(s->u.symb->val);
		break;
	case AGL_GRAMDEFN:
		if ((s->u.defn->val.vtype & AGL_STYPEALL) == AGL_STYPENONE)
		{
			agl_error2("cannot evaluate: %d", s->type);
			return -1;
		}
		if (agl_elemexp2u(&s->u.defn->val, &lval) < 0)
		{
			agl_error2("cannot evaluate: %d", s->type);
			return -1;
		}
		pval = &lval;
		break;
	}
	s->type = AGL_GRAMVAL;
	s->u.val = *pval;
	return agl_grampush(anal, s);
}

int
agl_listcrnpush(agl_analyse * anal)
{
	agl_graminstr *s;
	agl_graminstr *v;
	int lg, i;
	int ret;

	agl_rvassert(anal != NULL, -1);

	s = anal->codeptr->pnext;
	agl_rvassert(s != NULL, -1);

	anal->codeptr->pnext = s->pnext;

	agl_rvassert(s->type == AGL_GRAMLSTSZ, -1);
	lg = s->u.lstsz;
	s->type = AGL_GRAMVAL;
	s->u.val.vtype = AGL_STYPELIST;
	s->u.val.vu.list = NULL;	/* empty list */
	for (i = 0; i < lg; i++)
	{
		if ((ret = agl_grampop(anal, &v)) < 0)
		{
			agl_error("stack empty");
			return -1;
		}
		agl_rvassert(v->type == AGL_GRAMVAL, -1);
		v->u.val.pnext = s->u.val.vu.list;
		s->u.val.vu.list = &v->u.val;
	}
	return agl_grampush(anal, s);
}

int
agl_indicescrnpush(agl_analyse * anal)
{
	agl_graminstr *s, *cur;
	agl_graminstr *v;
	int lg, i;
	int ret;

	agl_rvassert(anal != NULL, -1);

	s = anal->codeptr->pnext;
	agl_rvassert(s != NULL, -1);

	anal->codeptr->pnext = s->pnext;

	agl_rvassert(s->type == AGL_GRAMLSTSZ, -1);
	lg = s->u.lstsz;
	s->type = AGL_GRAMVAL;
	s->u.val.vtype = AGL_STYPELIST;
	s->u.val.vu.list = NULL;	/* empty list */
	cur = s;
	for (i = 0; i < lg; i++)
	{
		if ((ret = agl_grampop(anal, &v)) < 0)
		{
			agl_error("stack empty");
			return -1;
		}
		agl_rvassert(v->type == AGL_GRAMVAL, -1);
		v->u.val.pnext = NULL;
		cur->u.val.vu.list = &v->u.val;
		cur = v;
	}
	return agl_grampush(anal, s);
}

int
agl_keywindex(agl_analyse * anal)
{
	agl_graminstr *idx;
	agl_graminstr *v;
	agl_elemtype *pidx, *pel, *pnext;
	int ret;

	agl_rvassert(anal != NULL, -1);

	if ((ret = agl_grampop(anal, &v)) < 0)
	{
		agl_error("stack empty");
		return -1;
	}
	agl_rvassert(v->type == AGL_GRAMVAL, -1);

	if ((ret = agl_grampop(anal, &idx)) < 0)
	{
		agl_error("stack empty");
		return -1;
	}
	agl_rvassert(idx->type == AGL_GRAMVAL, -1);

	agl_rvassert((idx->u.val.vtype & AGL_STYPELIST) && ((idx->u.val.itype & AGL_STYPEALL) == AGL_STYPENONE), -1);

	pel = &v->u.val;

	pidx = &idx->u.val;

	pidx = pidx->vu.list;

	while (pidx != NULL)
	{

		agl_rvassert((pidx->itype & AGL_STYPEALL) == AGL_STYPENONE, -1);

		pel->itype = pidx->vtype;
		pel->iu = pidx->vu;
		pidx->vtype = AGL_STYPELIST;
		pidx->vu.list = pel;
		pnext = pidx->pnext;
		pidx->pnext = NULL;

		pel = pidx;
		pidx = pnext;
	}
	return agl_grampush(anal, v);
}

int
agl_keywassign(agl_analyse * anal)
{
	agl_graminstr *s;
	agl_graminstr *v;
	agl_object *pobj;
	int ret, idx;
	agl_modelem *pmod;

	agl_rvassert(anal != NULL, -1);

	s = anal->codeptr->pnext;

	agl_rvassert(s != NULL, -1);

	anal->codeptr->pnext = s->pnext;

	agl_rvassert(s->type == AGL_GRAMKEYW, -1);

	if ((ret = agl_grampop(anal, &v)) < 0)
	{
		agl_error("stack empty");
		return -1;
	}
	agl_rvassert(v->type == AGL_GRAMVAL, -1);

	for (pmod = anal->curmodule->pmodelem; pmod->name; pmod++)
		if (s->u.keyw->index == pmod->index)
			idx = pmod->index;

	if (pmod->name == NULL)
	{
		agl_error("indexdoes not exist");
		return -1;
	}

	if (v->u.val.vtype & s->u.keyw->type & AGL_STYPEALL)
	{			/* there is a common type between value and element AND index value and element */
		pobj = anal->cursymb->val.vu.user;
		agl_rvassert(pobj != NULL, -1);

		if (s->u.keyw->gram_valid != NULL)
		{
			ret = s->u.keyw->gram_valid(pobj, idx, &v->u.val);
			if (ret <= 0)
				return ret;
		}
		if (agl_objaddelem(pobj, idx, &v->u.val) <= 0)
		{
			agl_error3("cannot add element for %s from %s cannot be created", s->u.keyw->name, anal->cursymb->name);
			return 0;
		}
	}
	else
	{
		agl_error3("invalid type of %s from %s cannot be created", s->u.keyw->name, anal->cursymb->name);
		return 0;
	}
	return 1;
}

int
agl_gramhash(agl_analyse * anal)
{
	agl_graminstr *h;
	agl_graminstr *v;
	int ret;

	agl_rvassert(anal != NULL, -1);

	if ((ret = agl_grampop(anal, &v)) < 0)
	{
		agl_error("stack empty");
		return -1;
	}
	agl_rvassert(v->type == AGL_GRAMVAL, -1);

	if ((ret = agl_grampop(anal, &h)) < 0)
	{
		agl_error("stack empty");
		return -1;
	}
	agl_rvassert(h->type == AGL_GRAMVAL, -1);

	v->u.val.iu = h->u.val.vu;
	v->u.val.itype = h->u.val.vtype;
	return agl_grampush(anal, v);
}

agl_graminstr *
agl_gramencode(agl_analyse * analyse, agl_graminstr * f)
{
	agl_graminstr *nptr;

	nptr = agl_gramnewinst(analyse);
	if (nptr == NULL)
		return NULL;
	nptr->type = f->type;
	nptr->u = f->u;		// copy a complete struct
	nptr->pnext = NULL;
	if (analyse->codeptr != NULL)
		analyse->codeptr->pnext = nptr;
	else
		analyse->codelist = nptr;

	analyse->codeptr = nptr;

	return analyse->codeptr;
}

static int
agl_gramexecute(agl_analyse * analyse)
{
	int ret;

	for (analyse->codeptr = analyse->codelist; analyse->codeptr; analyse->codeptr = analyse->codeptr->pnext)
	{
		if (analyse->codeptr->type != AGL_GRAMINSTR)
			return -1;
		if (analyse->codeptr->u.instr == GRAM_NONE)
			break;
		ret = (*(analyse->codeptr->u.instr)) (analyse);
		if (ret < 0)
			return -1;
	}
	return 0;
}

int
agl_gramparse(agl_handle handle, agl_ios * ops)
{
	agl_analyse anal;
	agl_container *ph;
	int ret;

	agl_rvassert(handle != NULL, -1);

	ph = (agl_container *) handle;

	ret = 0;
	if (agl_graminitcode(&anal, ph->moduleroot, ops) < 0)
		return -1;

	if (aglparse(&anal) != 0)
		ret = -1;
	else if (agl_gramexecute(&anal) < 0)
		ret = -1;
	else if (agl_cachesymb(&anal) < 0)
		ret = -1;

	if (agl_gramfreeinst(&anal, ret == -1) < 0)
		ret = -1;

	if (agl_deletesymb(&anal, ret == -1) < 0)
		ret = -1;

	return ret;
}
