
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/*! \file
 * \ingroup module
 * \short this is the elementary type module
 * it manages the elementary types: strings, integer, double.
 * it converts values from one type to another one.
 */

#include <string.h>
#include <stdarg.h>
#include "config.h"
#include "agl_private.h"

static int
agl_elemload(agl_object * pobj)
{
	agl_elemtype *elem, *defn;

	agl_rvassert((pobj != NULL), -1);

	if (pobj->flags & AGL_OBJ_LOADED)
		return 1;

	if ((pobj->flags & AGL_OBJ_VALID) == 0)
		return 0;

	defn = pobj->defelem;
	agl_rvassert(defn != NULL, -1);

	elem = (agl_elemtype *) agl_malloc(sizeof(agl_elemtype));
	if (elem == NULL)
	{
		agl_memerr();
		return (-1);
	}
	memcpy(elem, defn, sizeof(agl_elemtype));

	pobj->priv = elem;

	pobj->flags |= AGL_OBJ_LOADED;
	return 1;
}


//! treats events

/*! an object receives an event. we must treat it
 * \param pcur a pointer to the current object
 * \param event the event that occurs (please consult the defines)
 * \param name the name of the object concerned by 
 * \param pobj a pointer to the object concerned by or 
 * \return
 * - -1 if a fatal error occurs
 * - 0 if it fails but without fatal error
 * - 1 if success
 */


static int
agl_elemdepevt(agl_object * pcur, int event, char *name, agl_object * pobj)
{
	agl_rvassert(pcur != NULL, -1);
	agl_rvassert(pobj != NULL, -1);
	agl_rvassert(name != NULL, -1);

	// nothing to do
	switch (event)
	{
	case AGL_OBJEVT_FLUSHED:	// nothing to do this is a leaf object
		break;
	case AGL_OBJEVT_UPDATE:
		return agl_elemload(pcur);
		break;
	case AGL_OBJEVT_UNLOAD:
		agl_free(pcur->priv);
		pcur->priv = NULL;
		pcur->flags &= ~AGL_OBJ_LOADED;
		break;
	default:
		agl_error2("event %d not catched by the module", event);
		return 0;
	}
	return 1;
}

static agl_modelem agl_elemmodtab[] = {
	{"VAL", 1, NULL, AGL_STYPESTR | AGL_STYPEDBL | AGL_STYPEINT},
	{NULL, 0, 0, 0}
};

agl_module agl_modtab_elemmod = {
	"ELEM",
	"Elementary types",
	AGL_MOD_MOD,
	{{NULL, NULL}},
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	agl_elemdepevt,
	NULL,
	NULL,
	NULL,
	agl_elemmodtab,
	NULL,
	NULL,
	NULL
};

int
agl_typeidx(agl_object * psrc, int nb, ...)
{
	int i;
	char *name;
	va_list ap;

	va_start(ap, nb);
	for (i = 0; i < nb; i++)
	{
		name = va_arg(ap, char *);

		if (strcmp(name, psrc->pmod->name) == 0)
		{
			return i + 1;
		}
	}
	return 0;
}

int
agl_getelem(agl_handle handle, agl_elemtype * psrc, agl_elemtype * pdst)
{
	agl_object *pobj;
	agl_elemtype *ptmp;

	agl_rvassert(psrc != NULL, -1);
	agl_rvassert(pdst != NULL, -1);

	pdst->vtype = psrc->vtype;
	switch (psrc->vtype & AGL_STYPEALL)
	{
	case AGL_STYPENONE:
		break;
	case AGL_STYPESTR:
		pdst->vu.str = psrc->vu.str;
		break;
	case AGL_STYPEDBL:
		pdst->vu.dbl = psrc->vu.dbl;
		break;
	case AGL_STYPEINT:
		pdst->vu.num = psrc->vu.num;
		break;
	case AGL_STYPELIST:
		pdst->vu.list = psrc->vu.list;
		break;
	case AGL_STYPEOBJREF:
	case AGL_STYPEUSER:
		if (psrc->vtype == AGL_STYPEOBJREF)
		{
			if (agl_cacheobjget(psrc->vu.ref.name, handle, &pobj) != 1)
			{
				agl_error2("object %s not found or error", psrc->vu.ref.name);
				return -1;
			}
		}
		else
			pobj = psrc->vu.user;
		if (strcmp(pobj->pmod->name, agl_modtab_elemmod.name) == 0)
		{
			ptmp = (agl_elemtype *) pobj->priv;
			switch (ptmp->vtype & AGL_STYPEALL)
			{
			case AGL_STYPESTR:
				pdst->vu.str = ptmp->vu.str;
				break;
			case AGL_STYPEDBL:
				pdst->vu.dbl = ptmp->vu.dbl;
				break;
			case AGL_STYPEINT:
				pdst->vu.num = ptmp->vu.num;
				break;
			default:
				agl_error2("object %s is not a string, double or integer", psrc->vu.ref.name);
				return -1;
			}
			pdst->vtype = ptmp->vtype;
		}
		else
			pdst->vu.user = pobj;
		break;
	default:
		agl_error2("object has unknown type: %d ", psrc->vtype);
		return -1;
	}
	pdst->itype = psrc->itype;
	switch (psrc->itype & AGL_STYPEALL)
	{
	case AGL_STYPENONE:
		break;
	case AGL_STYPESTR:
		pdst->iu.str = psrc->iu.str;
		break;
	case AGL_STYPEDBL:
		pdst->iu.dbl = psrc->iu.dbl;
		break;
	case AGL_STYPEINT:
		pdst->iu.num = psrc->iu.num;
		break;
	case AGL_STYPELIST:
		pdst->iu.list = psrc->iu.list;
		break;
	case AGL_STYPEOBJREF:
	case AGL_STYPEUSER:
		if (psrc->vtype == AGL_STYPEOBJREF)
		{
			if (agl_cacheobjget(psrc->iu.ref.name, handle, &pobj) != 1)
			{
				agl_error2("object %s not found or error", psrc->iu.ref.name);
				return -1;
			}
		}
		else
			pobj = psrc->iu.user;
		if (strcmp(pobj->pmod->name, agl_modtab_elemmod.name) == 0)
		{
			ptmp = (agl_elemtype *) pobj->priv;
			switch (ptmp->vtype & AGL_STYPEALL)
			{
			case AGL_STYPESTR:
				pdst->iu.str = ptmp->vu.str;
				break;
			case AGL_STYPEDBL:
				pdst->iu.dbl = ptmp->vu.dbl;
				break;
			case AGL_STYPEINT:
				pdst->iu.num = ptmp->vu.num;
				break;
			default:
				agl_error2("object %s is not a string, double or integer", psrc->vu.ref.name);
				return -1;
			}
			pdst->vtype = ptmp->itype;
		}
		else
			pdst->iu.user = pobj;
		break;
	default:
		agl_error2("object has unknown type: %d ", psrc->itype);
		return -1;
	}
	return 1;
}
