
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

//! manipulation of channels of a pixmap

/*!\file agl_channel.c
 * \author Stephane APIOU
 * \ingroup video
 * this file manipulates channels and the pixmaps associated
 */

#include <string.h>
#include "config.h"
#include "agl_private.h"

//! this routine create a new channel structure

/*! if width or height are nul, the channel is created with no
 * data allocation (because there is no bitmap to store).
 * \param width the width of the channel to create
 * \param height the height of the channel to create
 * \return 
 * - NULL if an error occurs
 * - a pointer to the channel if success
 */


agl_channel *
agl_channew(int width, int height)
{
	agl_channel *channel;

	channel = (agl_channel *) agl_malloc(sizeof(agl_channel));
	if (channel == NULL)
	{
		agl_memerr();
		return NULL;
	}
	channel->size.height = height;
	channel->size.width = width;
	channel->row_size = width;
	channel->pixels = NULL;
	if (width * height)
	{
		channel->pixels = (unsigned char *) agl_malloc(height * width);
		if (channel->pixels == NULL)
		{
			agl_memerr();
			agl_free(channel);
			return NULL;
		}
	}
	return channel;
}

//! this routine destroys the channel \a ch

/*! it frees the space allocated for the bitmap associated.
 * \param ch the pointer to the channel to destroy
 * \return 
 * nothing
 */

void
agl_chanfree(agl_channel * ch)
{
	agl_channel *channel;

	channel = (agl_channel *) ch;

	agl_rassert(channel != NULL);

	if (channel->pixels != NULL)
		agl_free(channel->pixels);

	agl_free(channel);
}

agl_channel *
agl_chandup(agl_channel * channel)
{
	agl_channel *cpy;


	agl_rvassert(channel != NULL, NULL);

	cpy = (agl_channel *) agl_malloc(sizeof(agl_channel));
	if (cpy == NULL)
	{
		agl_memerr();
		return NULL;
	}
	cpy->size.height = channel->size.height;
	cpy->size.width = channel->size.width;
	cpy->row_size = channel->row_size;
	cpy->pixels = NULL;
	cpy->pixels = (unsigned char *) agl_malloc(channel->size.height * channel->row_size);
	if (channel->pixels == NULL)
	{
		agl_memerr();
		agl_free(channel);
		return NULL;
	}
	memcpy(cpy->pixels, channel->pixels, channel->row_size * channel->size.height);

	return cpy;
}

//! this routine gets channels from a pixmap

/*! this routine gets the channels from a pixmap.
 * the manner the channel are separated is specified by
 * the \a type parameter.
 * all the channel pointers must point to nothing at the entry
 * of this function. The objects pointed by are allocated in this function.
 * the dimension of these objects are those of the pix parameter.
 * if one size is null, the 4 channels are allocated but their size is null
 * and no copy are done.
 * \param type indicates how the channel are separated
 * \param rect select the zone to convert fron pix to channels
 * \param pix the source pixmap that generates the channels
 * \param alpha contains the alpha channel
 * \param channel1 the first channel (depends on the type of separation)
 * - red channel for a RGB separation
 * - cyan channel for a CMY separation
 * - hue channel for a HSV separation
 * - the mean of the three channels
 * \param channel2 the second channel (depends on the type of separation)
 * - green channel for a RGB separation
 * - magenta channel for a CMY separation
 * - saturation channel for a HSV separation
 * \param channel3 the second channel (depends on the type of separation)
 * - blue channel for a RGB separation
 * - yellow channel for a CMY separation
 * - value channel for a HSV separation
 * \return 
 * - -1 if an error occurs
 * - 0 if nothing bad appears
 */

int
agl_frompixmap(int type, agl_rect * rect, agl_pix * pix, agl_channel ** alpha, agl_channel ** ch1, agl_channel ** ch2,
	       agl_channel ** ch3)
{
	unsigned char *pa, *p1, *p2, *p3, *ppa, *pp1, *pp2, *pp3;
	unsigned char r, g, b, v;
	short i, f, j;
	unsigned char *ppix;
	unsigned long *pppix;
	unsigned long x, y, w, h, xbeg, ybeg;

	pa = p1 = p2 = p3 = NULL;
	agl_rvassert(pix != NULL, -1);
	if (rect != NULL)
	{
		xbeg = rect->x;
		ybeg = rect->y;
		w = rect->width;
		h = rect->height;
		if ((xbeg + w) >= pix->size.width)
			return -1;
		if ((ybeg + h) >= pix->size.height)
			return -1;
	}
	else
	{
		xbeg = 0;
		ybeg = 0;
		w = pix->size.width;
		h = pix->size.height;
	}
	if (alpha != NULL)
	{
		*alpha = agl_channew(w, h);
		if (*alpha == NULL)
		{
			agl_memerr();
			return -1;
		}
		pa = (*alpha)->pixels;
	}
	if (ch1 != NULL)
	{
		*ch1 = agl_channew(w, h);
		if (*ch1 == NULL)
		{
			agl_memerr();
			agl_chanfree(*alpha);
			return -1;
		}
		p1 = (*ch1)->pixels;
	}
	if (ch2 != NULL)
	{
		*ch2 = agl_channew(w, h);
		if (*ch2 == NULL)
		{
			agl_memerr();
			agl_chanfree(*alpha);
			agl_chanfree(*ch1);
			return -1;
		}
		p2 = (*ch2)->pixels;
	}
	if (ch3 != NULL)
	{
		*ch3 = agl_channew(w, h);
		if (*ch3 == NULL)
		{
			agl_memerr();
			agl_chanfree(*alpha);
			agl_chanfree(*ch1);
			agl_chanfree(*ch2);
			return -1;
		}
		p3 = (*ch3)->pixels;
	}
	if ((pa == NULL) || (p1 == NULL) || (p2 == NULL) || (p3 == NULL) || (w == 0) || (h == 0))
		return 0;
	ppix = (unsigned char *) (pix->pixels + pix->row_size * ybeg + xbeg);
	agl_rvassert(ppix != NULL, -1);
	for (y = 0; y < h; y++, ppix = (unsigned char *) (pppix + pix->row_size))
	{
		pppix = (unsigned long *) ppix;
		ppa = pa;
		pp1 = p1;
		pp2 = p2;
		pp3 = p3;
		switch (type)
		{
		case AGL_RGB:
			for (x = 0; x < w; x++)
			{
				if (p1 != NULL)
					*p1++ = *ppix;
				ppix++;
				if (p2 != NULL)
					*p2++ = *ppix;
				ppix++;
				if (p3 != NULL)
					*p3++ = *ppix;
				ppix++;
				if (pa != NULL)
					*pa++ = *ppix;
				ppix++;
			}
			break;
		case AGL_CMY:
			for (x = 0; x < w; x++)
			{
				r = *ppix++;
				g = *ppix++;
				b = *ppix++;
				if (p1 != NULL)
					*p1++ = (r + b) / 2;
				ppix++;
				if (p2 != NULL)
					*p2++ = (r + g) / 2;
				ppix++;
				if (p3 != NULL)
					*p3++ = (g + b) / 2;
				ppix++;
				if (pa != NULL)
					*pa++ = *ppix;
				ppix++;
			}
			break;
		case AGL_HSV:
			for (x = 0; x < w; x++)
			{
				r = *ppix++;
				g = *ppix++;
				b = *ppix++;
				i = min3(r, g, b);
				v = max3(r, g, b);
				if (v != i)
				{
					f = (r == i) ? g - b : ((g == i) ? b - r : r - g);
					j = (r == i) ? 3 : ((g == i) ? 5 : 1);
					r = ((255 * 256 * j) - (255 * 256 * f) / (v - i) + 128) / 6 / 256;
					g = (255 * 256 * (v - i) + 128) / v / 256;
				}
				else
				{
					r = g = 0;
				}
				b = v;
				if (p1 != NULL)
					*p1++ = r;
				ppix++;
				if (p2 != NULL)
					*p2++ = g;
				ppix++;
				if (p3 != NULL)
					*p3++ = b;
				ppix++;
				if (pa != NULL)
					*pa++ = *ppix;
				ppix++;
			}
			break;
		case AGL_HLS:
			for (x = 0; x < w; x++)
			{
				r = *ppix++;
				g = *ppix++;
				b = *ppix++;
				i = min3(r, g, b);
				v = max3(r, g, b);
				if (v != i)
				{
					f = (r == i) ? g - b : ((g == i) ? b - r : r - g);
					j = (r == i) ? 3 : ((g == i) ? 5 : 1);
					r = ((255 * 256 * j) - (255 * 256 * f) / (v - i) + 128) / 6 / 256;
					if ((i + v) < 256)
						b = (256 * 255 * (v - i) + 128) / (v + i) / 256;
					else
						b = (256 * 255 * (v - i) + 128) / (511 - v - i) / 256;
				}
				else
				{
					b = r = 0;
				}
				g = (256 * (i + v) + 128) / 2 / 256;
				if (p1 != NULL)
					*p1++ = r;
				ppix++;
				if (p2 != NULL)
					*p2++ = g;
				ppix++;
				if (p3 != NULL)
					*p3++ = b;
				ppix++;
				if (pa != NULL)
					*pa++ = *ppix;
				ppix++;
			}
			break;

		case AGL_AVG:
			for (x = 0; x < w; x++)
			{
				v = (*ppix + ppix[1] + ppix[2] + 1) / 3;
				ppix += 3;
				if (p1 != NULL)
					*p1++ = v;
				if (pa != NULL)
					*pa++ = *ppix;
				ppix++;
			}
			break;
		default:
			agl_error("invalid type of transformation");
			return -1;
		}
		if (pa != NULL)
		{
			pa = ppa + (*alpha)->row_size;
			if (pa >= ((*alpha)->pixels + (*alpha)->row_size * (*alpha)->size.height))
				pa = (*alpha)->pixels;
		}
		if (p1 != NULL)
		{
			p1 = pp1 + (*ch1)->row_size;
			if (p1 >= ((*ch1)->pixels + (*ch1)->row_size * (*ch1)->size.height))
				p1 = (*ch1)->pixels;
		}
		if (p2 != NULL)
		{
			p2 = pp2 + (*ch2)->row_size;
			if (p2 >= ((*ch2)->pixels + (*ch2)->row_size * (*ch2)->size.height))
				p2 = (*ch2)->pixels;
		}
		if (p3 != NULL)
		{
			p3 = pp3 + (*ch3)->row_size;
			if (p3 >= ((*ch3)->pixels + (*ch3)->row_size * (*ch3)->size.height))
				p3 = (*ch3)->pixels;
		}
	}
	return 0;
}

//
//! this routine creates a pixmap from channels

/*! this routine creates pixmap from channels. The pixmap must already
 * be created and could contains some valid image.
 * the specified channels are substituted to the valid ones
 * if the channels are too small to fit into the pixmap, they are replicated
 * along the pixmap. if the channels are too big for the pixmap, the channels
 * are troncated.
 * if a channel parameter is a null pointer, the original pixmap channel is
 * preserved according to the type of channel combination
 * \param type indicates how the channel are combinated
 * \param rect the starting position and the size of the zone. if null, the
 * size of the pix zone is used
 * \param pix the source and destination pixmap that generates the channels
 * \param alpha contains the alpha channel
 * \param channel1 the first channel (depends on the type of separation)
 * - red channel for a RGB separation
 * - cyan channel for a CMY separation
 * - hue channel for a HSV separation
 * - the mean of the three RGB channels (i.e black and white values).
 * \param channel2 the second channel (depends on the type of separation)
 * - green channel for a RGB separation
 * - magenta channel for a CMY separation
 * - saturation channel for a HSV separation
 * \param channel3 the second channel (depends on the type of separation)
 * - blue channel for a RGB separation
 * - yellow channel for a CMY separation
 * - value channel for a HSV separation
 * \return 
 * - -1 if an error occurs
 * - 0 if nothing bad appears
 * \note special algorithm are implemented when all channels are defined and when the size
 * of all the channels and the pixmap are the same
 */

int
agl_topixmap(int type, agl_rect * rect, agl_pix * pix, agl_channel * alpha, agl_channel * ch1, agl_channel * ch2, agl_channel * ch3)
{
	int pw, ph, x, y, xpos, ypos;
	unsigned char r, g, b, v, h, s, l;
	short c, i, f, j;
	unsigned long *pr;
	unsigned char *pp, *pa, *p1, *p2, *p3, *ppa, *pp1, *pp2, *pp3;

	agl_rvassert(pix != NULL, -1);
	pw = pix->size.width;
	ph = pix->size.height;
	if (rect != NULL)
	{
		xpos = rect->x;
		if ((xpos < 0) || ((xpos + rect->width) >= pw))
			return 0;
		ypos = rect->y;
		if ((ypos < 0) || ((ypos + rect->height) >= ph))
			return 0;
		if (rect->width > pw)
			return 0;
		pw = rect->width;
		if (rect->height > ph)
			return 0;
		ph = rect->height;
	}
	else
	{
		xpos = 0;
		ypos = 0;
	}
	if ((pw == 0) || (ph == 0))
		return 0;
	pp = (unsigned char *) pix->pixels;
	agl_rvassert(pp != NULL, -1);
	if ((alpha != NULL) && (alpha->size.width != 0) && (alpha->size.height != 0))
		pa = alpha->pixels;
	else
		pa = NULL;
	if ((ch1 != NULL) && (ch1->size.width != 0) && (ch1->size.height != 0))
		p1 = ch1->pixels;
	else
		p1 = NULL;
	if ((ch2 != NULL) && (ch2->size.width != 0) && (ch2->size.height != 0))
		p2 = ch2->pixels;
	else
		p2 = NULL;
	if ((ch3 != NULL) && (ch3->size.width != 0) && (ch3->size.height != 0))
		p3 = ch3->pixels;
	else
		p3 = NULL;
	if ((pa == NULL) && (p1 == NULL) && (p2 == NULL) && (p3 == NULL))
		return 0;
	if ((pa != NULL) && (p1 != NULL) && (p2 != NULL) && (p3 != NULL))
	{
		if ((alpha->size.width == pw) && (alpha->size.height == ph) && (ch1->size.width == pw) && (ch1->size.height == ph)
		    && (ch2->size.width == pw) && (ch2->size.height == ph) && (ch3->size.width == pw) && (ch3->size.height == ph))
		{
			switch (type)
			{
			case AGL_RGB:
				for (y = 0; y < ph; y++)
				{
					pp = (unsigned char *) (pix->pixels + pix->row_size * (y + ypos) + xpos);
					pa = (alpha->pixels + alpha->row_size * y);
					p1 = (ch1->pixels + ch1->row_size * y);
					p2 = (ch2->pixels + ch2->row_size * y);
					p3 = (ch3->pixels + ch3->row_size * y);
					for (x = 0; x < pw; x++)
					{
						*pp++ = *p1++;
						*pp++ = *p2++;
						*pp++ = *p3++;
						*pp++ = *pa++;
					}
				}
				break;
			case AGL_CMY:
				for (y = 0; y < ph; y++)
				{
					pp = (unsigned char *) (pix->pixels + pix->row_size * (y + ypos) + xpos);
					pa = (alpha->pixels + alpha->row_size * y);
					p1 = (ch1->pixels + ch1->row_size * y);
					p2 = (ch2->pixels + ch2->row_size * y);
					p3 = (ch3->pixels + ch3->row_size * y);
					for (x = 0; x < pw; x++)
					{
						*pp++ = *p1 + *p3 - *p2;
						*pp++ = *p3 + *p2 - *p1;
						*pp++ = *p2++ + *p1++ - *p3++;
						*pp++ = *pa++;
					}
				}
				break;
			case AGL_HSV:
				for (y = 0; y < ph; y++)
				{
					pp = (unsigned char *) (pix->pixels + pix->row_size * (y + ypos) + xpos);
					pa = (alpha->pixels + alpha->row_size * y);
					p1 = (ch1->pixels + ch1->row_size * y);
					p2 = (ch2->pixels + ch2->row_size * y);
					p3 = (ch3->pixels + ch3->row_size * y);
					for (x = 0; x < pw; x++)
					{
						h = *p1++;
						if (h == 0)
						{
							p2++;
							*pp++ = *p3;
							*pp++ = *p3;
							*pp++ = *p3++;
						}
						else
						{
							s = *p2++;
							v = *p3++;
							j = (h * 6) / 255;
							f = 6 * h - j * 255;
							f = (j & 1) ? f : (255 - f);
							i = (v * (255 - s) + 127) / 255;
							c = (v * (65025 - s * f) + 32512) / 65025;
							switch (j)
							{
							case 6:
							case 0:
								*pp++ = v;
								*pp++ = c;
								*pp++ = i;
								break;
							case 1:
								*pp++ = c;
								*pp++ = v;
								*pp++ = i;
								break;
							case 2:
								*pp++ = i;
								*pp++ = v;
								*pp++ = c;
								break;
							case 3:
								*pp++ = i;
								*pp++ = c;
								*pp++ = v;
								break;
							case 4:
								*pp++ = c;
								*pp++ = i;
								*pp++ = v;
								break;
							case 5:
								*pp++ = v;
								*pp++ = i;
								*pp++ = c;
								break;
							}
						}
						*pp++ = *pa++;
					}
				}
				break;
			case AGL_HLS:
				for (y = 0; y < ph; y++)
				{
					pp = (unsigned char *) (pix->pixels + pix->row_size * (y + ypos) + xpos);
					pa = (alpha->pixels + alpha->row_size * y);
					p1 = (ch1->pixels + ch1->row_size * y);
					p2 = (ch2->pixels + ch2->row_size * y);
					p3 = (ch3->pixels + ch3->row_size * y);
					for (x = 0; x < pw; x++)
					{
						s = *p3++;
						if (s == 0)
						{
							p1++;
							*pp++ = *p2;
							*pp++ = *p2;
							*pp++ = *p2++;
						}
						else
						{
							h = *p1++;
							l = *p2++;
							if (l < 128)
								c = (l * (255 + s) + 127) / 255;
							else
								c = l + s - (l * s) / 255;

							i = 2 * l - c;
							j = (h * 6) / 255;
							f = 6 * h - j * 255;
							f = (j & 1) ? f : (255 - f);
							switch (j)
							{
							case 6:
							case 0:
								*pp++ = c;
								*pp++ = c - f * (c - i) / 255;
								*pp++ = i;
								break;
							case 1:
								*pp++ = c - f * (c - i) / 255;
								*pp++ = c;
								*pp++ = i;
								break;
							case 2:
								*pp++ = i;
								*pp++ = c;
								*pp++ = c - f * (c - i) / 255;
								break;
							case 3:
								*pp++ = i;
								*pp++ = c - f * (c - i) / 255;
								*pp++ = c;
								break;
							case 4:
								*pp++ = c - f * (c - i) / 255;
								*pp++ = i;
								*pp++ = c;
								break;
							case 5:
								*pp++ = c;
								*pp++ = i;
								*pp++ = c - f * (c - i) / 255;
								break;
							}
						}
						*pp++ = *pa++;
					}
				}
				break;
			case AGL_AVG:
				for (y = 0; y < ph; y++)
				{
					pp = (unsigned char *) (pix->pixels + pix->row_size * (y + ypos) + xpos);
					pa = (alpha->pixels + alpha->row_size * y);
					p1 = (ch1->pixels + ch1->row_size * y);
					p2 = (ch2->pixels + ch2->row_size * y);
					p3 = (ch3->pixels + ch3->row_size * y);
					for (x = 0; x < pw; x++)
					{
						*pp++ = *p1++;
						*pp++ = *p1++;
						*pp++ = *p1++;
					}
				}
				break;
			default:
				agl_error("invalid type of transformation");
				return -1;
			}
		}
		else
		{
			pp = (unsigned char *) (pix->pixels + pix->row_size * ypos + xpos);
			for (y = 0; y < ph; y++, pp = (unsigned char *) (pr + pix->row_size))
			{
				pr = (unsigned long *) pp;
				ppa = pa;
				pp1 = p1;
				pp2 = p2;
				pp3 = p3;
				for (x = 0; x < pw; x++)
				{
					switch (type)
					{
					case AGL_RGB:
						*pp++ = *p1++;
						if (p1 >= (pp1 + ch1->size.width))
							p1 = pp1;
						*pp++ = *p2++;
						if (p2 >= (pp2 + ch2->size.width))
							p2 = pp2;
						*pp++ = *p3++;
						if (p3 >= (pp3 + ch2->size.width))
							p3 = pp3;
						break;
					case AGL_CMY:
						h = *p1++;
						if (p1 >= (pp1 + ch1->size.width))
							p1 = pp1;
						v = *p2++;
						if (p2 >= (pp2 + ch2->size.width))
							p2 = pp2;
						f = *p3++;
						if (p3 >= (pp3 + ch2->size.width))
							p3 = pp3;
						*pp++ = h + f - v;
						*pp++ = f + v - h;
						*pp++ = v + h - f;
						break;
					case AGL_HSV:
						h = *p1++;
						if (p1 >= (pp1 + ch1->size.width))
							p1 = pp1;
						s = *p2++;
						if (p2 >= (pp2 + ch2->size.width))
							p2 = pp2;
						v = *p3++;
						if (p3 >= (pp3 + ch2->size.width))
							p3 = pp3;
						if (h == 0)
						{
							*pp++ = v;
							*pp++ = v;
							*pp++ = v;
						}
						else
						{
							j = (h * 6) / 255;
							f = 6 * h - j * 255;
							f = (j & 1) ? f : (255 - f);
							i = (v * (255 - s) + 127) / 255;
							c = (v * (65025 - s * f) + 32512) / 65025;
							switch (j)
							{
							case 6:
							case 0:
								*pp++ = v;
								*pp++ = c;
								*pp++ = i;
								break;
							case 1:
								*pp++ = c;
								*pp++ = v;
								*pp++ = i;
								break;
							case 2:
								*pp++ = i;
								*pp++ = v;
								*pp++ = c;
								break;
							case 3:
								*pp++ = i;
								*pp++ = c;
								*pp++ = v;
								break;
							case 4:
								*pp++ = c;
								*pp++ = i;
								*pp++ = v;
								break;
							case 5:
								*pp++ = v;
								*pp++ = i;
								*pp++ = c;
								break;
							}
						}
						break;
					case AGL_HLS:
						h = *p1++;
						if (p1 >= (pp1 + ch1->size.width))
							p1 = pp1;
						l = *p2++;
						if (p2 >= (pp2 + ch2->size.width))
							p2 = pp2;
						s = *p3++;
						if (p3 >= (pp3 + ch2->size.width))
							p3 = pp3;
						if (s == 0)
						{
							*pp++ = l;
							*pp++ = l;
							*pp++ = l;
						}
						else
						{
							if (l < 128)
								c = (l * (255 + s) + 127) / 255;
							else
								c = l + s - (l * s) / 255;

							i = 2 * l - c;
							j = (h * 6) / 255;
							f = 6 * h - j * 255;
							f = (j & 1) ? f : (255 - f);
							switch (j)
							{
							case 6:
							case 0:
								*pp++ = c;
								*pp++ = c - f * (c - i) / 255;
								*pp++ = i;
								break;
							case 1:
								*pp++ = c - f * (c - i) / 255;
								*pp++ = c;
								*pp++ = i;
								break;
							case 2:
								*pp++ = i;
								*pp++ = c;
								*pp++ = c - f * (c - i) / 255;
								break;
							case 3:
								*pp++ = i;
								*pp++ = c - f * (c - i) / 255;
								*pp++ = c;
								break;
							case 4:
								*pp++ = c - f * (c - i) / 255;
								*pp++ = i;
								*pp++ = c;
								break;
							case 5:
								*pp++ = c;
								*pp++ = i;
								*pp++ = c - f * (c - i) / 255;
								break;
							}
						}
						break;

					case AGL_AVG:
						*pp++ = *p1++;
						*pp++ = *p1++;
						*pp++ = *p1++;
						if (p1 >= (pp1 + ch1->size.width))
							p1 = pp1;
						break;
					default:
						agl_error("invalid type of transformation");
						return -1;
					}
					*pp++ = *pa++;
					if (pa >= (ppa + alpha->size.width))
						pa = ppa;
				}
				pa = ppa + alpha->row_size;
				if (pa >= (alpha->pixels + alpha->row_size * alpha->size.height))
					pa = alpha->pixels;
				p1 = pp1 + ch1->row_size;
				if (p1 >= (ch1->pixels + ch1->row_size * ch1->size.height))
					p1 = ch1->pixels;
				p2 = pp2 + ch2->row_size;
				if (p2 >= (ch2->pixels + ch2->row_size * ch2->size.height))
					p2 = ch2->pixels;
				p3 = pp3 + ch3->row_size;
				if (p3 >= (ch3->pixels + ch3->row_size * ch3->size.height))
					p3 = ch3->pixels;
			}
		}
	}
	else
	{

		pp = (unsigned char *) (pix->pixels + pix->row_size * ypos + xpos);
		for (y = 0; y < ph; y++, pp = (unsigned char *) (pr + pix->row_size))
		{
			pr = (unsigned long *) pp;
			ppa = pa;
			pp1 = p1;
			pp2 = p2;
			pp3 = p3;
			for (x = 0; x < pw; x++)
			{
				switch (type)
				{
				case AGL_RGB:
					if (p1 != NULL)
					{
						*pp = *p1++;
						if (p1 >= (pp1 + ch1->size.width))
							p1 = pp1;
					}
					pp++;
					if (p2 != NULL)
					{
						*pp = *p2++;
						if (p2 >= (pp2 + ch2->size.width))
							p2 = pp2;
					}
					pp++;
					if (p3 != NULL)
					{
						*pp = *p3++;
						if (p3 >= (pp3 + ch2->size.width))
							p3 = pp3;
					}
					pp++;
					break;
				case AGL_CMY:
					h = (pp[0] + pp[2]) / 2;
					v = (pp[0] + pp[1]) / 2;
					f = (pp[1] + pp[2]) / 2;
					if (p1 != NULL)
					{
						h = *p1++;
						if (p1 >= (pp1 + ch1->size.width))
							p1 = pp1;
					}
					if (p2 != NULL)
					{
						v = *p2++;
						if (p2 >= (pp2 + ch2->size.width))
							p2 = pp2;
					}
					if (p3 != NULL)
					{
						f = *p3++;
						if (p3 >= (pp3 + ch2->size.width))
							p3 = pp3;
					}
					*pp++ = h + f - v;
					*pp++ = f + v - h;
					*pp++ = v + h - f;
					break;
				case AGL_HSV:
					r = pp[0];
					g = pp[1];
					b = pp[2];
					i = min3(r, g, b);
					v = max3(r, g, b);
					if (v != i)
					{
						f = (r == i) ? g - b : ((g == i) ? b - r : r - g);
						j = (r == i) ? 3 : ((g == i) ? 5 : 1);
						h = ((255 * 256 * j) - (255 * 256 * f) / (v - i) + 128) / 6 / 256;
						s = (255 * 256 * (v - i) + 128) / v / 256;
					}
					else
					{
						h = s = 0;
					}
					if (p1 != NULL)
					{
						h = *p1++;
						if (p1 >= (pp1 + ch1->size.width))
							p1 = pp1;
					}
					if (p2 != NULL)
					{
						s = *p2++;
						if (p2 >= (pp2 + ch2->size.width))
							p2 = pp2;
					}
					if (p3 != NULL)
					{
						v = *p3++;
						if (p3 >= (pp3 + ch2->size.width))
							p3 = pp3;
					}
					if (h == 0)
					{
						*pp++ = v;
						*pp++ = v;
						*pp++ = v;
					}
					else
					{
						j = (h * 6) / 255;
						f = 6 * h - j * 255;
						f = (j & 1) ? f : (255 - f);
						i = (v * (255 - s) + 127) / 255;
						c = (v * (65025 - s * f) + 32512) / 65025;
						switch (j)
						{
						case 6:
						case 0:
							*pp++ = v;
							*pp++ = c;
							*pp++ = i;
							break;
						case 1:
							*pp++ = c;
							*pp++ = v;
							*pp++ = i;
							break;
						case 2:
							*pp++ = i;
							*pp++ = v;
							*pp++ = c;
							break;
						case 3:
							*pp++ = i;
							*pp++ = c;
							*pp++ = v;
							break;
						case 4:
							*pp++ = c;
							*pp++ = i;
							*pp++ = v;
							break;
						case 5:
							*pp++ = v;
							*pp++ = i;
							*pp++ = c;
							break;
						}
					}
					break;
				case AGL_HLS:
					r = pp[0];
					g = pp[1];
					b = pp[2];
					i = min3(r, g, b);
					v = max3(r, g, b);
					if (v != i)
					{
						f = (r == i) ? g - b : ((g == i) ? b - r : r - g);
						j = (r == i) ? 3 : ((g == i) ? 5 : 1);
						h = ((255 * 256 * j) - (255 * 256 * f) / (v - i) + 128) / 6 / 256;
						if ((i + v) < 256)
							s = (256 * 255 * (v - i) + 128) / (v + i) / 256;
						else
							s = (256 * 255 * (v - i) + 128) / (511 - v - i) / 256;
					}
					else
					{
						s = h = 0;
					}
					l = (256 * (i + v) + 128) / 2 / 256;

					if (p1 != NULL)
					{
						h = *p1++;
						if (p1 >= (pp1 + ch1->size.width))
							p1 = pp1;
					}
					if (p2 != NULL)
					{
						l = *p2++;
						if (p2 >= (pp2 + ch2->size.width))
							p2 = pp2;
					}
					if (p3 != NULL)
					{
						s = *p3++;
						if (p3 >= (pp3 + ch2->size.width))
							p3 = pp3;
					}
					if (s == 0)
					{
						*pp++ = l;
						*pp++ = l;
						*pp++ = l;
					}
					else
					{
						if (l < 128)
							c = (l * (255 + s) + 127) / 255;
						else
							c = l + s - (l * s) / 255;

						i = 2 * l - c;
						j = (h * 6) / 255;
						f = 6 * h - j * 255;
						f = (j & 1) ? f : (255 - f);
						switch (j)
						{
						case 6:
						case 0:
							*pp++ = c;
							*pp++ = c - f * (c - i) / 255;
							*pp++ = i;
							break;
						case 1:
							*pp++ = c - f * (c - i) / 255;
							*pp++ = c;
							*pp++ = i;
							break;
						case 2:
							*pp++ = i;
							*pp++ = c;
							*pp++ = c - f * (c - i) / 255;
							break;
						case 3:
							*pp++ = i;
							*pp++ = c - f * (c - i) / 255;
							*pp++ = c;
							break;
						case 4:
							*pp++ = c - f * (c - i) / 255;
							*pp++ = i;
							*pp++ = c;
							break;
						case 5:
							*pp++ = c;
							*pp++ = i;
							*pp++ = c - f * (c - i) / 255;
							break;
						}
					}
					break;
				case AGL_AVG:
					if (p1 != NULL)
					{
						*pp++ = *p1++;
						*pp++ = *p1++;
						*pp++ = *p1++;
						if (p1 >= (pp1 + ch1->size.width))
							p1 = pp1;
					}
					break;
				default:
					agl_error("invalid type of transformation");
					return -1;
				}
				if (pa != NULL)
				{
					*pp = *pa++;
					if (pa >= (ppa + alpha->size.width))
						pa = ppa;
				}
				pp++;
			}
			if (pa != NULL)
			{
				pa = ppa + alpha->row_size;
				if (pa >= (alpha->pixels + alpha->row_size * alpha->size.height))
					pa = alpha->pixels;
			}
			if (p1 != NULL)
			{
				p1 = pp1 + ch1->row_size;
				if (p1 >= (ch1->pixels + ch1->row_size * ch1->size.height))
					p1 = ch1->pixels;
			}
			if (p2 != NULL)
			{
				p2 = pp2 + ch2->row_size;
				if (p2 >= (ch2->pixels + ch2->row_size * ch2->size.height))
					p2 = ch2->pixels;
			}
			if (p3 != NULL)
			{
				p3 = pp3 + ch3->row_size;
				if (p3 >= (ch3->pixels + ch3->row_size * ch3->size.height))
					p3 = ch3->pixels;
			}
		}
	}
	return 0;
}
