
/*
     ______     __                     ___     __
    /\  _  \   /\ \                   /\_ \   /\_\
    \ \ \_\ \  \ \ \      __     ___  \//\ \  \/_/_
     \ \  __ \ `/__ \   /'__`\  / __`\  \ \ \   /\ \
      \ \ \/\ \/\\_\ \_/\ \_\ \/\ \_\ \_ \_\ \_ \_\ \_
       \ \_\ \_\ \_____\ \___\ \ \______\/\____\/\____\
        \/_/\/_/\/_____/\/___L\ \/______/\/____/\/____/
                          /\____/
                          \_/__/




                  An advanced game library

                 (c) Stephane APIOU 2003


    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

/*!
 * \file agl_pix.c
 *
 * \ingroup video
 * \brief manipulation of pixmaps and blit routs
 *
 * this file manipulates pixmaps, blits and clip it
 *
 * $Id: agl_pix.c,v 1.9 2003/06/23 23:58:47 stef Exp $
 *
 * \author Stephane APIOU
 * \date   mar jui 30 01:40:18 CEST 2002
*/

#include <string.h>
#include "config.h"
#include "agl_private.h"

agl_pix *
agl_pixnew(int width, int height)
{
	unsigned long *pixels;
	agl_pix *ppix;

	ppix = (agl_pix *) agl_malloc(sizeof(agl_pix));
	if (ppix == NULL)
	{
		agl_memerr();
		return NULL;
	}

	if ((width == 0) || (height == 0))
		pixels = NULL;
	else
	{
		pixels = (unsigned long *) agl_malloc(width * height * sizeof(unsigned long));
		if (pixels == NULL)
		{
			agl_free(ppix);
			agl_memerr();
			return NULL;
		}
	}
	ppix->pixels = pixels;
	ppix->row_size = width;
	ppix->size.width = width;
	ppix->size.height = height;
	ppix->alpha = 255;

	return ppix;
}

void
agl_pixfree(agl_pix * ppix)
{

	agl_rassert(ppix != NULL);

	if (ppix->pixels != NULL)
		agl_free(ppix->pixels);

	agl_free(ppix);
}

agl_pix *
agl_pixdup(agl_pix * ppix)
{
	unsigned long *pixels;
	agl_pix *rpix;

	rpix = (agl_pix *) agl_malloc(sizeof(agl_pix));
	if (rpix == NULL)
	{
		agl_memerr();
		return NULL;
	}
	if (ppix->pixels == NULL)
		pixels = NULL;
	else
	{
		pixels = (unsigned long *) agl_malloc(ppix->row_size * ppix->size.height * sizeof(unsigned long));
		if (pixels == NULL)
		{
			agl_free(rpix);
			agl_memerr();
			return NULL;
		}
		memcpy(pixels, ppix->pixels, ppix->row_size * ppix->size.height * sizeof(unsigned long));
	}
	rpix->pixels = pixels;
	rpix->row_size = ppix->row_size;
	rpix->size.width = ppix->size.width;
	rpix->size.height = ppix->size.height;
	rpix->alpha = ppix->alpha;

	return rpix;

}

//! copy one zone into another

/*! copy one zone into another one with clipping of destination zone 
 *  	and source zone alpha of the destination zone is supposed to be opaque
 *	if psrcrect and pdstrect are NULL, no clipping is done
 *	alpha of original dest zone is 255 (opaque)
 *	apply transformation to source before copying to dest 
 * \param psrc the source zone to copy
 * \param pdst the destination zone
 * \param psrcrect the clipping rectangle for the source zone. This
 *  rectangle defines a clipping into the source zone. The coordonates
 *  refers into the source zone
 *  \param pdstrect the clipping rectangle for the destination zone.
 *  \param pos the position of the source zone into the destination zone.
 *  \param type the type of copy.
 *  \return 
 *  - -1 if an error occurs
 *  - 0 if nothing bad appears
 *  - 1 if success
 */

int
agl_pixtranscat(agl_pix * psrc, agl_pix * pdst, agl_rect * psrcrect, agl_rect * pdstrect, agl_pos * pos, int type)
{
	register unsigned long *pixdst, *pixsrc;
	register short n;
	register short nw, nh, sz, iter, y;
	register short srcrow, dstrow;

	short tx, ty, sw, sh;
	short sx, sy, dx, dy, ex, ey, dw, dh, bx, by;

	unsigned long *dpix, *spix;

	agl_rvassert(psrc != NULL, -1);

	agl_rvassert(pdst != NULL, -1);

	agl_rvassert(psrc->pixels != NULL, -1);

	agl_rvassert(pdst->pixels != NULL, -1);

	if (psrcrect == NULL)
	{
		sw = psrc->size.width;
		sh = psrc->size.height;
		sx = 0;
		sy = 0;
	}
	else
	{
		sw = psrcrect->width;
		sh = psrcrect->height;
		sx = psrcrect->x;
		sy = psrcrect->y;
	}
	if (pdstrect == NULL)
	{
		dw = pdst->size.width;
		dh = pdst->size.height;
		dx = 0;
		dy = 0;
	}
	else
	{
		dw = pdstrect->width;
		dh = pdstrect->height;
		dx = pdstrect->x;
		dy = pdstrect->y;
	}
	dstrow = pdst->row_size;
	srcrow = psrc->row_size;

	switch (type)
	{
	case PIX_ID:
		bx = max4(0, pos->x, pos->x + sx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.width, pos->x + sx + sw);
		nw = ex - bx;
		by = max4(0, pos->y, pos->y + sy, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.height, pos->y + sy + sh);
		nh = ey - by;
		if ((nw <= 0) || (nh <= 0))
			return 1;
		sx = bx - pos->x;
		sy = by - pos->y;

		dpix = pdst->pixels + by * dstrow + bx;
		spix = psrc->pixels + sy * srcrow + sx;
		for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = nw / 8; y < nh;
		     y++, pixdst = dpix += dstrow, pixsrc = spix += srcrow)
		{
			for (n = iter; n > 0; --n, pixdst += 8, pixsrc += 8)
			{
				pixdst[0] = pixsrc[0];
				pixdst[1] = pixsrc[1];
				pixdst[2] = pixsrc[2];
				pixdst[3] = pixsrc[3];
				pixdst[4] = pixsrc[4];
				pixdst[5] = pixsrc[5];
				pixdst[6] = pixsrc[6];
				pixdst[7] = pixsrc[7];
			}
			switch (sz)
			{
			case 7:
				pixdst[6] = pixsrc[6];
			case 6:
				pixdst[5] = pixsrc[5];
			case 5:
				pixdst[4] = pixsrc[4];
			case 4:
				pixdst[3] = pixsrc[3];
			case 3:
				pixdst[2] = pixsrc[2];
			case 2:
				pixdst[1] = pixsrc[1];
			case 1:
				pixdst[0] = pixsrc[0];
			case 0:
				break;
			}
		}
		break;
	case PIX_SYM0:
		tx = sx;
		ty = psrc->size.height - sh - sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.width, pos->x + tx + sw);
		nw = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.height, pos->y + ty + sh);
		nh = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = pdst->pixels + (by + nh - 1) * dstrow + bx;

		tx = bx - pos->x;
		ty = by - pos->y;
		sx = tx;
		sy = psrc->size.height - nh - ty;
		spix = psrc->pixels + sy * srcrow + sx;

		for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = nw / 8; y < nh;
		     y++, pixdst = dpix -= dstrow, pixsrc = spix += srcrow)
		{
			for (n = iter; n > 0; --n, pixdst += 8, pixsrc += 8)
			{
				pixdst[0] = pixsrc[0];
				pixdst[1] = pixsrc[1];
				pixdst[2] = pixsrc[2];
				pixdst[3] = pixsrc[3];
				pixdst[4] = pixsrc[4];
				pixdst[5] = pixsrc[5];
				pixdst[6] = pixsrc[6];
				pixdst[7] = pixsrc[7];
			}
			switch (sz)
			{
			case 7:
				pixdst[6] = pixsrc[6];
			case 6:
				pixdst[5] = pixsrc[5];
			case 5:
				pixdst[4] = pixsrc[4];
			case 4:
				pixdst[3] = pixsrc[3];
			case 3:
				pixdst[2] = pixsrc[2];
			case 2:
				pixdst[1] = pixsrc[1];
			case 1:
				pixdst[0] = pixsrc[0];
			case 0:
				break;
			}
		}
		break;
	case PIX_SYM45:
		ty = psrc->size.width - sw - sx;
		tx = psrc->size.height - sh - sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.height, pos->x + tx + sh);
		nh = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.width, pos->y + sw + ty);
		nw = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = pdst->pixels + (by + nw - 1) * dstrow + bx + nh - 1;

		tx += bx - pos->x;
		ty += by - pos->y;
		sx = psrc->size.width - nw - ty;
		sy = psrc->size.height - nh - tx;
		spix = psrc->pixels + sy * srcrow + sx;


		for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = nw / 8; y < nh;
		     y++, pixdst = dpix -= 1, pixsrc = spix += srcrow)
		{
			for (n = iter; n > 0; --n, pixsrc += 8)
			{
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				pixdst -= dstrow;
				*pixdst = pixsrc[2];
				pixdst -= dstrow;
				*pixdst = pixsrc[3];
				pixdst -= dstrow;
				*pixdst = pixsrc[4];
				pixdst -= dstrow;
				*pixdst = pixsrc[5];
				pixdst -= dstrow;
				*pixdst = pixsrc[6];
				pixdst -= dstrow;
				*pixdst = pixsrc[7];
				pixdst -= dstrow;
			}
			switch (sz)
			{
			case 7:
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				pixdst -= dstrow;
				*pixdst = pixsrc[2];
				pixdst -= dstrow;
				*pixdst = pixsrc[3];
				pixdst -= dstrow;
				*pixdst = pixsrc[4];
				pixdst -= dstrow;
				*pixdst = pixsrc[5];
				pixdst -= dstrow;
				*pixdst = pixsrc[6];
				break;
			case 6:
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				pixdst -= dstrow;
				*pixdst = pixsrc[2];
				pixdst -= dstrow;
				*pixdst = pixsrc[3];
				pixdst -= dstrow;
				*pixdst = pixsrc[4];
				pixdst -= dstrow;
				*pixdst = pixsrc[5];
				break;
			case 5:
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				pixdst -= dstrow;
				*pixdst = pixsrc[2];
				pixdst -= dstrow;
				*pixdst = pixsrc[3];
				pixdst -= dstrow;
				*pixdst = pixsrc[4];
				break;
			case 4:
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				pixdst -= dstrow;
				*pixdst = pixsrc[2];
				pixdst -= dstrow;
				*pixdst = pixsrc[3];
				break;
			case 3:
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				pixdst -= dstrow;
				*pixdst = pixsrc[2];
				break;
			case 2:
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				break;
			case 1:
				*pixdst = pixsrc[0];
				break;
			case 0:
				break;
			}
		}
		break;
	case PIX_SYM90:
		tx = psrc->size.width - sw - sx;
		ty = sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.width, pos->x + tx + sw);
		nw = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.height, pos->y + ty + sh);
		nh = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = pdst->pixels + by * dstrow + bx + nw - 1;

		tx += bx - pos->x;
		ty += by - pos->y;
		sx = psrc->size.width - nw - tx;
		sy = ty;
		spix = psrc->pixels + sy * srcrow + sx;


		for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = nw / 8; y < nh;
		     y++, pixdst = dpix += dstrow, pixsrc = spix += srcrow)
		{
			for (n = iter; n > 0; --n, pixdst -= 8, pixsrc += 8)
			{
				pixdst[0] = pixsrc[0];
				pixdst[-1] = pixsrc[1];
				pixdst[-2] = pixsrc[2];
				pixdst[-3] = pixsrc[3];
				pixdst[-4] = pixsrc[4];
				pixdst[-5] = pixsrc[5];
				pixdst[-6] = pixsrc[6];
				pixdst[-7] = pixsrc[7];
			}
			switch (sz)
			{
			case 7:
				pixdst[-6] = pixsrc[6];
			case 6:
				pixdst[-5] = pixsrc[5];
			case 5:
				pixdst[-4] = pixsrc[4];
			case 4:
				pixdst[-3] = pixsrc[3];
			case 3:
				pixdst[-2] = pixsrc[2];
			case 2:
				pixdst[-1] = pixsrc[1];
			case 1:
				pixdst[0] = pixsrc[0];
			case 0:
				break;
			}
		}
		break;
	case PIX_SYM135:
		ty = sx;
		tx = sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.height, pos->x + tx + sh);
		nh = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.width, pos->y + ty + sw);
		nw = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = pdst->pixels + by * dstrow + bx;

		tx += bx - pos->x;
		ty += by - pos->y;
		sx = ty;
		sy = tx;
		spix = psrc->pixels + sy * srcrow + sx;

		for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = nw / 8; y < nh;
		     y++, pixdst = dpix += 1, pixsrc = spix += srcrow)
		{
			for (n = iter; n > 0; --n, pixsrc += 8)
			{
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				pixdst += dstrow;
				*pixdst = pixsrc[2];
				pixdst += dstrow;
				*pixdst = pixsrc[3];
				pixdst += dstrow;
				*pixdst = pixsrc[4];
				pixdst += dstrow;
				*pixdst = pixsrc[5];
				pixdst += dstrow;
				*pixdst = pixsrc[6];
				pixdst += dstrow;
				*pixdst = pixsrc[7];
				pixdst += dstrow;
			}
			switch (sz)
			{
			case 7:
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				pixdst += dstrow;
				*pixdst = pixsrc[2];
				pixdst += dstrow;
				*pixdst = pixsrc[3];
				pixdst += dstrow;
				*pixdst = pixsrc[4];
				pixdst += dstrow;
				*pixdst = pixsrc[5];
				pixdst += dstrow;
				*pixdst = pixsrc[6];
				break;
			case 6:
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				pixdst += dstrow;
				*pixdst = pixsrc[2];
				pixdst += dstrow;
				*pixdst = pixsrc[3];
				pixdst += dstrow;
				*pixdst = pixsrc[4];
				pixdst += dstrow;
				*pixdst = pixsrc[5];
				break;
			case 5:
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				pixdst += dstrow;
				*pixdst = pixsrc[2];
				pixdst += dstrow;
				*pixdst = pixsrc[3];
				pixdst += dstrow;
				*pixdst = pixsrc[4];
				break;
			case 4:
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				pixdst += dstrow;
				*pixdst = pixsrc[2];
				pixdst += dstrow;
				*pixdst = pixsrc[3];
				break;
			case 3:
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				pixdst += dstrow;
				*pixdst = pixsrc[2];
				break;
			case 2:
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				break;
			case 1:
				*pixdst = pixsrc[0];
				break;
			case 0:
				break;
			}
		}
		break;
	case PIX_ROT90:
		ty = sx;
		tx = psrc->size.height - sh - sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.height, pos->x + tx + sh);
		nh = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.width, pos->y + ty + sw);
		nw = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = pdst->pixels + by * dstrow + bx + nh - 1;

		tx += bx - pos->x;
		ty += by - pos->y;
		sx = ty;
		sy = psrc->size.height - nh - tx;
		spix = psrc->pixels + sy * srcrow + sx;


		for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = nw / 8; y < nh;
		     y++, pixdst = dpix -= 1, pixsrc = spix += srcrow)
		{
			for (n = iter; n > 0; --n, pixsrc += 8)
			{
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				pixdst += dstrow;
				*pixdst = pixsrc[2];
				pixdst += dstrow;
				*pixdst = pixsrc[3];
				pixdst += dstrow;
				*pixdst = pixsrc[4];
				pixdst += dstrow;
				*pixdst = pixsrc[5];
				pixdst += dstrow;
				*pixdst = pixsrc[6];
				pixdst += dstrow;
				*pixdst = pixsrc[7];
				pixdst += dstrow;
			}
			switch (sz)
			{
			case 7:
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				pixdst += dstrow;
				*pixdst = pixsrc[2];
				pixdst += dstrow;
				*pixdst = pixsrc[3];
				pixdst += dstrow;
				*pixdst = pixsrc[4];
				pixdst += dstrow;
				*pixdst = pixsrc[5];
				pixdst += dstrow;
				*pixdst = pixsrc[6];
				break;
			case 6:
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				pixdst += dstrow;
				*pixdst = pixsrc[2];
				pixdst += dstrow;
				*pixdst = pixsrc[3];
				pixdst += dstrow;
				*pixdst = pixsrc[4];
				pixdst += dstrow;
				*pixdst = pixsrc[5];
				break;
			case 5:
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				pixdst += dstrow;
				*pixdst = pixsrc[2];
				pixdst += dstrow;
				*pixdst = pixsrc[3];
				pixdst += dstrow;
				*pixdst = pixsrc[4];
				break;
			case 4:
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				pixdst += dstrow;
				*pixdst = pixsrc[2];
				pixdst += dstrow;
				*pixdst = pixsrc[3];
				break;
			case 3:
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				pixdst += dstrow;
				*pixdst = pixsrc[2];
				break;
			case 2:
				*pixdst = pixsrc[0];
				pixdst += dstrow;
				*pixdst = pixsrc[1];
				break;
			case 1:
				*pixdst = pixsrc[0];
				break;
			case 0:
				break;
			}
		}

		break;
	case PIX_ROT180:
		tx = psrc->size.width - sw - sx;
		ty = psrc->size.height - sh - sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.width, pos->x + tx + sw);
		nw = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.height, pos->y + ty + sh);
		nh = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = pdst->pixels + (by + nh - 1) * dstrow + bx + nw - 1;

		tx += bx - pos->x;
		ty += by - pos->y;
		sx = psrc->size.width - nw - tx;
		sy = psrc->size.height - nh - ty;
		spix = psrc->pixels + sy * srcrow + sx;

		for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = nw / 8; y < nh;
		     y++, pixdst = dpix -= dstrow, pixsrc = spix += srcrow)
		{
			for (n = iter; n > 0; --n, pixdst -= 8, pixsrc += 8)
			{
				pixdst[0] = pixsrc[0];
				pixdst[-1] = pixsrc[1];
				pixdst[-2] = pixsrc[2];
				pixdst[-3] = pixsrc[3];
				pixdst[-4] = pixsrc[4];
				pixdst[-5] = pixsrc[5];
				pixdst[-6] = pixsrc[6];
				pixdst[-7] = pixsrc[7];
			}
			switch (sz)
			{
			case 7:
				pixdst[-6] = pixsrc[6];
			case 6:
				pixdst[-5] = pixsrc[5];
			case 5:
				pixdst[-4] = pixsrc[4];
			case 4:
				pixdst[-3] = pixsrc[3];
			case 3:
				pixdst[-2] = pixsrc[2];
			case 2:
				pixdst[-1] = pixsrc[1];
			case 1:
				pixdst[0] = pixsrc[0];
			case 0:
				break;
			}
		}
		break;
	case PIX_ROT270:
		ty = psrc->size.width - sw - sx;
		tx = sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.height, pos->x + tx + sh);
		nh = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.width, pos->y + ty + sw);
		nw = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = pdst->pixels + (by + nw - 1) * dstrow + bx;

		tx += bx - pos->x;
		ty += by - pos->y;
		sx = psrc->size.width - sw - ty;
		sy = tx;
		spix = psrc->pixels + sy * srcrow + sx;


		for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = nw / 8; y < nh;
		     y++, pixdst = dpix += 1, pixsrc = spix += srcrow)
		{
			for (n = iter; n > 0; --n, pixsrc += 8)
			{
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				pixdst -= dstrow;
				*pixdst = pixsrc[2];
				pixdst -= dstrow;
				*pixdst = pixsrc[3];
				pixdst -= dstrow;
				*pixdst = pixsrc[4];
				pixdst -= dstrow;
				*pixdst = pixsrc[5];
				pixdst -= dstrow;
				*pixdst = pixsrc[6];
				pixdst -= dstrow;
				*pixdst = pixsrc[7];
				pixdst -= dstrow;
			}
			switch (sz)
			{
			case 7:
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				pixdst -= dstrow;
				*pixdst = pixsrc[2];
				pixdst -= dstrow;
				*pixdst = pixsrc[3];
				pixdst -= dstrow;
				*pixdst = pixsrc[4];
				pixdst -= dstrow;
				*pixdst = pixsrc[5];
				pixdst -= dstrow;
				*pixdst = pixsrc[6];
				break;
			case 6:
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				pixdst -= dstrow;
				*pixdst = pixsrc[2];
				pixdst -= dstrow;
				*pixdst = pixsrc[3];
				pixdst -= dstrow;
				*pixdst = pixsrc[4];
				pixdst -= dstrow;
				*pixdst = pixsrc[5];
				break;
			case 5:
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				pixdst -= dstrow;
				*pixdst = pixsrc[2];
				pixdst -= dstrow;
				*pixdst = pixsrc[3];
				pixdst -= dstrow;
				*pixdst = pixsrc[4];
				break;
			case 4:
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				pixdst -= dstrow;
				*pixdst = pixsrc[2];
				pixdst -= dstrow;
				*pixdst = pixsrc[3];
				break;
			case 3:
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				pixdst -= dstrow;
				*pixdst = pixsrc[2];
				break;
			case 2:
				*pixdst = pixsrc[0];
				pixdst -= dstrow;
				*pixdst = pixsrc[1];
				break;
			case 1:
				*pixdst = pixsrc[0];
				break;
			case 0:
				break;
			}
		}
		break;
	default:
		agl_error("invalid type of transformation");
		return -1;
	}
	return 1;
}

//! copy one zone into another with alpha calculation

/*! copy one zone into another one with clipping of destination zone 
   	and source zone and apply alpha to zones
	if psrcrect and pdstrect are NULL, this is a real alpha copy
 	apply transformation to source before copying to dest
 * \param psrc the source zone to copy
 * \param pdst the destination zone
 * \param psrcrect the clipping rectangle for the source zone. This
 *  rectangle defines a clipping into the source zone. The coordonates
 *  refers into the source zone
 *  \param pdstrect the clipping rectangle for the destination zone.
 *  \param pos the position of the source zone into the destination zone.
 *  \param type the type of copy.
 *  \return 
 *  - -1 if an error occurs
 *  - 0 if nothing bad appears
 *  - 1 if success
 */

int
agl_pixtransalpha(agl_pix * psrc, agl_pix * pdst, agl_rect * psrcrect, agl_rect * pdstrect, agl_pos * pos, int type)
{
	register unsigned char *pixdst, *pixsrc;
	register unsigned char *pixdsttab[8];
	register int n;
	register long al, alpha;
	register int nw, nh, sz, iter, y;
	register unsigned long srcrow, dstrow;

	int tx, ty, sw, sh;
	int sx, sy, dx, dy, ex, ey, dw, dh, bx, by;

	unsigned char *dpix, *spix;

	agl_rvassert(psrc != NULL, -1);

	agl_rvassert(pdst != NULL, -1);

	agl_rvassert(psrc->pixels != NULL, -1);

	agl_rvassert(pdst->pixels != NULL, -1);

	if (psrcrect == NULL)
	{
		sw = psrc->size.width;
		sh = psrc->size.height;
		sx = 0;
		sy = 0;
	}
	else
	{
		sw = psrcrect->width;
		sh = psrcrect->height;
		sx = psrcrect->x;
		sy = psrcrect->y;
	}
	if (pdstrect == NULL)
	{
		dw = pdst->size.width;
		dh = pdst->size.height;
		dx = 0;
		dy = 0;
	}
	else
	{
		dw = pdstrect->width;
		dh = pdstrect->height;
		dx = pdstrect->x;
		dy = pdstrect->y;
	}
	dstrow = pdst->row_size;
	srcrow = psrc->row_size;
	alpha = psrc->alpha;

#define dopixalpha(dst, src) \
	al = (long)(src)[3] * alpha; \
	if (al > (255 << 8)) \
		*((unsigned long *)(dst)) = *((unsigned long *)(src)); \
	else \
	(dst)[0] = ((((long)(src)[0] - (dst)[0]) * al) >> 16) + (dst)[0], \
	(dst)[1] = ((((long)(src)[1] - (dst)[1]) * al) >> 16) + (dst)[1], \
	(dst)[2] = ((((long)(src)[2] - (dst)[2]) * al) >> 16) + (dst)[2], \
	(dst)[3] = (((long)((src)[3] - (dst)[3]) * al) >> 16) + (dst)[3]

	switch (type)
	{
	case PIX_ID:
		bx = max4(0, pos->x, pos->x + sx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.width, pos->x + sx + sw);
		nw = ex - bx;
		by = max4(0, pos->y, pos->y + sy, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.height, pos->y + sy + sh);
		nh = ey - by;
		if ((nw <= 0) || (nh <= 0))
			return 1;
		sx = bx - pos->x;
		sy = by - pos->y;
		dpix = (unsigned char *) (pdst->pixels + by * dstrow + bx);
		dpix = (unsigned char *) (pdst->pixels + by * dstrow + bx);
		spix = (unsigned char *) (psrc->pixels + sy * srcrow + sx);
		dstrow *= 4;
		srcrow *= 4;

		if (nw >= 8)
		{
			for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = (nw + 7) / 8; y < nh;
			     y++, pixdst = dpix += dstrow, pixsrc = spix += srcrow)
			{
				n = iter;
				switch (sz)
				{
				case 0:
					do
					{
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 7:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 6:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 5:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 4:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 3:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 2:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 1:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
					}
					while (--n > 0);
				}
			}
		}
		else
		{
			for (pixdst = dpix, pixsrc = spix, y = 0; y < nh; y++, pixdst += dstrow, pixsrc += srcrow)
			{
				switch (nw)
				{
				case 0:
					dopixalpha(pixdst + 4 * 7, pixsrc + 4 * 7);
				case 7:
					dopixalpha(pixdst + 4 * 6, pixsrc + 4 * 6);
				case 6:
					dopixalpha(pixdst + 4 * 5, pixsrc + 4 * 5);
				case 5:
					dopixalpha(pixdst + 4 * 4, pixsrc + 4 * 4);
				case 4:
					dopixalpha(pixdst + 4 * 3, pixsrc + 4 * 3);
				case 3:
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 2);
				case 2:
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 1);
				case 1:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 0);
				}
			}
		}
		break;
	case PIX_SYM0:
		tx = sx;
		ty = psrc->size.height - sh - sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.width, pos->x + tx + sw);
		nw = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.height, pos->y + ty + sh);
		nh = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = (unsigned char *) (pdst->pixels + (by + nh - 1) * dstrow + bx);

		tx = bx - pos->x;
		ty = by - pos->y;
		sx = tx;
		sy = psrc->size.height - nh - ty;
		spix = (unsigned char *) (psrc->pixels + sy * srcrow + sx);
		dstrow *= 4;
		srcrow *= 4;

		if (nw >= 8)
		{
			for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = (nw + 7) / 8; y < nh;
			     y++, pixdst = dpix -= dstrow, pixsrc = spix += srcrow)
			{
				n = iter;
				switch (sz)
				{
				case 0:
					do
					{
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 7:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 6:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 5:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 4:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 3:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 2:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
				case 1:
						dopixalpha(pixdst, pixsrc);
						pixdst += 4;
						pixsrc += 4;
					}
					while (--n > 0);
				}
			}
		}
		else
		{
			for (pixdst = dpix, pixsrc = spix, y = 0; y < nh; y++, pixdst -= dstrow, pixsrc += srcrow)
			{
				switch (nw)
				{
				case 0:
					dopixalpha(pixdst + 4 * 7, pixsrc + 4 * 7);
				case 7:
					dopixalpha(pixdst + 4 * 6, pixsrc + 4 * 6);
				case 6:
					dopixalpha(pixdst + 4 * 5, pixsrc + 4 * 5);
				case 5:
					dopixalpha(pixdst + 4 * 4, pixsrc + 4 * 4);
				case 4:
					dopixalpha(pixdst + 4 * 3, pixsrc + 4 * 3);
				case 3:
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 2);
				case 2:
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 1);
				case 1:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 0);
				}
			}
		}
		break;
	case PIX_SYM45:
		ty = psrc->size.width - sw - sx;
		tx = psrc->size.height - sh - sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.height, pos->x + tx + sh);
		nh = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.width, pos->y + sw + ty);
		nw = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = (unsigned char *) (pdst->pixels + (by + nw - 1) * dstrow + bx + nh - 1);

		tx = bx - pos->x;
		ty = by - pos->y;
		sx = psrc->size.width - nw - ty;
		sy = psrc->size.height - nh - tx;
		spix = (unsigned char *) (psrc->pixels + sy * srcrow + sx);
		dstrow *= 4;
		srcrow *= 4;


		if (nw >= 8)
		{
			for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = (nw + 7) / 8; y < nh;
			     y++, pixdst = dpix -= 4, pixsrc = spix += srcrow)
			{
				n = iter;
				switch (sz)
				{
				case 0:
					do
					{
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 7:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 6:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 5:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 4:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 3:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 2:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 1:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
					}
					while (--n > 0);
				}
			}
		}
		else
		{
			pixdsttab[0] = dpix;
			pixdsttab[1] = dpix -= dstrow;
			pixdsttab[2] = dpix -= dstrow;
			pixdsttab[3] = dpix -= dstrow;
			pixdsttab[4] = dpix -= dstrow;
			pixdsttab[5] = dpix -= dstrow;
			pixdsttab[6] = dpix -= dstrow;
			pixdsttab[7] = dpix -= dstrow;
			for (pixsrc = spix, y = 0; y < nh; y++, pixsrc += srcrow)
			{
				switch (nw)
				{
				case 0:
					dopixalpha(pixdsttab[7], pixsrc);
					pixdsttab[7] -= 4;
				case 7:
					dopixalpha(pixdsttab[6], pixsrc);
					pixdsttab[6] -= 4;
				case 6:
					dopixalpha(pixdsttab[5], pixsrc);
					pixdsttab[5] -= 4;
				case 5:
					dopixalpha(pixdsttab[4], pixsrc);
					pixdsttab[4] -= 4;
				case 4:
					dopixalpha(pixdsttab[3], pixsrc);
					pixdsttab[3] -= 4;
				case 3:
					dopixalpha(pixdsttab[2], pixsrc);
					pixdsttab[2] -= 4;
				case 2:
					dopixalpha(pixdsttab[1], pixsrc);
					pixdsttab[1] -= 4;
				case 1:
					dopixalpha(pixdsttab[0], pixsrc);
					pixdsttab[0] -= 4;
				}
			}
		}
		break;
	case PIX_SYM90:
		tx = psrc->size.width - sw - sx;
		ty = sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.width, pos->x + tx + sw);
		nw = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.height, pos->y + ty + sh);
		nh = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = (unsigned char *) (pdst->pixels + by * dstrow + bx + nw - 1);

		tx = bx - pos->x;
		ty = by - pos->y;
		sx = psrc->size.width - nw - tx;
		sy = ty;
		spix = (unsigned char *) (psrc->pixels + sy * srcrow + sx);
		dstrow *= 4;
		srcrow *= 4;


		if (nw >= 8)
		{
			for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = (nw + 7) / 8; y < nh;
			     y++, pixdst = dpix += dstrow, pixsrc = spix += srcrow)
			{
				n = iter;
				switch (sz)
				{
				case 0:
					do
					{
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 7:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 6:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 5:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 4:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 3:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 2:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 1:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
					}
					while (--n > 0);
				}
			}
		}
		else
		{
			for (pixdst = dpix, pixsrc = spix, y = 0; y < nh; y++, pixdst += dstrow, pixsrc += srcrow)
			{
				switch (nw)
				{
				case 0:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 7);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 6);
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 5);
					dopixalpha(pixdst + 4 * 3, pixsrc + 4 * 4);
					dopixalpha(pixdst + 4 * 4, pixsrc + 4 * 3);
					dopixalpha(pixdst + 4 * 5, pixsrc + 4 * 2);
					dopixalpha(pixdst + 4 * 6, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 7, pixsrc + 4 * 0);
					break;
				case 7:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 6);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 5);
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 4);
					dopixalpha(pixdst + 4 * 3, pixsrc + 4 * 3);
					dopixalpha(pixdst + 4 * 4, pixsrc + 4 * 2);
					dopixalpha(pixdst + 4 * 5, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 6, pixsrc + 4 * 0);
					break;
				case 6:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 5);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 4);
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 3);
					dopixalpha(pixdst + 4 * 3, pixsrc + 4 * 2);
					dopixalpha(pixdst + 4 * 4, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 5, pixsrc + 4 * 0);
					break;
				case 5:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 4);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 3);
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 2);
					dopixalpha(pixdst + 4 * 3, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 4, pixsrc + 4 * 0);
					break;
				case 4:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 3);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 2);
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 3, pixsrc + 4 * 0);
					break;
				case 3:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 2);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 0);
					break;
				case 2:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 0);
					break;
				case 1:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 0);
					break;
				}
			}
		}
		break;
	case PIX_SYM135:
		ty = sx;
		tx = sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.height, pos->x + tx + sh);
		nh = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.width, pos->y + ty + sw);
		nw = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = (unsigned char *) (pdst->pixels + by * dstrow + bx);

		tx = bx - pos->x;
		ty = by - pos->y;
		sx = ty;
		sy = tx;
		spix = (unsigned char *) (psrc->pixels + sy * srcrow + sx);
		dstrow *= 4;
		srcrow *= 4;

		if (nw >= 8)
		{
			for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = (nw + 7) / 8; y < nh;
			     y++, pixdst = dpix += 4, pixsrc = spix += srcrow)
			{
				n = iter;
				switch (sz)
				{
				case 0:
					do
					{
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 7:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 6:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 5:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 4:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 3:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 2:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 1:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
					}
					while (--n > 0);
				}
			}
		}
		else
		{
			pixdsttab[0] = dpix;
			pixdsttab[1] = dpix += dstrow;
			pixdsttab[2] = dpix += dstrow;
			pixdsttab[3] = dpix += dstrow;
			pixdsttab[4] = dpix += dstrow;
			pixdsttab[5] = dpix += dstrow;
			pixdsttab[6] = dpix += dstrow;
			pixdsttab[7] = dpix += dstrow;
			for (pixsrc = spix, y = 0; y < nh; y++, pixsrc += srcrow)
			{
				switch (nw)
				{
				case 0:
					dopixalpha(pixdsttab[7], pixsrc);
					pixdsttab[7] += 4;
				case 7:
					dopixalpha(pixdsttab[6], pixsrc);
					pixdsttab[6] += 4;
				case 6:
					dopixalpha(pixdsttab[5], pixsrc);
					pixdsttab[5] += 4;
				case 5:
					dopixalpha(pixdsttab[4], pixsrc);
					pixdsttab[4] += 4;
				case 4:
					dopixalpha(pixdsttab[3], pixsrc);
					pixdsttab[3] += 4;
				case 3:
					dopixalpha(pixdsttab[2], pixsrc);
					pixdsttab[2] += 4;
				case 2:
					dopixalpha(pixdsttab[1], pixsrc);
					pixdsttab[1] += 4;
				case 1:
					dopixalpha(pixdsttab[0], pixsrc);
					pixdsttab[0] += 4;
				}
			}
		}
		break;
	case PIX_ROT90:
		ty = sx;
		tx = psrc->size.height - sh - sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.height, pos->x + tx + sh);
		nh = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.width, pos->y + ty + sw);
		nw = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = (unsigned char *) (pdst->pixels + by * dstrow + bx + nh - 1);

		tx = bx - pos->x;
		ty = by - pos->y;
		sx = ty;
		sy = psrc->size.height - nh - tx;
		spix = (unsigned char *) (psrc->pixels + sy * srcrow + sx);
		dstrow *= 4;
		srcrow *= 4;

		if (nw >= 8)
		{
			for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = (nw + 7) / 8; y < nh;
			     y++, pixdst = dpix -= 4, pixsrc = spix += srcrow)
			{
				n = iter;
				switch (sz)
				{
				case 0:
					do
					{
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 7:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 6:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 5:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 4:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 3:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 2:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
				case 1:
						dopixalpha(pixdst, pixsrc);
						pixdst += dstrow;
						pixsrc += 4;
					}
					while (--n > 0);
				}
			}
		}
		else
		{
			pixdsttab[0] = dpix;
			pixdsttab[1] = dpix += dstrow;
			pixdsttab[2] = dpix += dstrow;
			pixdsttab[3] = dpix += dstrow;
			pixdsttab[4] = dpix += dstrow;
			pixdsttab[5] = dpix += dstrow;
			pixdsttab[6] = dpix += dstrow;
			pixdsttab[7] = dpix += dstrow;
			for (pixsrc = spix, y = 0; y < nh; y++, pixsrc += srcrow)
			{
				switch (nw)
				{
				case 0:
					dopixalpha(pixdsttab[7], pixsrc);
					pixdsttab[7] -= 4;
				case 7:
					dopixalpha(pixdsttab[6], pixsrc);
					pixdsttab[6] -= 4;
				case 6:
					dopixalpha(pixdsttab[5], pixsrc);
					pixdsttab[5] -= 4;
				case 5:
					dopixalpha(pixdsttab[4], pixsrc);
					pixdsttab[4] -= 4;
				case 4:
					dopixalpha(pixdsttab[3], pixsrc);
					pixdsttab[3] -= 4;
				case 3:
					dopixalpha(pixdsttab[2], pixsrc);
					pixdsttab[2] -= 4;
				case 2:
					dopixalpha(pixdsttab[1], pixsrc);
					pixdsttab[1] -= 4;
				case 1:
					dopixalpha(pixdsttab[0], pixsrc);
					pixdsttab[0] -= 4;
				}
			}
		}
		break;
	case PIX_ROT180:
		tx = psrc->size.width - sw - sx;
		ty = psrc->size.height - sh - sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.width, pos->x + tx + sw);
		nw = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.height, pos->y + ty + sh);
		nh = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = (unsigned char *) (pdst->pixels + (by + nh - 1) * dstrow + bx + nw - 1);

		tx = bx - pos->x;
		ty = by - pos->y;
		sx = psrc->size.width - nw - tx;
		sy = psrc->size.height - nh - ty;
		spix = (unsigned char *) (psrc->pixels + sy * srcrow + sx);
		dstrow *= 4;
		srcrow *= 4;


		if (nw >= 8)
		{
			for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = (nw + 7) / 8; y < nh;
			     y++, pixdst = dpix -= dstrow, pixsrc = spix += srcrow)
			{
				n = iter;
				switch (sz)
				{
				case 0:
					do
					{
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 7:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 6:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 5:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 4:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 3:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 2:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
				case 1:
						dopixalpha(pixdst, pixsrc);
						pixdst -= 4;
						pixsrc += 4;
					}
					while (--n > 0);
				}
			}
		}
		else
		{
			for (pixdst = dpix, pixsrc = spix, y = 0; y < nh; y++, pixdst -= dstrow, pixsrc += srcrow)
			{
				switch (nw)
				{
				case 0:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 7);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 6);
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 5);
					dopixalpha(pixdst + 4 * 3, pixsrc + 4 * 4);
					dopixalpha(pixdst + 4 * 4, pixsrc + 4 * 3);
					dopixalpha(pixdst + 4 * 5, pixsrc + 4 * 2);
					dopixalpha(pixdst + 4 * 6, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 7, pixsrc + 4 * 0);
					break;
				case 7:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 6);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 5);
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 4);
					dopixalpha(pixdst + 4 * 3, pixsrc + 4 * 3);
					dopixalpha(pixdst + 4 * 4, pixsrc + 4 * 2);
					dopixalpha(pixdst + 4 * 5, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 6, pixsrc + 4 * 0);
					break;
				case 6:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 5);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 4);
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 3);
					dopixalpha(pixdst + 4 * 3, pixsrc + 4 * 2);
					dopixalpha(pixdst + 4 * 4, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 5, pixsrc + 4 * 0);
					break;
				case 5:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 4);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 3);
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 2);
					dopixalpha(pixdst + 4 * 3, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 4, pixsrc + 4 * 0);
					break;
				case 4:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 3);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 2);
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 3, pixsrc + 4 * 0);
					break;
				case 3:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 2);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 2, pixsrc + 4 * 0);
					break;
				case 2:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 1);
					dopixalpha(pixdst + 4 * 1, pixsrc + 4 * 0);
					break;
				case 1:
					dopixalpha(pixdst + 4 * 0, pixsrc + 4 * 0);
					break;
				}
			}
		}
		break;
	case PIX_ROT270:
		ty = psrc->size.width - sw - sx;
		tx = sy;

		bx = max4(0, pos->x, pos->x + tx, dx);
		ex = min4(pdst->size.width, dx + dw, pos->x + psrc->size.height, pos->x + tx + sh);
		nh = ex - bx;
		by = max4(0, pos->y, pos->y + ty, dy);
		ey = min4(pdst->size.height, dy + dh, pos->y + psrc->size.width, pos->y + ty + sw);
		nw = ey - by;

		if ((nw <= 0) || (nh <= 0))
			return 1;
		dpix = (unsigned char *) (pdst->pixels + (by + nw - 1) * dstrow + bx);

		tx = bx - pos->x;
		ty = by - pos->y;
		sx = psrc->size.width - sw - ty;
		sy = tx;
		spix = (unsigned char *) (psrc->pixels + sy * srcrow + sx);
		dstrow *= 4;
		srcrow *= 4;

		if (nw >= 8)
		{
			for (pixdst = dpix, pixsrc = spix, y = 0, sz = nw % 8, iter = (nw + 7) / 8; y < nh;
			     y++, pixdst = dpix += 4, pixsrc = spix += srcrow)
			{
				n = iter;
				switch (sz)
				{
				case 0:
					do
					{
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 7:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 6:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 5:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 4:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 3:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 2:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
				case 1:
						dopixalpha(pixdst, pixsrc);
						pixdst -= dstrow;
						pixsrc += 4;
					}
					while (--n > 0);
				}
			}
		}
		else
		{
			pixdsttab[0] = dpix;
			pixdsttab[1] = dpix -= dstrow;
			pixdsttab[2] = dpix -= dstrow;
			pixdsttab[3] = dpix -= dstrow;
			pixdsttab[4] = dpix -= dstrow;
			pixdsttab[5] = dpix -= dstrow;
			pixdsttab[6] = dpix -= dstrow;
			pixdsttab[7] = dpix -= dstrow;
			for (pixsrc = spix, y = 0; y < nh; y++, pixsrc += srcrow)
			{
				switch (nw)
				{
				case 0:
					dopixalpha(pixdsttab[7], pixsrc);
					pixdsttab[7] += 4;
				case 7:
					dopixalpha(pixdsttab[6], pixsrc);
					pixdsttab[6] += 4;
				case 6:
					dopixalpha(pixdsttab[5], pixsrc);
					pixdsttab[5] += 4;
				case 5:
					dopixalpha(pixdsttab[4], pixsrc);
					pixdsttab[4] += 4;
				case 4:
					dopixalpha(pixdsttab[3], pixsrc);
					pixdsttab[3] += 4;
				case 3:
					dopixalpha(pixdsttab[2], pixsrc);
					pixdsttab[2] += 4;
				case 2:
					dopixalpha(pixdsttab[1], pixsrc);
					pixdsttab[1] += 4;
				case 1:
					dopixalpha(pixdsttab[0], pixsrc);
					pixdsttab[0] += 4;
				}
			}
		}
		break;
	default:
		agl_error("invalid type of transformation");
		return -1;
	}
	return 1;
}

int
agl_pixfill(agl_pix * ptr, agl_rect * rect, unsigned long color)
{
	register unsigned long *pixdst;
	register short n;
	register short nw, nh, sz, iter, y;
	register short dstrow;
	short dx, dy, ex, ey, dw, dh, bx, by;

	unsigned long *dpix;

	agl_rvassert(ptr != NULL, -1);

	agl_rvassert(ptr->pixels != NULL, -1);

	if (rect == NULL)
	{
		dw = ptr->size.width;
		dh = ptr->size.height;
		dx = 0;
		dy = 0;
	}
	else
	{
		dw = rect->width;
		dh = rect->height;
		dx = rect->x;
		dy = rect->y;
	}
	dstrow = ptr->row_size;

	bx = max(0, dx);
	ex = min(ptr->size.width, dx + dw);
	nw = ex - bx;
	by = max(0, dy);
	ey = min(ptr->size.height, dy + dh);
	nh = ey - by;
	if ((nw <= 0) || (nh <= 0))
		return 1;
	dpix = ptr->pixels + by * dstrow + bx;
	for (pixdst = dpix, y = 0, sz = nw % 8, iter = nw / 8; y < nh; y++, pixdst = dpix += dstrow)
	{
		for (n = iter; n > 0; --n, pixdst += 8)
		{
			pixdst[0] = pixdst[1] = pixdst[2] = pixdst[3] = pixdst[4] = pixdst[5] = pixdst[6] = pixdst[7] = color;
		}
		switch (sz)
		{
		case 7:
			pixdst[6] = color;
		case 6:
			pixdst[5] = color;
		case 5:
			pixdst[4] = color;
		case 4:
			pixdst[3] = color;
		case 3:
			pixdst[2] = color;
		case 2:
			pixdst[1] = color;
		case 1:
			pixdst[0] = color;
		case 0:
			break;
		}
	}
	return 1;
}

int
agl_pixset(agl_pix * ptr, agl_rect * rect, unsigned long mask, unsigned long color)
{
	register unsigned long *pixdst;
	register short n;
	register short nw, nh, sz, iter, y;
	register short dstrow;
	short dx, dy, ex, ey, dw, dh, bx, by;

	unsigned long *dpix;

	agl_rvassert(ptr != NULL, -1);

	agl_rvassert(ptr->pixels != NULL, -1);

	if (rect == NULL)
	{
		dw = ptr->size.width;
		dh = ptr->size.height;
		dx = 0;
		dy = 0;
	}
	else
	{
		dw = rect->width;
		dh = rect->height;
		dx = rect->x;
		dy = rect->y;
	}
	dstrow = ptr->row_size;

	bx = max(0, dx);
	ex = min(ptr->size.width, dx + dw);
	nw = ex - bx;
	by = max(0, dy);
	ey = min(ptr->size.height, dy + dh);
	nh = ey - by;
	if ((nw <= 0) || (nh <= 0))
		return 1;
	dpix = ptr->pixels + by * dstrow + bx;
	color = color & ~mask;
	for (pixdst = dpix, y = 0, sz = nw % 8, iter = nw / 8; y < nh; y++, pixdst = dpix += dstrow)
	{
		for (n = iter; n > 0; --n, pixdst += 8)
		{
			pixdst[0] = (pixdst[0] & mask) | color;
			pixdst[1] = (pixdst[1] & mask) | color;
			pixdst[2] = (pixdst[2] & mask) | color;
			pixdst[3] = (pixdst[3] & mask) | color;
			pixdst[4] = (pixdst[4] & mask) | color;
			pixdst[5] = (pixdst[5] & mask) | color;
			pixdst[6] = (pixdst[6] & mask) | color;
			pixdst[7] = (pixdst[7] & mask) | color;
		}
		switch (sz)
		{
		case 7:
			pixdst[6] = (pixdst[6] & mask) | color;
		case 6:
			pixdst[5] = (pixdst[5] & mask) | color;
		case 5:
			pixdst[4] = (pixdst[4] & mask) | color;
		case 4:
			pixdst[3] = (pixdst[3] & mask) | color;
		case 3:
			pixdst[2] = (pixdst[2] & mask) | color;
		case 2:
			pixdst[1] = (pixdst[1] & mask) | color;
		case 1:
			pixdst[0] = (pixdst[0] & mask) | color;
		case 0:
			break;
		}
	}
	return 1;
}

int
agl_pixkey(agl_pix * ptr, agl_rect * rect, unsigned long color)
{
	register unsigned long *pixdst;
	register short n;
	register short nw, nh, y;
	register short dstrow;
	short dx, dy, ex, ey, dw, dh, bx, by;

	unsigned long *dpix;

	agl_rvassert(ptr != NULL, -1);

	agl_rvassert(ptr->pixels != NULL, -1);

	if (rect == NULL)
	{
		dw = ptr->size.width;
		dh = ptr->size.height;
		dx = 0;
		dy = 0;
	}
	else
	{
		dw = rect->width;
		dh = rect->height;
		dx = rect->x;
		dy = rect->y;
	}
	dstrow = ptr->row_size;

	bx = max(0, dx);
	ex = min(ptr->size.width, dx + dw);
	nw = ex - bx;
	by = max(0, dy);
	ey = min(ptr->size.height, dy + dh);
	nh = ey - by;
	if ((nw <= 0) || (nh <= 0))
		return 1;
	dpix = ptr->pixels + by * dstrow + bx;
	color = color & ~AGL_ALPHA;
	for (pixdst = dpix, y = 0; y < nh; y++, pixdst = dpix += dstrow)
	{
		for (n = nw; n > 0; --n, pixdst++)
		{
			if ((*pixdst & ~AGL_ALPHA) == color)
				*pixdst = 0;
		}
	}
	return 1;
}
